//-----------------------------------------------------------------------------
//
//  $Logfile:: /GameSource/script.cpp                                         $
// $Revision:: 3                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 8/8/01 4:47 PM                                                 $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// C++ implementaion of tokenizing text interpretation.  Class accepts filename
// to load or pointer to preloaded text data.  Standard tokenizing operations
// such as skip white-space, get string, get integer, get float, get token,
// and skip line are implemented.  
//
// Note: all '//', '#', and ';' are treated as comments.  Probably should
// make this behaviour toggleable.
// 

#include "g_local.h"
#include "script.h"
#include "archive.h"

#define TOKENCOMMENT      (';')
#define TOKENCOMMENT2     ('#')
#define TOKENEOL          ('\n')
#define TOKENNULL         ('\0')
#define TOKENSPACE        (' ')

CLASS_DECLARATION( Class, Script, NULL );

ResponseDef Script::Responses[] =
{
	{ NULL, NULL }
};

Script::~Script()
{
	Close();
}

Script::Script()
{
	buffer			= NULL;
	script_p			= NULL;
	end_p				= NULL;
	line				= 0;
	releaseBuffer	= FALSE;
	tokenready		= FALSE;
	FastMemSet( token, 0, sizeof( token ) );
}

void Script::Close(void)
{
	if ( releaseBuffer && buffer ) {
		gi.TagFree( ( void * )buffer );
	}
	
	buffer			= NULL;
	script_p			= NULL;
	end_p				= NULL;
	line				= 0;
	releaseBuffer	= FALSE;
	tokenready		= FALSE;
	FastMemSet( token, 0, sizeof( token ) );
}

/*
==============
=
= Reset
=
==============
*/

void Script::Reset(void)
{
	script_p = buffer;
	line = 1;
	tokenready = FALSE;
}
	
/*
==============
=
= MarkPosition
=
==============
*/

void Script::MarkPosition(scriptmarker_t *mark)
{
	assert( mark );
	
	mark->tokenready = tokenready;
	mark->offset     = script_p - buffer;
	mark->line		  = line;
	Burger::StringCopy( mark->token, token );
}

/*
==============
=
= RestorePosition
=
==============
*/

void Script::RestorePosition(scriptmarker_t *mark)
{
	assert( mark );
	
	tokenready = mark->tokenready;
	script_p = buffer + mark->offset;
	line = mark->line;
	Burger::StringCopy( token, mark->token );
	
	assert( script_p <= end_p );
	if ( script_p > end_p ) {
		script_p = end_p;
	}
}

/*
==============
=
= SkipToEOL
=
==============
*/

qboolean Script::SkipToEOL(void)
{
	if ( script_p >= end_p ) {
		return TRUE;
	}
	
	while( *script_p != TOKENEOL ) {
		if ( script_p >= end_p ) {
			return TRUE;
		}
		script_p++;
	}
	return FALSE;
}

/*
==============
=
= CheckOverflow
=
==============
*/

void Script::CheckOverflow(void)
{
	if ( script_p >= end_p ) {
		gi.error( "End of token file reached prematurely reading %s\n", filename.c_str() );
	}
}

/*
==============
=
= SkipWhiteSpace
=
==============
*/

void Script::SkipWhiteSpace(qboolean crossline)
{
	//
	// skip space
	//
	CheckOverflow();
	
	while( *script_p <= TOKENSPACE ) {
		if ( *script_p++ == TOKENEOL ) {
			if ( !crossline ) {
				gi.error( "Line %i is incomplete in file %s\n", line, filename.c_str() );
			}
	
			line++;
		}
		CheckOverflow();
	}
}

qboolean Script::AtComment(void)
{
	if ( script_p >= end_p ) {
		return FALSE;
	}
	
	if ( *script_p == TOKENCOMMENT ) {
		return TRUE;
	}
	
	if ( *script_p == TOKENCOMMENT2 ) {
		return TRUE;
	}
	
	// Two or more character comment specifiers
	if ( ( script_p + 1 ) >= end_p ) {
		return FALSE;
	}
	
	if ( ( *script_p == '/' ) && ( *( script_p + 1 ) == '/' ) ) {
		return TRUE;
	}
	return FALSE;
}

/*
==============
=
= SkipNonToken
=
==============
*/

void Script::SkipNonToken(qboolean crossline)
{
	//
	// skip space and comments
	//
	SkipWhiteSpace( crossline );
	while( AtComment() ) {
		SkipToEOL();
		SkipWhiteSpace( crossline );
	}
}

/*
=============================================================================
=
= Token section
=
=============================================================================
*/

/*
==============
=
= TokenAvailable
=
==============
*/

qboolean Script::TokenAvailable (qboolean crossline)
{
	if ( script_p >= end_p ) {
		return FALSE;
	}
	
	while ( 1 ) {
		while ( *script_p <= TOKENSPACE ) {
			if ( *script_p == TOKENEOL ) {
				if ( crossline==FALSE ) {
					return( FALSE );
				}
				line++;
			}
	
			script_p++;
			if ( script_p >= end_p ) {
				return FALSE;
			}
		}
	
		if ( AtComment() ) {
			if ( SkipToEOL() ) {
				return FALSE;
			}
		} else {
			break;
		}
	}	
	return TRUE;
}

/*
==============
=
= CommentAvailable
=
==============
*/

qboolean Script::CommentAvailable(qboolean crossline)
{
	const char *searchptr;
	
	searchptr = script_p;
	
	if ( searchptr >= end_p ) {
		return FALSE;
	}
	
	while ( *searchptr <= TOKENSPACE ) {
		if ( ( *searchptr == TOKENEOL ) && ( !crossline ) ) {
			return FALSE;
		}
		searchptr++;
		if ( searchptr >= end_p ) {
			return FALSE;
		}
	}
	
	return TRUE;
}


/*
==============
=
= Get
=
==============
*/
qboolean Script::AtString(qboolean crossline)
{
	//
	// skip space
	//
	SkipNonToken( crossline );
	
	return ( *script_p == '"' );
}

/*
==============
=
= Get
=
==============
*/

const char *Script::GetToken(qboolean crossline)
{
	char *token_p;
	
	// is a token already waiting?
	if ( tokenready ) {
		tokenready = FALSE;
		return token;
	}
	
	//
	// skip space
	//
	SkipNonToken( crossline );
	
	//
	// copy token
	//
	
	if ( *script_p == '"' ) {
		return GetString( crossline );
	}
	
	token_p = token;	
	while( *script_p > TOKENSPACE && !AtComment() ) {
		if ( ( *script_p == '\\' ) && ( script_p < end_p - 1 ) ) {
			script_p++;
			switch( *script_p ) {
			case 'n' :	*token_p++ = '\n'; break;
			case 'r' :	*token_p++ = '\n'; break;
			case '\'' : *token_p++ = '\''; break;
			case '\"' : *token_p++ = '\"'; break;
			case '\\' : *token_p++ = '\\'; break;
			default:		*token_p++ = *script_p; break;
			}
			script_p++;
		} else {
			*token_p++ = *script_p++;
		}
	
		if ( token_p == &token[ MAXTOKEN ] ) {
			gi.error( "Token too large on line %i in file %s\n", line, filename.c_str() );
		}
	
		if ( script_p == end_p ) {
			break;
		}
	}
	
	*token_p = 0;
	
	return token;
}

/*
==============
=
= GetLine
=
==============
*/

const char *Script::GetLine(qboolean crossline)
{
	const char	*start;
	int			size;
	
	// is a token already waiting?
	if ( tokenready ) {
		tokenready = FALSE;
		return token;
	}
	
	//
	// skip space
	//
	SkipNonToken( crossline );
	
	//
	// copy token
	//
	start = script_p;
	SkipToEOL();
	size = script_p - start;
	if ( size < MAXTOKEN - 1 ) {
		FastMemCpy( token, start, size );
		token[ size ] = '\0';
	} else {
		gi.error( "Token too large on line %i in file %s\n", line, filename.c_str() );
	}
	return token;
}

/*
==============
=
= GetRaw
=
==============
*/

const char *Script::GetRaw(void)
{
	const char	*start;
	int size;
	
	//
	// skip white space
	//
	SkipWhiteSpace( TRUE );
	
	//
	// copy token
	//
	start = script_p;
	SkipToEOL();
	size = script_p - start;
	if ( size < MAXTOKEN - 1 ) {
		FastMemSet( token, 0, sizeof( token ) );
		FastMemCpy( token, start, size );
	} else {
		gi.error( "Token too large on line %i in file %s\n", line, filename.c_str() );
	}
	return token;
}

/*
==============
=
= GetString
=
==============
*/

const char *Script::GetString(qboolean crossline)
{
	int startline;
	char *token_p;
	
	// is a token already waiting?
	if ( tokenready ) {
		tokenready = FALSE;
		return token;
	}
	
	//
	// skip space
	//
	SkipNonToken( crossline );
	
	if ( *script_p != '"' ) {
		gi.error( "Expecting string on line %i in file %s\n", line, filename.c_str() );
	}
	
	script_p++;
	
	startline = line;
	token_p = token;
	while( *script_p != '"' ) {
		if ( *script_p == TOKENEOL ) {
			gi.error( "Line %i is incomplete while reading string in file %s\n", line, filename.c_str() );
		}
	
		if ( ( *script_p == '\\' ) && ( script_p < end_p - 1 ) ) {
			script_p++;
			switch( *script_p ) {
			case 'n' :	*token_p++ = '\n'; break;
			case 'r' :	*token_p++ = '\n'; break;
			case '\'' : *token_p++ = '\''; break;	
			case '\"' : *token_p++ = '\"'; break;	
			case '\\' : *token_p++ = '\\'; break;
			default:		*token_p++ = *script_p; break;
			}
			script_p++;
		} else {
			*token_p++ = *script_p++;
		}
		
		if ( script_p >= end_p ) {
			gi.error( "End of token file reached prematurely while reading string on\n"
				"line %d in file %s\n", startline, filename.c_str() );
		}
		
		if ( token_p == &token[ MAXTOKEN ] ) {
			gi.error( "String too large on line %i in file %s\n", line, filename.c_str() );
		}
	}
	
	*token_p = 0;
	
	// skip last quote
	script_p++;
	
	return token;
}

/*
==============
=
= GetSpecific
=
==============
*/

qboolean Script::GetSpecific(const char *string)
{
	do {
		if ( !TokenAvailable( TRUE ) ) {
			return FALSE;
		}
		GetToken( TRUE );
	} while( strcmp( token, string ) );
	return TRUE;
}

/*
==============
=
= GetInteger
=
==============
*/

int Script::GetInteger(qboolean crossline)
{
	GetToken( crossline );
	return atoi( token );
}

/*
==============
=
= GetDouble
=
==============
*/

double Script::GetDouble(qboolean crossline)
{
	GetToken( crossline );
	return (double)Sinatof( token );
}

/*
==============
=
= GetFloat
=
==============
*/

float Script::GetFloat(qboolean crossline)
{
	GetToken( crossline );
	return Sinatof( token );
}

/*
==============
=
= GetVector
=
==============
*/

void Script::GetVector(Vector *Output,qboolean crossline)
{
	Output->x = GetFloat(crossline);
	Output->y = GetFloat(crossline);
	Output->z = GetFloat(crossline);
}

/*
===================
=
= LinesInFile
=
===================
*/
int Script::LinesInFile(void)
{
	qboolean		temp_tokenready;
	const char	*temp_script_p;
	int			temp_line;		
	char			temp_token[ MAXTOKEN ];
	int			numentries;
	
	temp_tokenready = tokenready;
	temp_script_p	= script_p;
	temp_line		= line;
	Burger::StringCopy( temp_token, token );
	
	numentries = 0;
	
	Reset();
	while( TokenAvailable( TRUE ) ) {
		GetLine( TRUE );
		numentries++;
	}
	
	tokenready	= temp_tokenready;
	script_p	= temp_script_p;
	line		= temp_line;
	Burger::StringCopy( token, temp_token );
	
	return numentries;
}

/*
==============
=
= Parse
=
==============
*/

void Script::Parse(const char *data, int length,const char *name )
{
	Close();
	
	buffer = data;
	Reset();
	end_p = script_p + length;
	this->length = length;
	filename = name;
}

/*
==============
=
= Load
=
==============
*/

void Script::LoadFile(const char *name)
{
	int			length;
	const char	*buffer;
	
	Close();
	
	length = gi.LoadFile( name, ( void ** )&buffer, TAG_GAME );
	if ( length < 0 ) {
		error( "LoadFile", "Couldn't load %s\n", name );
	}
	Parse( buffer, length, name );
	releaseBuffer = TRUE;
}

void Script::Archive(Archiver &arc)
{
	Class::Archive( arc );
	
	arc.WriteCString( filename.c_str() );
	arc.WriteBoolean( tokenready );
	//
	// save out current pointer as an offset
	//
	arc.WriteInteger( script_p - buffer );
	arc.WriteInteger( line );
	arc.WriteRaw( token, sizeof( token ) );
}

void Script::Unarchive(Archiver &arc)
{
	int i;
	
	Class::Unarchive( arc );
	
	arc.ReadString( &filename );
	//
	// load the file in
	//
	LoadFile( filename.c_str() );
	
	tokenready  = arc.ReadBoolean();
	i = arc.ReadInteger();
	//
	// restore the script pointer
	//
	script_p = buffer + i;
	line = arc.ReadInteger();
	arc.ReadRaw( token, sizeof( token ) );
}
