/*
	CD_MAC.C
	
	Mac implementation of Quake CD-ROM access routines

*/

/* Includes */
#include <burger.h>

#include "client.h"

/* Internal globals */

static Redbook_t gRedbook;		/* BurgerLib CD Audio reference struct */

static qboolean cdValid;		/* Valid Audio CD mounted? */
static qboolean playing;		/* CD Track playing */
static qboolean wasPlaying;		/* Was a track playing when pause occured? */
static qboolean playLooping;	/* Are we looping the current track? */
static qboolean enabled;		/* CD-ROM access is enabled */
static byte		playTrack;		/* Currently playing audio track */
static byte		maxTrack;		/* Number of tracks on mounted CD-ROM */

static byte		remap[100];		/* track remap array for track selection */

cvar_t *cd_nocd;				/* cvar for whether or not to access CD */
cvar_t *cd_loopcount;			/* cvar for count of times to loop a looped track */
cvar_t *cd_looptrack;			/* track to switch to when loop track finishes */

/* Prototypes */

void CDAudio_Pause(void);

/* ---------------------------------------------------------------
	CDAudio_Eject
	
	Opens the drive drawer, ejecting the CD-ROM
*/
static void CDAudio_Eject(void)
{
	if (gRedbook.Active == FALSE)	/* If redbook isn't active, just bail */
		return;

	RedbookOpenDrawer(&gRedbook);	/* Open the tray */
}


/* ---------------------------------------------------------------
	CDAudio_CloseDoor
	
	Closes the drive drawer
*/
static void CDAudio_CloseDoor(void)
{
	if (gRedbook.Active == FALSE)	/* If redbook isn't active, just bail */
		return;
		
	RedbookCloseDrawer(&gRedbook);	/* Close the tray */
}

/* ---------------------------------------------------------------
	CDAudio_GetAudioDiskInfo
	
	Determines if there is a valid CD-ROM with audio tracks 
	in the drive, returns 0 if there is
*/
static int CDAudio_GetAudioDiskInfo(void)
{
int trax;

	cdValid = FALSE;	// assume failure
	
	if (gRedbook.Active == FALSE)	/* If redbook isn't active, just bail */
		return(-1);

	trax = RedbookGetTrackCount(&gRedbook);	/* get track count */
	if (trax == 0)							/* If no tracks, it's not a valid CD */
		return(-1);
	
	maxTrack = trax;						/* Set max track var */
	cdValid = TRUE;							/* Set current cd is valid */
	return(0);								/* Return success */
}

/* ---------------------------------------------------------------
	CDAudio_Play
	
	Plays the track specified by 'track', 'looping' is TRUE if
	the track is to be looped
*/
void CDAudio_Play(int track, qboolean looping)
{
int mode;

	if (gRedbook.Active == FALSE)	/* If redbook isn't active, just bail */
		return;

	if (!enabled)					/* If we're not enabled, just bail */
		return;

	if (!cdValid)					/* If current CD isn't valid, try and validate it */
	{
		CDAudio_GetAudioDiskInfo();
		if (!cdValid)				/* Did it work? */
			return;					/* If not, just bail */
	}
	
	playLooping = looping;			/* Set the global looping flag */
	track = remap[track];			/* remap the track using the remap table */
	
	if (track < 1 || track > maxTrack)	/* If the track isn't in a valid range */
	{
		CDAudio_Stop();					/* Stop any currently playing track */
		return;
	}
	
	if (playing)					/* If there is a track playing */
	{
		if (playTrack == track)		/* And it isn't the same track as what we want to play */
			return;	
		CDAudio_Stop();				/* Then stop the currently playing track */
	}
	
	if (looping == TRUE)			/* Set the loop or playonce play mode */
		mode = REDBOOKPLAYLOOP;
	else
		mode = REDBOOKPLAYONCE;
	
	RedbookPlay(&gRedbook, mode, track);	/* Call our redbook track play */
	
	playTrack = track;				/* Set the currently playing track */
	playing = TRUE;					/* We are now playing! */
	
	if (Cvar_VariableValue("cd_nocd"))	/* But if nocd cvar is set */
		CDAudio_Pause();				/* immediately pause */
}

/* ---------------------------------------------------------------
	CDAudio_Stop
	
	Stop the currently playing track, if there is one
*/
void CDAudio_Stop(void)
{
	if (gRedbook.Active == FALSE)	/* If redbook isn't active, just bail */
		return;
	if (!enabled)					/* if we're not enabled, just bail */
		return;
	if (!playing)					/* if we're not playing, just bail */
		return;
		
	RedbookStop(&gRedbook);			/* Stop the currently playing track */
	
	playing = FALSE;				/* We're not playing no more */
	wasPlaying = FALSE;				/* we weren't paused when we stopped playing */
}

/* ---------------------------------------------------------------
	CDAudio_Resume
	
	Resume from pause, if there was a playing track
*/
static void CDAudio_Resume(void)
{
	if (gRedbook.Active == FALSE)	/* If redbook isn't active, just bail */
		return;
	if (!enabled)					/* if we're not enabled, just bail */
		return;
	if (!cdValid)					/* if no valid CD mounted, just bail */
		return;
	if (!wasPlaying)				/* if we weren't even paused, just bail */
		return;
		
	RedbookResume(&gRedbook);		/* resume the paused track */
	playing = TRUE;					/* we're playing again */
}

/* ---------------------------------------------------------------
	CDAudio_Pause
	
	Pause the currently playing track, if there is one
*/
void CDAudio_Pause(void)
{
	if (gRedbook.Active == FALSE)	/* If redbook isn't active, just bail */
		return;
	if (!enabled)					/* if we're not enabled, just bail */
		return;
	if (!playing)					/* If we aren't even playing a track, just bail */
		return;
	
	RedbookPause(&gRedbook);		/* pause the currently playing track */
	
	wasPlaying = playing;			/* yes, we were playing before the pause */
	playing = FALSE;				/* not playing no more */
}

/* ---------------------------------------------------------------
	CDAudio_Update
	
	Updates the status of the CD-ROM drive depending on the
	cd_nocd cvar
*/
void CDAudio_Update(void)
{
	if (gRedbook.Active == FALSE)	/* If redbook isn't active, just bail */
		return;

	if (cd_nocd->value != !enabled)	/* does the cvar match our enabled state */
	{
		if (cd_nocd->value)			/* If not, update the play state based on no_cd */
		{
			CDAudio_Stop();			/* turn in off! */
			enabled = FALSE;
		}
		else
		{
			enabled = TRUE;			/* turn it on! */
			CDAudio_Resume();
		}
	}
}

/* ---------------------------------------------------------------
	CD_f
	
	function that handles the 'cd' console command
	valid commands:
					info
					eject
					loop (track)
					pause
					resume
					stop
					play (track)
					on
					off
					reset
					remap (tracks)
					open
					close
*/
static void CD_f(void)
{
const char *command;
int ret;
int n;

	if (Cmd_Argc() < 2)		/* must have at least 'cd' and the 'command' verb */
		return;
		
	command = Cmd_Argv(1);	/* get the 'command' word */
	
	if (Burger::StringCaseCompare(command, "on") == 0)		/* turn CD-ROM on? */
	{
		enabled = TRUE;							/* turn it on then */
		return;
	}
	
	if (Burger::StringCaseCompare(command, "off") == 0)		/* turn CD-ROM off? */
	{
		if (playing)
			CDAudio_Stop();						/* stop any playing track and turn it off */
		enabled = FALSE;
		return;
	}
	
	if (Burger::StringCaseCompare(command, "reset") == 0)	/* reset the CD-ROM interface? */
	{
		enabled = TRUE;							/* turn us on if we were off */
		if (playing)							/* Stop any playing track */
			CDAudio_Stop();
		for (n = 0; n < 100; n++)				/* reset the track remap table */
			remap[n] = n;						/* to a direct correspondence */
		CDAudio_GetAudioDiskInfo();				/* revalidate current CD */
		return;		
	}
	
	if (Burger::StringCaseCompare(command, "remap") == 0)	/* remap track list? */
	{
		ret = Cmd_Argc() - 2;					/* Get the number of remap args */
		if (ret <= 0)							/* if there aren't any, just list */
		{
			for (n = 1; n < 100; n++)			/* the remap table entries */
				if (remap[n] != n)
					Com_Printf("  %u -> %u\n", n, remap[n]);
			return;
		}
		for (n=1; n<=ret; n++)					/* otherwise, set the remap list */
			remap[n] = atoi(Cmd_Argv(n+1));		/* with the command arguments */
		return;
	}
	
	if (Burger::StringCaseCompare(command, "close") == 0)	/* close the CD-ROM drawer? */
	{
		CDAudio_CloseDoor();
		return;
	}
	
	if (!cdValid)								/* check if current CD is valid before parsing */
	{											/* the next commands ... */
		CDAudio_GetAudioDiskInfo();				/* if not valid, try to validate it */
		if (!cdValid)							/* if we can't, print a message and bail */
		{
			Com_Printf("No CD in player\n");
			return;
		}
	}
	
	if (Burger::StringCaseCompare(command, "play") == 0)		/* play a track (unlooped)? */
	{
		CDAudio_Play(atoi(Cmd_Argv(2)), FALSE);
		return;
	}

	if (Burger::StringCaseCompare(command, "loop") == 0)		/* play a track (looped)? */
	{
		CDAudio_Play(atoi(Cmd_Argv(2)), TRUE);
		return;
	}
	
	if (Burger::StringCaseCompare(command, "stop") == 0)		/* stop a currently playing track */
	{
		CDAudio_Stop();
		return;
	}

	if (Burger::StringCaseCompare(command, "pause") == 0)	/* pause a currently playing track */
	{
		CDAudio_Pause();
		return;
	}

	if (Burger::StringCaseCompare(command, "resume") == 0)	/* resume a paused track */
	{
		CDAudio_Resume();
		return;
	}

	if (Burger::StringCaseCompare(command, "eject") == 0)	/* eject the current CD */
	{
		if (playing)
			CDAudio_Stop();
		CDAudio_Eject();
		cdValid = FALSE;
		return;
	}

	if (Burger::StringCaseCompare(command, "info") == 0)		/* dump the number of tracks and play status */
	{
		Com_Printf("%u tracks\n",maxTrack);
		if (playing)
			Com_Printf("Currently %s track %u\n", playLooping ? "looping" : "playing", playTrack);
		else if (wasPlaying)
			Com_Printf("Paused %s track %u\n", playLooping ? "looping" : "playing", playTrack);
		return;
	}

}

/* ---------------------------------------------------------------
	CDAudio_Init
	
	Inits the CD-ROM system for playing back audio
*/
int CDAudio_Init(void)
{
int n;
int res;
	
	enabled = FALSE;	// presume failure

	cd_nocd = Cvar_Get("cd_nocd", "0", CVAR_ARCHIVE);	/* access the cvars */
	cd_loopcount = Cvar_Get("cd_loopcount", "4", 0);
	cd_looptrack = Cvar_Get("cd_looptrack", "11", 0);
	
	if (cd_nocd->value)									/* if no_cd set, bail with error */
		return(-1);
	
	res = RedbookInit(&gRedbook);						/* Init BurgerLib redbook audio */
	if (res == FALSE)
	{
		enabled = TRUE;									/* if success, we are enabled */
		if (CDAudio_GetAudioDiskInfo())					/* try to validate current CD */
		{
			cdValid = FALSE;							/* if it fails, disable us */
			enabled = FALSE;
		}

		for (n = 0; n < 100; n++)						/* init the remap table to a direct */
			remap[n] = n;								/* track correspondence */
	
		Cmd_AddCommand("cdplayer", CD_f);						/* add the 'CD' console command */
		Com_Printf("CD Audio Initialized\n");
	}
	else
	{
		Com_Printf("CDAudio_Init failed\n");
	}
		
	return(res);										/* return our status */
}

/* ---------------------------------------------------------------
	CDAudio_Shutdown
	
	Shuts down the CD-ROM system for playing back audio
*/
void CDAudio_Shutdown(void)
{
	if (gRedbook.Active == TRUE)	/* If redbook isn't active, just bail */
	{
		CDAudio_Stop();				/* stop any playing track */
		RedbookDestroy(&gRedbook);	/* shut down burgerlib Redbook audio */
	}
	enabled = FALSE;				/* not enabled any more */
}

