#include "Burger.h"
#include "qcommon.h"
#include "rw_mac.h"
#include "errno.h"
#include "mac_isp.h"
#include "client.h"

void VID_Activate(qboolean isActive);

void InitBackground(void)
{
	IN_Activate(FALSE);
	VID_Activate(FALSE);
	Input_Suspend();	/* inputsprocket */
//	S_Activate(FALSE);
}


void KillBackground(void)
{
	IN_Activate(TRUE);
	VID_Activate(TRUE);
	Input_Resume();	/* inputsprocket */
}


void Mac_Resume_f(void)
{
	if (gMacBackground == FALSE)
		return;
		
	CL_Pause_f();			// unpause if necessary
	gMacBackground = FALSE;
	
	Com_Printf("Resuming from Mac desktop\n");
}

void Mac_Desktop_f(void)
{
cvar_t	*isPaused;

	// Already in background?
	if (gMacBackground == TRUE)
		return;

/*
	First, check if game is paused.  If it isn't, try to pause
	the game.  Since we can't pause in multiplayer, and the
	background issues would be nasty on the Mac, sorry, but no
	switching to the desktop when you're connected to a server
*/

	isPaused = Cvar_Get ("paused", "0", 0);
	
	if (isPaused == 0)
	{
		// never pause in multiplayer
		if (Cvar_VariableValue ("maxclients") > 1 || !Com_ServerState ())
		{
			Com_Printf("Can't switch to desktop in multiplayer");
			Cvar_SetValue ("paused", !isPaused->value);
			return;
		}

		Cvar_SetValue ("paused", !isPaused->value);
	}
	Com_Printf("Switching to Mac desktop\n");
	
	gMacBackground = TRUE;
}

// NCT_HACK function to convert path formats so the mac is happy

int	fprintfx(FILE * file, const char * format, ...)
{
	va_list		argptr;
	char		msg[2048];
	int			startoffset,endoffset,leng;
	
	va_start (argptr,format);
	vsprintf (msg,format,argptr);
	va_end (argptr);

	leng = Burger::StringLength(msg);
	startoffset = 0;
	while (msg[startoffset] != 0)
	{
		if (msg[startoffset] == 0xA)	//CR (converting Mac to 0x0D0A
		{
			for (endoffset = leng; endoffset > startoffset; endoffset--)
			{
				msg[endoffset+1] = msg[endoffset];
			}
			msg[startoffset] = 0xD;
			++startoffset;
			msg[startoffset] = 0xA;
		}
		++startoffset;
	}
	return(fwrite(msg,1,Burger::StringLength(msg),file));
}

/**********************************

	Memory pool of randomly sized memory chunks
	
**********************************/

typedef struct MemGroup_t {	/* Array of arbitary memory chunks */
	void ***MemList;		/* Handle to list of memory */
	Word MemListSize;		/* Size of the handle list */
	Word Count;				/* Number of memory pointers */
	Word32 Total;			/* Total bytes allocated */
} MemGroup_t;

/**********************************

	Init the memory group.
	
**********************************/

static void MemGroupInit(MemGroup_t *Input)
{
	Input->MemList = 0;
	Input->MemListSize = 0;
	Input->Count = 0;
	Input->Total = 0;
}

/**********************************

	Dispose of a memory group.
	
**********************************/

static void MemGroupDestroy(MemGroup_t *Input)
{
	Word i;
	i = Input->Count;
	if (i) {
		void **WorkPtr;
		WorkPtr = (void **)LockAHandle((void **)Input->MemList);
		do {
			Free(WorkPtr[0]);
			++WorkPtr;
		} while (--i);
		UnlockAHandle((void **)Input->MemList);
	}
	DeallocAHandle((void **)Input->MemList);
	Input->MemList = 0;
	Input->Count = 0;
	Input->Total = 0;
	Input->MemListSize = 0;
}

/**********************************

	Return the total memory allocated
	
**********************************/

static Word32 MemGroupTotalAlloc(const MemGroup_t *Input)
{
	return Input->Total;
}

/**********************************

	Allocate memory for the group
	
**********************************/

static void *MemGroupAlloc(MemGroup_t *Input,Word32 Size)
{
	Word i;
	void *MemPtr;
	i = Input->Count;
	if (i>=Input->MemListSize) {
		Input->MemList = (void ***)ResizeAHandle((void **)Input->MemList,(i+64)*sizeof(void *));
		if (!Input->MemList) {
			Input->MemListSize = 0;
			Input->Count = 0;
			Input->Total = 0;
			return 0;
		}
		Input->MemListSize = i+64;
	}
	MemPtr = AllocClear(Size);
	if (MemPtr) {
		Input->MemList[0][i] = MemPtr;
		++i;
		Input->Count = i;
		Input->Total += Size;
	}
	return MemPtr;
}

/**********************************

	Sin Memory group manager.
	
**********************************/

typedef struct MemGroupList_t {
	LinkedListEntry_t Entry;		/* Linked list entry */
	void *IDPtr;					/* Memory key */
	MemGroup_t Group;				/* Memory attached */
} MemGroupList_t;

static MemGroup_t CurrentGroup;		/* Current hunk being processed */
static LinkedList_t GroupList;		/* Linked list of active hunks */
static void *FirstEntry;			/* First memory entry */

/**********************************

	Create a new global memory group to
	allocate a series of chunks of memory.
	
**********************************/

void Hunk_Begin(int maxsize)
{
	MemGroupInit(&CurrentGroup);		/* Init the current group */
	FirstEntry = 0;						/* Mark the first allocation */
}

/**********************************

	After calling Hunk_Begin, I will log every memory
	allocation. The first allocation is important
	since it is the root pointer for Sin
	models. This must match the pointer returned
	from Hunk_End() or bad things happen
	
**********************************/

void *Hunk_Alloc (int size)
{
	void *Result;
	
	Result = MemGroupAlloc(&CurrentGroup,size);
	if (!Result) {
		Sys_Error ("Not enough free memory.  If you experience this during extended gameplay, consider increasing Sin's preferred RAM size.  You can do this in the Finder by selecting Get Info on the Sin application icon.");
		DebugString("Out of memory!\n");
	}
	if (!FirstEntry) {
		FirstEntry = Result;		/* Mark the first allocation */
	}
	return Result;
}

/**********************************

	Close up the allocations and return the root pointer
	as the ID. I then put away the real chain into a linked
	list and use the list to do the real disposal
	
**********************************/

int Hunk_End(void **DestPtr)
{
	MemGroupList_t *NewEntry;
	NewEntry = (MemGroupList_t *)Alloc(sizeof(MemGroupList_t));
	if (NewEntry) {
		Word32 cursize;
		NewEntry->Entry.KillProc = 0;
		NewEntry->Entry.Data = NewEntry;
		NewEntry->IDPtr = FirstEntry;
		NewEntry->Group = CurrentGroup;
		LinkedListAddEntryEnd(&GroupList,&NewEntry->Entry);
		DestPtr[0] = FirstEntry;
		cursize = MemGroupTotalAlloc(&NewEntry->Group);		/* Get the memory size */
		return cursize;
	}
	MemGroupDestroy(&CurrentGroup);
	DestPtr[0] = 0;
	return 0;
}

/**********************************

	Scan the linked list until the memory is found.
	Then dispose of the data and remove it from the list
	
**********************************/

void Hunk_Delete(void *DestPtr)
{
	LinkedListEntry_t *EntryPtr;
	EntryPtr = LinkedListGetFirst(&GroupList);
	if (EntryPtr) {
		do {
			MemGroupList_t *WorkPtr;
			WorkPtr = (MemGroupList_t *)EntryPtr->Data;
			if (WorkPtr->IDPtr == DestPtr) {
				MemGroupDestroy(&WorkPtr->Group);				/* Destroy the memory */
				LinkedListDeleteEntry(&GroupList,EntryPtr);		/* Remove from the list */
				return;
			}
			EntryPtr = EntryPtr->Next;
		} while (EntryPtr);
	}
	Free(DestPtr);
}