/**********************************

	Contains:	Functions to enable Mac OS device gamma adjustments using Windows common 3 channel 256 element 8 bit gamma ramps

	Written by:	Geoff Stahl

	Copyright:	Copyright  1999 Apple Computer, Inc., All Rights Reserved

**********************************/

#include "macgamma.h"
#include <Devices.h>
#include <Video.h>

/**********************************

	Returns the device gamma table pointer in ppDeviceTable

**********************************/

static OSErr GetGammaTable(GDHandle hGD, GammaTblPtr * ppTableGammaOut)
{
#if !TARGET_API_MAC_CARBON
	VDGammaRecord DeviceGammaRec;
	CntrlParam cParam;
	OSErr err;
	
	FastMemSet(&cParam,0,sizeof(cParam));
	cParam.ioCRefNum = hGD[0]->gdRefNum;
	cParam.csCode = cscGetGamma;								// Get Gamma commnd to device
	*(Ptr *)cParam.csParam = (Ptr) &DeviceGammaRec;				// record for gamma
	err = PBStatusSync( (ParmBlkPtr)&cParam);					// get gamma
	ppTableGammaOut[0] = (GammaTblPtr)(DeviceGammaRec.csGTable);	// pull table out of record
	return err;	
#else
	ppTableGammaOut[0] = 0;
	return FALSE;
#endif
}

/**********************************

	creates an empty gamma table of a given size, assume no formula data will be used

**********************************/

static Word8 *CreateEmptyGammaTable(Word channels,Word entries,Word bits)
{
	GammaTblPtr pTableGammaOut;
	Word tableSize, dataWidth;

	dataWidth = (bits + 7)>>3;										// number of bytes per entry
	tableSize = sizeof (GammaTbl) + (channels * entries * dataWidth);
	pTableGammaOut = (GammaTblPtr) AllocClear(tableSize);			// allocate new tabel

	if (pTableGammaOut)	{											// if we successfully allocated
//		pTableGammaOut->gVersion = 0;								// set parameters based on input
//		pTableGammaOut->gType = 0;
//		pTableGammaOut->gFormulaSize = 0;
		pTableGammaOut->gChanCnt = channels;
		pTableGammaOut->gDataCnt = entries;
		pTableGammaOut->gDataWidth = bits;
	}
	return (Word8 *)pTableGammaOut;										// return whatever we allocated
}

/**********************************

	given a pointer toa device gamma table properly iterates and copies

**********************************/

static Word8 *CopyGammaTable (GammaTblPtr pTableGammaIn)
{
	GammaTblPtr pTableGammaOut = NULL;
	Word tableSize, dataWidth;

	if (pTableGammaIn) {												// if there is a table to copy 
		dataWidth = (pTableGammaIn->gDataWidth + 7) >> 3;			// number of bytes per entry
		tableSize = sizeof (GammaTbl) + pTableGammaIn->gFormulaSize +
					(pTableGammaIn->gChanCnt * pTableGammaIn->gDataCnt * dataWidth);
		pTableGammaOut = (GammaTblPtr) Alloc(tableSize);			// allocate new table
		if (pTableGammaOut) {
			FastMemCpy(pTableGammaOut,pTableGammaIn,tableSize);	// move everything
		}
	}
	return (Word8 *)pTableGammaOut;										// return whatever we allocated, could be NULL
}

/**********************************

	disposes gamma table returned from GetGammaTable, GetDeviceGamma, or CopyGammaTable 

**********************************/

static void DisposeGammaTable(Word8 *pGamma)
{
	Free(pGamma);		/* Kill the pointer */
}

/**********************************

	returns pointer to copy of orginal device gamma table in native format (allocates memory for gamma table, call DisposeDeviceGamma to delete)
	5/20/99: (GGS) change spec to return the allocated pointer vice storing internally

**********************************/

static Word8 *GetDeviceGamma (GDHandle hGD)
{
	GammaTblPtr pTableGammaDevice = NULL;
	GammaTblPtr pTableGammaReturn;	
	OSErr err;
	
	err = GetGammaTable (hGD, &pTableGammaDevice);					// get a pointer to the devices table
	pTableGammaReturn = 0;
	if ((err == noErr) && pTableGammaDevice) {						// if succesful
		pTableGammaReturn = (GammaTblPtr) CopyGammaTable (pTableGammaDevice); // copy to global
	}
	return (Word8 *)pTableGammaReturn;
}

/**********************************

	sets device to saved table
	5/20/99: (GGS) now does not delete table, avoids confusion

**********************************/

static void RestoreDeviceGamma (GDHandle hGD,Word8 *pGammaTable)
{
#if !TARGET_API_MAC_CARBON
	VDSetEntryRecord setEntriesRec;
	VDGammaRecord gameRecRestore;
	CTabHandle hCTabDeviceColors;
	Ptr csPtr;
	OSErr err;
	
	if (pGammaTable) {												// if we have a table to restore								
		gameRecRestore.csGTable = (Ptr)pGammaTable;						// setup restore record
		csPtr = (Ptr) &gameRecRestore;
		err = Control(hGD[0]->gdRefNum, cscSetGamma, (Ptr) &csPtr);	// restore gamma
		
		if (err == noErr) {
			GDevice *GD;
			PixMap *Pix;
			GD = hGD[0];
			Pix = GD->gdPMap[0];
			if (Pix->pixelSize == 8) {	// if successful and on an 8 bit device
				hCTabDeviceColors = Pix->pmTable;			// do SetEntries to force CLUT update
				setEntriesRec.csTable = (ColorSpec *) &hCTabDeviceColors[0]->ctTable;
				setEntriesRec.csStart = 0;
				setEntriesRec.csCount = hCTabDeviceColors[0]->ctSize;
				csPtr = (Ptr) &setEntriesRec;
				Control(GD->gdRefNum, cscSetEntries, (Ptr) &csPtr); // SetEntries in CLUT
			}
		}
	}
#endif
}

/**********************************

	returns a pointer to a set of all current device gammas in native format (returns NULL on failure, which means reseting gamma will not be possible)
	5/20/99: (GGS) added

**********************************/

Word8 *MacGetSystemGammas (void)										
{
	recSystemGamma *pSysGammaOut;									// return pointer to system device gamma info
	int devCount;												// number of devices attached
	Bool fail;
	GDHandle hGDevice;
	
	hGDevice = GetDeviceList ();							// top of device list
	devCount = 0;
	if (hGDevice) {
		do {													// iterate
			++devCount;											// count devices					
			hGDevice = GetNextDevice (hGDevice);				// next device
		} while (hGDevice);
	}
	pSysGammaOut = (recSystemGamma *) AllocClear(sizeof (recSystemGamma)); // allocate for structure
	if (pSysGammaOut) {
		recDeviceGamma **GammaPtr;
		GammaPtr = (recDeviceGamma **) AllocClear(sizeof(recDeviceGamma *) * devCount); // allocate for array of pointers to device records
		fail = FALSE;
		if (GammaPtr) {
			pSysGammaOut->devGamma = GammaPtr;
			pSysGammaOut->numDevices = devCount;						// stuff count
			devCount = 0;												// reset iteration
			hGDevice = GetDeviceList();
			do {
				recDeviceGamma *NewGammaPtr;
				NewGammaPtr = (recDeviceGamma *) Alloc(sizeof(recDeviceGamma));	  // new device record
				if (NewGammaPtr) {				// if we actually allocated memory
					GammaPtr[0] = NewGammaPtr;
					NewGammaPtr->hGD = hGDevice;										  // stuff handle
					NewGammaPtr->pDeviceGamma = (GammaTblPtr)GetDeviceGamma (hGDevice); // copy gamma table
				} else {													// otherwise dump record on exit
					fail = TRUE;
				}
				++GammaPtr;
				if (++devCount>=pSysGammaOut->numDevices) {			// Failsafe
					break;
				}
				hGDevice = GetNextDevice(hGDevice);
			} while (hGDevice);
		}
	}
	if (!fail) {														// if we did not fail
		return (Word8 *) pSysGammaOut;									// return pointer to structure
	}
	MacDisposeSystemGammas ((Word8 *)pSysGammaOut);					// otherwise dump the current structures (dispose does error checking)
	return NULL;											// could not complete
}

/**********************************

	restores all system devices to saved gamma setting
	5/20/99: (GGS) added

**********************************/

void MacRestoreSystemGammas (Word8 *pSystemGammas)
{
	Word i;
	if (pSystemGammas) {
		i = ((recSystemGamma *)pSystemGammas)->numDevices;
		if (i) {
			recDeviceGamma **SrcPtr;
			SrcPtr = ((recSystemGamma *)pSystemGammas)->devGamma;
			do {
				recDeviceGamma *WorkPtr;
				WorkPtr = SrcPtr[0];
				if (WorkPtr) {
					RestoreDeviceGamma(WorkPtr->hGD,(Word8 *)WorkPtr->pDeviceGamma);	// restore gamma
				}
				++SrcPtr;
			} while (--i);
		}
	}
}

/**********************************

	iterates through and deletes stored gamma settings
	5/20/99: (GGS) added

**********************************/

void MacDisposeSystemGammas (Word8 *pSystemGammas)
{
	if (pSystemGammas) {
		Word i;
		i = ((recSystemGamma *)pSystemGammas)->numDevices;
		if (i) {
			recDeviceGamma **SrcPtr;
			SrcPtr = ((recSystemGamma *)pSystemGammas)->devGamma;
			do {
				recDeviceGamma *WorkPtr;
				WorkPtr = SrcPtr[0];
				if (WorkPtr) {					// if pointer is valid
					DisposeGammaTable((Word8 *)WorkPtr->pDeviceGamma); // dump gamma table
					Free(WorkPtr);						 // dump device info
				}
				++SrcPtr;
			} while (--i);
		}
		Free(((recSystemGamma *)pSystemGammas)->devGamma);				// dump device pointer array		
		Free(pSystemGammas);							// dump system structure
	}	
}

/**********************************

	retrieves the gamma ramp from a graphics device (pRamp: 3 arrays of 256 elements each)

**********************************/

Word MacGetDeviceGammaRampGD (GDHandle hGD,Word8 *pRamp)
{
	GammaTblPtr pTableGammaTemp;
	long indexChan, indexEntry;
	OSErr err;
	
	if (pRamp && hGD) {													// ensure pRamp is allocated
		pTableGammaTemp = 0;
		err = GetGammaTable (hGD, &pTableGammaTemp);					// get a pointer to the current gamma
		if ((err == noErr) && pTableGammaTemp) {							// if successful
			// fill ramp
			Word8 * pEntry = (Word8 *)&pTableGammaTemp->gFormulaData + pTableGammaTemp->gFormulaSize; // base of table
			Word bytesPerEntry = (pTableGammaTemp->gDataWidth + 7) / 8; // size, in bytes, of the device table entries
			Word shiftRightValue = pTableGammaTemp->gDataWidth - 8; 	 // number of right shifts device -> ramp
			Word channels = pTableGammaTemp->gChanCnt;	
			Word entries = pTableGammaTemp->gDataCnt;									
			if (channels == 3) {											// RGB format
																		// note, this will create runs of entries if dest. is bigger (not linear interpolate)
				indexChan = 0;
				do {
					indexEntry = 0;
					do {
						pRamp[(indexChan << 8) + indexEntry] = 
						  *(pEntry + (indexChan * entries * bytesPerEntry) + indexEntry * ((entries * bytesPerEntry) >> 8)) >> shiftRightValue;
					} while (++indexEntry<256);
				} while (++indexChan<3);
			} else {													// single channel format
				indexEntry = 0;
				do {	// for all entries set vramp value
					for (indexChan = 0; indexChan < channels; indexChan++) {	// repeat for all channels
						pRamp[(indexChan << 8) + indexEntry] = 
						  *(pEntry + ((indexEntry * entries * bytesPerEntry) >> 8)) >> shiftRightValue;
					}
				} while (++indexEntry<256);
			}
			return TRUE;
		}
	}
	return FALSE;
}

/**********************************

	sets the gamma ramp for a graphics device (pRamp: 3 arrays of 256 elements each (R,G,B))

**********************************/

#define CHANNELS 3

Word MacSetDeviceGammaRampGD (GDHandle hGD,const Word8*pRamp)
{
#if !TARGET_API_MAC_CARBON
	VDSetEntryRecord setEntriesRec;
	VDGammaRecord	gameRecRestore;
	GammaTblPtr		pTableGammaNew;
	CTabHandle      hCTabDeviceColors;
	Ptr				csPtr;
	OSErr			err;
	Word dataBits, entries;						// force three channels in the gamma table
	
	if (hGD) {
		if (pRamp) {															// ensure pRamp is allocated
			GammaTblPtr	pTableGammaCurrent;
			pTableGammaCurrent = 0;
			err= GetGammaTable (hGD, &pTableGammaCurrent);						// get pointer to current table
			if ((err == noErr) && pTableGammaCurrent) {
				dataBits = pTableGammaCurrent->gDataWidth;						// table must have same data width
				entries = pTableGammaCurrent->gDataCnt;							// table must be same size
				pTableGammaNew = (GammaTblPtr) CreateEmptyGammaTable (CHANNELS, entries, dataBits); // our new table
				if (pTableGammaNew)	{											// if successful fill table
					Word8 * pGammaBase = (unsigned char *)&pTableGammaNew->gFormulaData + pTableGammaNew->gFormulaSize; // base of table
					if (entries == 256 && dataBits == 8) {						// simple case: direct mapping
						FastMemCpy (pGammaBase,pRamp, CHANNELS * entries);		// move everything
					} else {													// tough case handle entry, channel and data size disparities
						Word bytesPerEntry = (dataBits + 7) / 8; 				// size, in bytes, of the device table entries
						Word shiftRightValue = 8 - dataBits;					// number of right shifts ramp -> device
						Word indexChan;
						Word indexEntry;
						Word indexByte;
						
						shiftRightValue += ((bytesPerEntry - 1) * 8);  			// multibyte entries and the need to map a byte at a time most sig. to least sig.
						for ( indexChan = 0; indexChan < CHANNELS; indexChan++) {	// for all the channels
							for ( indexEntry = 0; indexEntry < entries; indexEntry++) { // for all the entries
								Word currentShift = shiftRightValue;			// reset current bit shift
								Word32 temp = *((unsigned char *)pRamp + (indexChan << 8) + (indexEntry << 8) / entries); // get data from ramp
								for ( indexByte = 0; indexByte < bytesPerEntry; indexByte++) { // for all bytes
									if (currentShift < 0) {						// shift data correctly for current byte
										*pGammaBase = temp << -currentShift;
									} else {
										*pGammaBase = temp >> currentShift;
									}
									++pGammaBase;
									currentShift -= 8;							// increment shift to align to next less sig. byte
								}
							}
						}
					}
					
					// set gamma
					gameRecRestore.csGTable = (Ptr) pTableGammaNew;				// setup restore record
					csPtr = (Ptr) &gameRecRestore;
					err = Control((**hGD).gdRefNum, cscSetGamma, (Ptr) &csPtr);	// restore gamma
					
					if ((hGD[0]->gdPMap[0]->pixelSize == 8) && (err == noErr))	// if successful and on an 8 bit device
					{
						hCTabDeviceColors = (**(**hGD).gdPMap).pmTable;			// do SetEntries to force CLUT update
						setEntriesRec.csTable = (ColorSpec *) &(**hCTabDeviceColors).ctTable;
						setEntriesRec.csStart = 0;
						setEntriesRec.csCount = (**hCTabDeviceColors).ctSize;
						csPtr = (Ptr) &setEntriesRec;
						err = Control((**hGD).gdRefNum, cscSetEntries, (Ptr) &csPtr);	// SetEntries in CLUT
					}
					DisposeGammaTable ((Word8 *) pTableGammaNew);					// dump table
					if (err == noErr)
						return TRUE;
				}
			}
		} else {																// set NULL gamma -> results in linear map
			gameRecRestore.csGTable = (Ptr) NULL;								// setup restore record
			csPtr = (Ptr) &gameRecRestore;
			err = Control(hGD[0]->gdRefNum, cscSetGamma, (Ptr) &csPtr);			// restore gamma
			
			if ((hGD[0]->gdPMap[0]->pixelSize == 8) && (err == noErr)) {		// if successful and on an 8 bit device
				hCTabDeviceColors = hGD[0]->gdPMap[0]->pmTable;					// do SetEntries to force CLUT update
				setEntriesRec.csTable = (ColorSpec *) &hCTabDeviceColors[0]->ctTable;
				setEntriesRec.csStart = 0;
				setEntriesRec.csCount = hCTabDeviceColors[0]->ctSize;
				csPtr = (Ptr) &setEntriesRec;
				err = Control(hGD[0]->gdRefNum, cscSetEntries, (Ptr) &csPtr);	// SetEntries in CLUT
			}
			if (err == noErr) {
				return TRUE;
			}
		}
	}
	return FALSE;															// memory allocation or device control failed if we get here
#else
	return TRUE;
#endif
}

