/*
	vid_mac.c
	
	This file is the front end for handling video display on Quake II Mac
	Responsible for setting up and initializing/shutting down the
	proper rendering library
	
	Also includes stub setup stuff for input routines that I moved out of
	the renderer lib into the main executable

*/

#include <burger.h>
#include <errno.h>
#include <assert.h>

#include "client.h"
#include "rw_mac.h"
#include "mac_libutil.h"
#include "mac_isp.h"
#include "macfuncs.h"

/*****************************************************************************/
/* Extern protos/vars                                                        */
/*****************************************************************************/

/*****************************************************************************/
/* Internal function prototypes                                              */
/*****************************************************************************/

/** KEYBOARD ************************************************************/
refexport_t GetRefAPI (refimport_t rimp);
void VID_Activate(qboolean isActive);

void Do_Key_Event(int key, qboolean down);

void (*KBD_Update_fp)(void);
void (*KBD_Init_fp)(Key_Event_fp_t fp);
void (*KBD_Close_fp)(void);

/** MOUSE ***************************************************************/

in_state_t	in_state;

void (*RW_IN_Init_fp)(in_state_t *in_state_p);
void (*RW_IN_Shutdown_fp)(void);
void (*RW_IN_Activate_fp)(qboolean active);
void (*RW_IN_Commands_fp)(void);
void (*RW_IN_Move_fp)(usercmd_t *cmd);
void (*RW_IN_Frame_fp)(void);

static void Real_IN_Init (void);

/*****************************************************************************/
/* Misc. globals for file                                                    */
/*****************************************************************************/

// structure containing functions exported from refresh DLL
refexport_t	re;

// Console variables that we need to access from this module
cvar_t	*vid_xpos;		// x coordinate of window position
cvar_t	*vid_ypos;		// y coordinate of window position
cvar_t	*vid_fullscreen;

// Global variables used internally by this module

viddef_t	viddef;				// global video state
void 		*reflib_library;	// handle to refresh DLL
qboolean	reflib_active = 0;
char		reflib_name[64];


// NCT_ADD trying to toggle Voodoo on/off before we throw up the InputSprocket
// Configure dialog!

void VID_Restart_f (void);
void VID_FreeReflib(void);

void ToggleVoodoo(int mybool)
{
	int isSoft;

	// Call the renderer to tell it who is boss, but only in GL!

	isSoft = (strcmp("soft", vid_ref->CVarString)) ? 0 : 1;

	if (!isSoft) {
		if (mybool == 0) {
			re.Shutdown();
			VID_FreeReflib ();
		} else {
			VID_Restart_f();
		}
	}
}

int VID_IsFullscreen(void)
{
	return(vid_fullscreen->value);
}

/*
==========================================================================

DIRECT LINK GLUE

==========================================================================
*/

#define	MAXPRINTMSG	4096
static void VID_Printf (int print_level,const char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];

	va_start (argptr,fmt);
	vsprintf (msg,fmt,argptr);
	va_end (argptr);

	if (print_level == PRINT_ALL) {
		Com_Printf ("%s", msg);
	} else {
		Com_DPrintf ("%s", msg);
	}
}

static void VID_Error (int err_level, const char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];

	va_start (argptr,fmt);
	vsprintf (msg,fmt,argptr);
	va_end (argptr);

	Com_Error (err_level, "%s", msg);
}


/*
============
VID_Activate

Command I added to send a suspend/resume event through to the renderer
============
*/
void VID_Activate(qboolean isActive)
{
	re.AppActivate(isActive);
}

/*
============
VID_Restart_f

Console command to re-start the video mode and refresh DLL. We do this
simply by setting the modified flag for the vid_ref variable, which will
cause the entire video mode and refresh DLL to be reset on the next frame.
============
*/
void VID_Restart_f (void)
{
	vid_ref->modified = TRUE;
}


/*
** VID_GetModeInfo
*/
typedef struct vidmode_s
{
    const char *description;
    int         width, height;
    int         mode;
} vidmode_t;

vidmode_t vid_modes[] =
{
	{ "Mode 0: 640x480",   640, 480,   0 },
	{ "Mode 1: 800x600",   800, 600,   1 },
	{ "Mode 2: 960x720",   960, 720,   2 },
	{ "Mode 3: 1024x768",  1024, 768,  3 },
	{ "Mode 4: 1152x864",  1152, 864,  4 },
	{ "Mode 5: 1280x960",  1280, 960, 5 },
	{ "Mode 6: 1600x1200", 1600, 1200, 6 }
};
#define VID_NUM_MODES ( sizeof( vid_modes ) / sizeof( vid_modes[0] ) )

static qboolean VID_GetModeInfo( int *width, int *height, int mode )
{
    if ( mode < 0 || mode >= VID_NUM_MODES ) {
        return FALSE;
	}
    *width  = vid_modes[mode].width;
    *height = vid_modes[mode].height;

    return TRUE;
}

static void VID_NewWindow (int width, int height)
{
	viddef.width = width;
	viddef.height = height;

	CONSOLE_CheckResize();
	if (sound_started) {
		CL_Snd_Restart_f();
	}
}

void VID_FreeReflib (void)
{
	if (reflib_library) {
		if (KBD_Close_fp) {
			KBD_Close_fp();
		}
		if (RW_IN_Shutdown_fp) {
			RW_IN_Shutdown_fp();
		}
#ifndef REF_HARD_LINKED
		FreeMacLibrary(reflib_library);	// release the renderer shared library
#endif
	}

	KBD_Init_fp = NULL;
	KBD_Update_fp = NULL;
	KBD_Close_fp = NULL;
	RW_IN_Init_fp = NULL;
	RW_IN_Shutdown_fp = NULL;
	RW_IN_Activate_fp = NULL;
	RW_IN_Commands_fp = NULL;
	RW_IN_Move_fp = NULL;
	RW_IN_Frame_fp = NULL;

	FastMemSet(&re, 0, sizeof(re));
	reflib_library = NULL;
	reflib_active  = FALSE;
}


/*
==============
VID_LoadRefresh
==============
*/
static qboolean VID_LoadRefresh( char *name )
{
	refimport_t	ri;
	extern refexport_t GL_GetRefAPI (refimport_t rimp );
	extern refexport_t GetRefAPI (refimport_t rimp );
	extern debugline_t *DebugLines;
	extern int numDebugLines;
	if ( reflib_active ) {
		//	We have the library loaded... if it's the same, don't actually do all this
		//	junk.  Instead, just change modes.  Neat, eh?
		if (!strcmp(name, reflib_name)) {
			re.Shutdown();
				
			if ( re.Init( 0, 0 ) == -1 ) {
				re.Shutdown();
				VID_FreeReflib ();
				return FALSE;
			}
			return TRUE;
		}
	
		if (KBD_Close_fp) {
			KBD_Close_fp();
		}
		if (RW_IN_Shutdown_fp) {
			RW_IN_Shutdown_fp();
		}
		KBD_Close_fp = NULL;
		RW_IN_Shutdown_fp = NULL;
		re.Shutdown();
		VID_FreeReflib ();
	}

#if 0
	Com_Printf( "------- Loading %s -------\n", name );
	
	reflib_library = LoadMacLibrary(name);
	if (reflib_library) {
		Com_DPrintf("LoadMacLibrary (%s)\n",name);
	} else {
		return(FALSE);	// no lib!  Eeek!
	}
#endif

	ri.Cmd_AddCommand = Cmd_AddCommand;
	ri.Cmd_RemoveCommand = Cmd_RemoveCommand;
	ri.Cmd_Argc = Cmd_Argc;
	ri.Cmd_Argv = Cmd_Argv;
	ri.Cmd_ExecuteText = Cbuf_ExecuteText;
	ri.Con_Printf = VID_Printf;
	ri.Sys_Error = VID_Error;
	ri.FS_LoadFile = FS_LoadFile;
	ri.FS_FreeFile = FS_FreeFile;
	ri.FS_Gamedir = FS_Gamedir;
	ri.Cvar_Get = Cvar_Get;
	ri.Cvar_Set = Cvar_Set;
	ri.Cvar_SetValue = Cvar_SetValue;
	ri.Vid_GetModeInfo = VID_GetModeInfo;
	ri.Vid_MenuInit = VID_MenuInit;
	ri.Vid_NewWindow = VID_NewWindow;

	ri.SCR_CenterPrintf = SCR_CenterPrintf;
	ri.FS_Playerdir = FS_Playerdir;
	ri.SINMDL_GetAnim = SINMDL_GetAnim;
	ri.SINMDL_GetBaseModel = SINMDL_GetBaseModel;
	ri.SINMDL_GetDef = SINMDL_GetDef;
	ri.SINMDL_FreeDef = SINMDL_FreeDef;
	ri.SINMDL_RegisterDef = SINMDL_RegisterDef;
	ri.CONSOLE_Process = CONSOLE_Process;

	ri.DebugLines		= &DebugLines;
	ri.numDebugLines	= &numDebugLines;

	if (strstr(name,"gl")) {
		re = GL_GetRefAPI( ri );
	} else {
		re = GetRefAPI( ri );
	}

	if (re.api_version != API_VERSION) {
		VID_FreeReflib ();
		Com_Error (ERR_FATAL, "%s has incompatible api_version", name);
	}

	/* Init silly input function stubs, this stuff was originally in the
	   renderer lib but I moved it into the app itself */

	/* Init IN (Mouse) */
	in_state.IN_CenterView_fp = IN_CenterView;
	in_state.Key_Event_fp = Do_Key_Event;
	in_state.viewangles = cl.viewangles;
	in_state.in_strafe_state = &in_strafe.state;
    RW_IN_Init_fp = RW_IN_Init;
    RW_IN_Shutdown_fp = RW_IN_Shutdown;
    RW_IN_Activate_fp = RW_IN_Activate;
    RW_IN_Commands_fp = RW_IN_Commands;
    RW_IN_Move_fp = RW_IN_Move;
    RW_IN_Frame_fp = RW_IN_Frame;

// NCT_COMMENT, 1.0.3  This seemed to cause problems on Heretic II, being paranoid
//	Input_Suspend();
	if ( re.Init( 0, 0 ) == -1 ) {
		re.Shutdown();
		VID_FreeReflib ();
		return FALSE;
	}

	KBD_Init_fp = KBD_Init;
	KBD_Update_fp = KBD_Update;
	KBD_Close_fp = KBD_Close;
//#endif
	KBD_Init_fp(Do_Key_Event);
	Real_IN_Init();

	Com_Printf( "------------------------------------\n");
	reflib_active = TRUE;
	Burger::StringCopy(reflib_name, name);
	return TRUE;
}

/*
============
VID_CheckChanges

This function gets called once just before drawing each frame, and it's sole purpose in life
is to check to see if any of the video mode parameters have changed, and if they have to 
update the rendering DLL and/or video mode to match.
============
*/
void VID_CheckChanges (void)
{
	char name[100];

	if ( vid_ref->modified ) {
		S_StopAllSounds();
#ifdef SIN
		CL_ClearTriangles();
		CL_ClearTempmodels();
#endif
	}

	while (vid_ref->modified) {
		/*
		** refresh has changed
		*/
		vid_ref->modified = FALSE;
		vid_fullscreen->modified = TRUE;
		cl.refresh_prepped = FALSE;
#ifndef SIN
		cls.disable_screen = TRUE;
#else
//		S_StopAllSounds();	//	redundant?
#endif

		Com_sprintf( name, sizeof(name), "ref_%s.lib", vid_ref->CVarString );
		if ( !VID_LoadRefresh( name ) ) {
		    if ( strcmp (vid_ref->CVarString, "soft") == 0 ) {
			/* Com_Printf("Refresh failed\n");
				sw_mode = Cvar_Get( "sw_mode", "0", 0 );
				if (sw_mode->value != 0) {
				        Com_Printf("Trying mode 0\n");
					Cvar_SetValue("sw_mode", 0);
					if ( !VID_LoadRefresh( name ) )
						Com_Error (ERR_FATAL, "Couldn't fall back to software refresh!");
				} else*/
					Com_Error (ERR_FATAL, "Couldn't fall back to software refresh!");
			}

			Cvar_Set( "vid_ref", "soft" );

			/*
			** drop the console if we fail to load a refresh
			*/
			if ( cls.key_dest != key_console ) {
				Con_ToggleConsole_f();
			}
		}
#ifndef SIN
		cls.disable_screen = FALSE;
#endif
	}

}

/*
============
VID_Init
============
*/
void VID_Init (void)
{
	/* Create the video variables so we know how to start the graphics drivers */
    vid_ref = Cvar_Get ("vid_ref", "soft", CVAR_ARCHIVE);
	vid_xpos = Cvar_Get ("vid_xpos", "3", CVAR_ARCHIVE);
	vid_ypos = Cvar_Get ("vid_ypos", "22", CVAR_ARCHIVE);
	vid_fullscreen = Cvar_Get ("vid_fullscreen", "1", CVAR_ARCHIVE);
	vid_gamma = Cvar_Get( "vid_gamma", "1", CVAR_ARCHIVE );

#ifdef SIN
	scr_shotname = Cvar_Get( "scr_shotname", SCREEN_SHOT_NAME, CVAR_ARCHIVE );
#endif

	/* Add some console commands that we want to handle */
	Cmd_AddCommand ("vid_restart", VID_Restart_f);

	/* Disable the 3Dfx splash screen */
	//putenv("FX_GLIDE_NO_SPLASH=0");
		
	/* Start the graphics mode and load refresh DLL */
	VID_CheckChanges();
	
	SCR_EndLoadingPlaque();
}

/*
============
VID_Shutdown
============
*/
void VID_Shutdown (void)
{
	if ( reflib_active )
	{
		if (KBD_Close_fp)
			KBD_Close_fp();
		if (RW_IN_Shutdown_fp)
			RW_IN_Shutdown_fp();
		KBD_Close_fp = NULL;
		RW_IN_Shutdown_fp = NULL;
		re.Shutdown ();
		VID_FreeReflib ();
	}
}


/*****************************************************************************/
/* INPUT                                                                     */
/*****************************************************************************/

/*
	NCT, these are just stubs, I decided to move the input for the Mac version
	out of the renderer libraries into the main executable
*/

int	gIn_Active = TRUE;

// This if fake, it's acutally done by the Refresh load
void IN_Init (void)
{
}

static void Real_IN_Init (void)
{
	if (RW_IN_Init_fp)
		RW_IN_Init_fp(&in_state);
}

void IN_Shutdown (void)
{
	if (RW_IN_Shutdown_fp)
		RW_IN_Shutdown_fp();
	RW_IN_Shutdown_fp = NULL;
}

void IN_Commands (void)
{
	if (RW_IN_Commands_fp)
		RW_IN_Commands_fp();
}

void IN_Move (usercmd_t *cmd)
{
	if (RW_IN_Move_fp)
		RW_IN_Move_fp(cmd);
}

void IN_Frame (void)
{
	if (gIn_Active == FALSE)
		return;
	
	Input_GetButtonEvents(Sys_Milliseconds());	/* Read InputSprocket buttons */
		
	if (RW_IN_Frame_fp)
		RW_IN_Frame_fp();
}

void IN_Activate (qboolean active)
{
	gIn_Active = active;
	
	// InputSprocket disable/enable hook needs to go here
	if (RW_IN_Activate_fp)
		RW_IN_Activate_fp(active);
}

void Do_Key_Event(int key, qboolean down)
{
	Key_Event(key, down, Sys_Milliseconds());
}

void Do_Key_Event_Down(int key)
{
	Do_Key_Event(key,TRUE);
}
void Do_Key_Event_Up(int key)
{
	Do_Key_Event(key,FALSE);
}

