//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/common.c                                               $
// $Revision:: 5                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 8/10/01 6:01 PM                                                $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// misc functions used in client and server
//

#include "qcommon.h"
#include "crc.h"
#include "screen.h"
#include "keys.h"
#include <setjmp.h>
#
#define	MAXPRINTMSG	4096

#define MAX_NUM_ARGVS	50

int		com_argc;
char	*com_argv[MAX_NUM_ARGVS+1];

int		realtime;

jmp_buf abortframe;		// an ERR_DROP occured, exit the entire frame


FILE	*log_stats_file;

cvar_t	*host_speeds;
cvar_t	*log_stats;
cvar_t	*timescale;
cvar_t	*fixedtime;
cvar_t	*logfile_active;	// 1 = buffer log, 2 = flush after each print
cvar_t	*showtrace;

#ifndef REF_HARD_LINKED
cvar_t	*developer;
cvar_t	*dedicated;
#else
extern cvar_t	*developer;
extern cvar_t	*dedicated;
#endif

#ifdef SIN
cvar_t  *fps;
cvar_t  *console_debug;
cvar_t  *surface_debug;
cvar_t  *copy_prot;
cvar_t  *sindemo;
// 2015 code 
cvar_t  *timescalenopitch; // disables sound pitching in timescale
cvar_t  *game2015; // specifies we are running the 2015 MP
#endif
FILE	*logfile;

int			server_state;

// host_speeds times
int		time_before_game;
int		time_after_game;
int		time_before_ref;
int		time_after_ref;

/*
============================================================================

CLIENT / SERVER interactions

============================================================================
*/

static int  rd_target;
static char	*rd_buffer;
static int	rd_buffersize;
static void	(*rd_flush)(int target, char *buffer);

void Com_BeginRedirect (int target, char *buffer, int buffersize, void (*flush)(int, char *))
{
	if (!target || !buffer || !buffersize || !flush)
		return;
	rd_target = target;
	rd_buffer = buffer;
	rd_buffersize = buffersize;
	rd_flush = flush;

	*rd_buffer = 0;
}

void Com_EndRedirect (void)
{
	rd_flush(rd_target, rd_buffer);

	rd_target = 0;
	rd_buffer = NULL;
	rd_buffersize = 0;
	rd_flush = NULL;
}

/*
=============
Com_Printf

Both client and server can use this, and it will output
to the apropriate place.
=============
*/
void Com_Printf (const char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];

	va_start (argptr,fmt);
	vsprintf (msg,fmt,argptr);
	va_end (argptr);

   if (rd_target)
	{
		if ((Burger::StringLength (msg) + Burger::StringLength(rd_buffer)) > (rd_buffersize - 1))
		{
			rd_flush(rd_target, rd_buffer);
			*rd_buffer = 0;
		}
		strcat (rd_buffer, msg);
		return;
	}

   Con_Print (msg);

	// also echo to debugging console
	Sys_ConsoleOutput (msg);

	// logfile
	if (logfile_active && logfile_active->value) {
		Debug::String(msg);
	}
}


/*
================
Com_DPrintf

A Com_Printf that only shows up if the "developer" cvar is set
================
*/
void Com_DPrintf (const char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];

	if (!developer || !developer->value)
		return;			// don't confuse non-developers with techie stuff...

	va_start (argptr,fmt);
	vsprintf (msg,fmt,argptr);
	va_end (argptr);

	Com_Printf ("%s", msg);
}


/*
=============
Com_Error

Both client and server can use this, and it will
do the apropriate things.
=============
*/
void Com_Error (int code, const char *fmt, ...)
{
	va_list		argptr;
	static char		msg[MAXPRINTMSG];
	static	qboolean	recursive;

#ifdef SIN
   //
   // always stop in the debugger
   //
//   assert( 0 );
#endif

	if (recursive)
		Sys_Error ("recursive error after: %s", msg);
	recursive = TRUE;

	va_start (argptr,fmt);
	vsprintf (msg,fmt,argptr);
	va_end (argptr);

	if (code == ERR_DISCONNECT)
	{
		CL_Drop ();
		recursive = FALSE;
		longjmp (abortframe, -1);
	}
	else if (code == ERR_DROP)
	{
		Com_Printf ("********************\nERROR: %s\n********************\n", msg);
		SV_Shutdown (va("Server crashed: %s\n", msg), FALSE);
		CL_Drop ();
		recursive = FALSE;
		longjmp (abortframe, -1);
	}
	else
	{
		SV_Shutdown (va("Server fatal crashed: %s\n", msg), FALSE);
		CL_Shutdown ();
	}

	if (logfile)
	{
		fclose (logfile);
		logfile = NULL;
	}

	Sys_Error ("%s", msg);
}


/*
=============
Com_Quit

Both client and server can use this, and it will
do the apropriate things.
=============
*/
void Com_Quit (void)
{
	SV_Shutdown ("Server quit\n", FALSE);
	CL_Shutdown ();

	if (logfile)
	{
		fclose (logfile);
		logfile = NULL;
	}

	Sys_Quit ();
}


/*
==================
Com_ServerState
==================
*/
int Com_ServerState (void)
{
	return server_state;
}

/*
==================
Com_SetServerState
==================
*/
void Com_SetServerState (int state)
{
	server_state = state;
}


/*
==============================================================================

			MESSAGE IO FUNCTIONS

Handles byte ordering and avoids alignment errors
==============================================================================
*/

vec3_t	bytedirs[NUMVERTEXNORMALS] =
{
#include "anorms.h"
};

//
// writing functions
//

void MSG_WriteChar (sizebuf_t *sb, int c)
{
	byte	*buf;

#ifdef PARANOID
	if (c < -128 || c > 127)
		Com_Error (ERR_FATAL, "MSG_WriteChar: range error");
#endif

	buf = (Word8*)SZ_GetSpace (sb, 1);
	buf[0] = c;
}

void MSG_WriteByte (sizebuf_t *sb, int c)
{
	byte	*buf;

#ifdef PARANOID
	if (c < 0 || c > 255)
		Com_Error (ERR_FATAL, "MSG_WriteByte: range error");
#endif

	buf = (Word8*)SZ_GetSpace (sb, 1);
	buf[0] = c;
}

void MSG_WriteShort (sizebuf_t *sb, int c)
{
	byte	*buf;

#ifdef PARANOID
	if (c < ((short)0x8000) || c > (short)0x7fff)
		Com_Error (ERR_FATAL, "MSG_WriteShort: range error");
#endif

	buf = (Word8*)SZ_GetSpace (sb, 2);
	buf[0] = c&0xff;
	buf[1] = c>>8;
}

void MSG_WriteLong (sizebuf_t *sb, int c)
{
	byte	*buf;

	buf = (Word8*)SZ_GetSpace (sb, 4);
	buf[0] = c&0xff;
	buf[1] = (c>>8)&0xff;
	buf[2] = (c>>16)&0xff;
	buf[3] = c>>24;
}

void MSG_WriteFloat (sizebuf_t *sb, float f)
{
	union
	{
		float	f;
		int	l;
	} dat;


	dat.f = f;
	dat.l = LittleLong (dat.l);

	SZ_Write (sb, &dat.l, 4);
}

void MSG_WriteString (sizebuf_t *sb, const char *s)
{
	if (!s)
		SZ_Write (sb, "", 1);
	else
		SZ_Write (sb, s, Burger::StringLength(s)+1);
}

void MSG_WriteCoord (sizebuf_t *sb, float f)
{
	MSG_WriteShort (sb, (int)(f*8));
}

void MSG_WritePos (sizebuf_t *sb,const vec3_t pos)
{
	MSG_WriteShort (sb, (int)(pos[0]*8));
	MSG_WriteShort (sb, (int)(pos[1]*8));
	MSG_WriteShort (sb, (int)(pos[2]*8));
}

void MSG_WriteAngle (sizebuf_t *sb, float f)
{
	MSG_WriteByte (sb, (int)(f*256/360) & 255);
}

void MSG_WriteAngle16 (sizebuf_t *sb, float f)
{
	MSG_WriteShort (sb, ANGLE2SHORT(f));
}


void MSG_WriteDeltaUsercmd (sizebuf_t *buf, usercmd_t *from, usercmd_t *cmd)
{
	int		bits;

//
// send the movement message
//
	bits = 0;
   if (cmd->angles[0] != from->angles[0])
		bits |= CM_ANGLE1;
	if (cmd->angles[1] != from->angles[1])
		bits |= CM_ANGLE2;
	if (cmd->angles[2] != from->angles[2])
		bits |= CM_ANGLE3;
	if (cmd->forwardmove != from->forwardmove)
		bits |= CM_FORWARD;
	if (cmd->sidemove != from->sidemove)
		bits |= CM_SIDE;
	if (cmd->upmove != from->upmove)
		bits |= CM_UP;
	if (cmd->buttons != from->buttons)
		bits |= CM_BUTTONS;
	if (cmd->impulse != from->impulse)
		bits |= CM_IMPULSE;

    MSG_WriteByte (buf, bits);

	if (bits & CM_ANGLE1)
		MSG_WriteShort (buf, cmd->angles[0]);
   if (bits & CM_ANGLE2)
		MSG_WriteShort (buf, cmd->angles[1]);
	if (bits & CM_ANGLE3)
		MSG_WriteShort (buf, cmd->angles[2]);

	if (bits & CM_FORWARD)
		MSG_WriteShort (buf, cmd->forwardmove);
	if (bits & CM_SIDE)
	  	MSG_WriteShort (buf, cmd->sidemove);
	if (bits & CM_UP)
		MSG_WriteShort (buf, cmd->upmove);

 	if (bits & CM_BUTTONS)
	  	MSG_WriteByte (buf, cmd->buttons);
 	if (bits & CM_IMPULSE)
	    MSG_WriteByte (buf, cmd->impulse);
   MSG_WriteByte (buf, cmd->msec);
#ifdef SIN
   {
   int templight;

   templight = cmd->lightlevel;
   if ( templight > 127 )
      templight = 127;
   templight >>= 3;
   templight |= (cmd->debris << 4);
	MSG_WriteByte (buf, templight);
   }
#else
	MSG_WriteByte (buf, cmd->lightlevel);
#endif
}


void MSG_WriteDir (sizebuf_t *sb, vec3_t dir)
{
	int		i, best;
	float	d, bestd;

	if (!dir)
	{
		MSG_WriteByte (sb, 0);
		return;
	}

	bestd = 0;
	best = 0;
	for (i=0 ; i<NUMVERTEXNORMALS ; i++)
	{
		d = DotProduct (dir, bytedirs[i]);
		if (d > bestd)
		{
			bestd = d;
			best = i;
		}
	}
	MSG_WriteByte (sb, best);
}


void MSG_ReadDir (sizebuf_t *sb, vec3_t dir)
{
	int		b;

	b = MSG_ReadByte (sb);
	if (b >= NUMVERTEXNORMALS)
		Com_Error (ERR_DROP, "MSF_ReadDir: out of range");
	VectorCopy (bytedirs[b], dir);
}


/*
==================
MSG_WriteDeltaEntity

Writes part of a packetentities message.
Can delta from either a baseline or a previous packet_entity
==================
*/
#ifdef SIN

void MSG_WriteDeltaEntity (entity_state_t *from, entity_state_t *to, sizebuf_t *msg, qboolean force, qboolean newentity)
{
	int		bits;

	if (!to->number)
		Com_Error (ERR_FATAL, "Unset entity number");
	if (to->number >= MAX_EDICTS)
		Com_Error (ERR_FATAL, "Entity number >= MAX_EDICTS");

// send an update
	bits = 0;

	if (to->number >= 256)
		bits |= U_NUMBER16;		// number8 is implicit otherwise

	if (to->origin[0] != from->origin[0])
		bits |= U_ORIGIN1;
	if (to->origin[1] != from->origin[1])
		bits |= U_ORIGIN2;
	if (to->origin[2] != from->origin[2])
		bits |= U_ORIGIN3;

	if ( to->angles[0] != from->angles[0] )
		bits |= U_ANGLE1;
	if ( to->angles[1] != from->angles[1] )
		bits |= U_ANGLE2;
	if ( to->angles[2] != from->angles[2] )
		bits |= U_ANGLE3;

	if ( to->skinnum != from->skinnum )
	{
		if ((unsigned)to->skinnum < 256)
			bits |= U_SKIN8;
		else if ((unsigned)to->skinnum < 0x10000)
			bits |= U_SKIN16;
		else
			bits |= (U_SKIN8|U_SKIN16);
	}

   if ( to->anim != from->anim )
      {
		bits |= U_FRAME16;
      }
	else if ( to->frame != from->frame )
   	{
		bits |= U_FRAME8;
	   }
   if ( SCALE_TO_PKTSCALE( to->scale )!= SCALE_TO_PKTSCALE( from->scale ) )
      {
		bits |= U_SCALE;
      }
   if (ALPHA_TO_PKTALPHA(to->alpha) != ALPHA_TO_PKTALPHA(from->alpha))
      {
      bits |= U_ALPHA;
      }
   if ( (COLOR_TO_PKTCOLOR(to->color_r) != COLOR_TO_PKTCOLOR(from->color_r)) ||
        (COLOR_TO_PKTCOLOR(to->color_g) != COLOR_TO_PKTCOLOR(from->color_g)) ||
        (COLOR_TO_PKTCOLOR(to->color_b) != COLOR_TO_PKTCOLOR(from->color_b)) ||
        (to->radius != from->radius))
      {
      bits |= U_COLOR;
      }
   if (
         (to->parent != from->parent) ||
         (to->bone.group_num != from->bone.group_num) ||
         (to->bone.tri_num != from->bone.tri_num) ||
         (to->bone.orientation[0] != from->bone.orientation[0] ) ||
         (to->bone.orientation[1] != from->bone.orientation[1] ) ||
         (to->bone.orientation[2] != from->bone.orientation[2] )
      )
      {
      bits |= U_BONE;
      }
   if (
         (to->numgroups != from->numgroups) ||
         (memcmp( to->groups, from->groups, to->numgroups))
      )
      {
      bits |= U_MODELGROUPS;
      }
   if (
         ( to->gunframe != from->gunframe ) ||
         ( to->gunanim != from->gunanim )
      )
      {
		bits |= U_GUNANIMFRAME;
      }

	if ( to->effects != from->effects )
	{
			bits |= U_EFFECTS;
	}

	if ( to->renderfx != from->renderfx )
	{
		bits |= U_RENDERFX16;
	}

	if ( to->solid != from->solid )
		bits |= U_SOLID;

	// event is not delta compressed, just 0 compressed
//	if ( to->event  )
//		bits |= U_EVENT;

	if ( (to->modelindex&0xff) != (from->modelindex&0xff) )
		bits |= U_MODEL;
	if ( (to->gunmodelindex&0xff) != (from->gunmodelindex&0xff) )
		bits |= U_GUNMODEL;
	if ( (to->modelindex>>8) != (from->modelindex>>8) )
		bits |= U_MODEL_16;
	if ( (to->gunmodelindex>>8) != (from->gunmodelindex>>8) )
		bits |= U_GUNMODEL_16;

	if ( to->sound != from->sound )
		bits |= U_SOUND;

   if (newentity || (to->renderfx & RF_BEAM) || (to->effects & EF_BEAMEFFECT))
      bits |= U_OLDORIGIN;

   if ( 
         !newentity && 
	      (  
            to->old_origin[0] != from->old_origin[0] ||
		      to->old_origin[1] != from->old_origin[1] ||
		      to->old_origin[2] != from->old_origin[2] 
         )
      )
		bits |= U_OLDORIGIN;		// FIXME: only when newly seen!!!

   if (to->lightofs != from->lightofs)
      bits |= U_LIGHTOFS;

	if ( to->prevframe != -1 )
		bits |= U_PREVFRAME;
	//
	// write the message
	//
	if (!bits && !force)
		return;		// nothing to send!

	//----------

	if (bits & 0xff000000)
		bits |= U_MOREBITS3 | U_MOREBITS2 | U_MOREBITS1;
	else if (bits & 0x00ff0000)
		bits |= U_MOREBITS2 | U_MOREBITS1;
	else if (bits & 0x0000ff00)
		bits |= U_MOREBITS1;

   MSG_WriteByte (msg,	bits&255 );

	if (bits & 0xff000000)
	{
		MSG_WriteByte (msg,	(bits>>8)&255 );
		MSG_WriteByte (msg,	(bits>>16)&255 );
		MSG_WriteByte (msg,	(bits>>24)&255 );
	}
	else if (bits & 0x00ff0000)
	{
		MSG_WriteByte (msg,	(bits>>8)&255 );
		MSG_WriteByte (msg,	(bits>>16)&255 );
	}
	else if (bits & 0x0000ff00)
	{
		MSG_WriteByte (msg,	(bits>>8)&255 );
	}

	//----------

	if (bits & U_NUMBER16)
		MSG_WriteShort (msg, to->number);
	else
		MSG_WriteByte (msg,	to->number);

	if (bits & U_MODEL)
		MSG_WriteByte (msg,	to->modelindex&0xff);
	if (bits & U_GUNMODEL)
		MSG_WriteByte (msg,	to->gunmodelindex&0xff);
	if (bits & U_MODEL_16)
		MSG_WriteByte (msg,	to->modelindex>>8);
	if (bits & U_GUNMODEL_16)
		MSG_WriteByte (msg,	to->gunmodelindex>>8);
	if (bits & U_FRAME8)
		MSG_WriteByte (msg, to->frame);
	if (bits & U_FRAME16)
		MSG_WriteShort (msg, (to->anim<<8) + to->frame);
	if (bits & U_SCALE)
		MSG_WriteShort (msg, SCALE_TO_PKTSCALE( to->scale ) );
   if (bits & U_ALPHA)
      MSG_WriteByte (msg, ALPHA_TO_PKTALPHA( to->alpha ) );
    if (bits & U_COLOR)
       {
       MSG_WriteByte (msg, COLOR_TO_PKTCOLOR( to->color_r ) );
       MSG_WriteByte (msg, COLOR_TO_PKTCOLOR( to->color_g ) );
       MSG_WriteByte (msg, COLOR_TO_PKTCOLOR( to->color_b ) );
       MSG_WriteShort (msg, to->radius);
       }
    if (bits & U_BONE)
       {
       MSG_WriteShort (msg, to->parent );
       MSG_WriteShort (msg, ((int)to->bone.group_num<<12) + to->bone.tri_num );
       MSG_WriteAngle (msg, to->bone.orientation[0] );
       MSG_WriteAngle (msg, to->bone.orientation[1] );
       MSG_WriteAngle (msg, to->bone.orientation[2] );
       }
   if (bits & U_MODELGROUPS)
      {
      int i;
      MSG_WriteByte (msg, to->numgroups );
      for (i=0;i<to->numgroups;i++)
         {
         MSG_WriteByte (msg, to->groups[i] );
         }
      }
	if (bits & U_GUNANIMFRAME)
      {
		MSG_WriteByte (msg, to->gunanim);
		MSG_WriteByte (msg, to->gunframe);
      }

	if ((bits & U_SKIN8) && (bits & U_SKIN16))		//used for laser colors
		MSG_WriteLong (msg, to->skinnum);
	else if (bits & U_SKIN8)
		MSG_WriteByte (msg, to->skinnum);
	else if (bits & U_SKIN16)
		MSG_WriteShort (msg, to->skinnum);

	if (bits & U_EFFECTS)
		{
		to->effects &= ~( EF_EFFECTS_16 | EF_EFFECTS_32 );
		if ( to->effects & 0xffff0000 )
			{
			to->effects |= EF_EFFECTS_32 | EF_EFFECTS_16;
			}
		else if ( to->effects & 0xff00 )
			{
			to->effects |= EF_EFFECTS_16;
			}
		MSG_WriteByte (msg, to->effects & 0xff );
		if ( to->effects & EF_EFFECTS_16 )
			{
			MSG_WriteByte (msg, ( to->effects >> 8 ) & 0xff );
			}
		if ( to->effects & EF_EFFECTS_32 )
			{
			MSG_WriteShort (msg, ( to->effects >> 16 ) & 0xffff );
			}
		}

	if (bits & U_RENDERFX16)
		MSG_WriteShort (msg, to->renderfx);

	if (bits & U_ORIGIN1)
		MSG_WriteCoord (msg, to->origin[0]);
	if (bits & U_ORIGIN2)
		MSG_WriteCoord (msg, to->origin[1]);
	if (bits & U_ORIGIN3)
		MSG_WriteCoord (msg, to->origin[2]);

	if (to->effects & EF_SMOOTHANGLES)
		{
		if (bits & U_ANGLE1)
			MSG_WriteAngle16(msg, to->angles[0]);
		if (bits & U_ANGLE2)
			MSG_WriteAngle16(msg, to->angles[1]);
		if (bits & U_ANGLE3)
			MSG_WriteAngle16(msg, to->angles[2]);
		}
	else
		{
		if (bits & U_ANGLE1)
			MSG_WriteAngle(msg, to->angles[0]);
		if (bits & U_ANGLE2)
			MSG_WriteAngle(msg, to->angles[1]);
		if (bits & U_ANGLE3)
			MSG_WriteAngle(msg, to->angles[2]);
		}

	if (bits & U_OLDORIGIN)
	{
		MSG_WriteCoord (msg, to->old_origin[0]);
		MSG_WriteCoord (msg, to->old_origin[1]);
		MSG_WriteCoord (msg, to->old_origin[2]);
	}

	if (bits & U_SOUND)
#ifdef SIN
		MSG_WriteShort (msg, to->sound);
#else
		MSG_WriteByte (msg, to->sound);
#endif
//	if (bits & U_EVENT)
//		MSG_WriteByte (msg, to->event);
	if (bits & U_SOLID)
		MSG_WriteLong (msg, to->solid);
	if (bits & U_LIGHTOFS)
		MSG_WriteByte (msg, to->lightofs);
	if (bits & U_PREVFRAME)
		MSG_WriteByte (msg, to->prevframe);
}


#else

void MSG_WriteDeltaEntity (entity_state_t *from, entity_state_t *to, sizebuf_t *msg, qboolean force)
{
	int		bits;

	if (!to->number)
		Com_Error (ERR_FATAL, "Unset entity number");
	if (to->number >= MAX_EDICTS)
		Com_Error (ERR_FATAL, "Entity number >= MAX_EDICTS");

// send an update
	bits = 0;

	if (to->number >= 256)
		bits |= U_NUMBER16;		// number8 is implicit otherwise

	if (to->origin[0] != from->origin[0])
		bits |= U_ORIGIN1;
	if (to->origin[1] != from->origin[1])
		bits |= U_ORIGIN2;
	if (to->origin[2] != from->origin[2])
		bits |= U_ORIGIN3;

	if ( to->angles[0] != from->angles[0] )
		bits |= U_ANGLE1;
	if ( to->angles[1] != from->angles[1] )
		bits |= U_ANGLE2;
	if ( to->angles[2] != from->angles[2] )
		bits |= U_ANGLE3;

	if ( to->skinnum != from->skinnum )
	{
		if ((unsigned)to->skinnum < 256)
			bits |= U_SKIN8;
		else if ((unsigned)to->skinnum < 0x10000)
			bits |= U_SKIN16;
		else
			bits |= (U_SKIN8|U_SKIN16);
	}

	if ( to->frame != from->frame )
	{
		if (to->frame < 256)
			bits |= U_FRAME8;
		else
			bits |= U_FRAME16;
	}

	if ( to->effects != from->effects )
	{
		if (to->effects < 256)
			bits |= U_EFFECTS8;
		else if (to->effects < 0x8000)
			bits |= U_EFFECTS16;
		else
			bits |= U_EFFECTS8|U_EFFECTS16;
	}

	if ( to->renderfx != from->renderfx )
	{
		if (to->renderfx < 256)
			bits |= U_RENDERFX8;
		else if (to->renderfx < 0x8000)
			bits |= U_RENDERFX16;
		else
			bits |= U_RENDERFX8|U_RENDERFX16;
	}

	if ( to->solid != from->solid )
		bits |= U_SOLID;

	// event is not delta compressed, just 0 compressed
	if ( to->event  )
		bits |= U_EVENT;

	if ( to->modelindex != from->modelindex )
		bits |= U_MODEL;
	if ( to->modelindex2 != from->modelindex2 )
		bits |= U_MODEL2;
	if ( to->modelindex3 != from->modelindex3 )
		bits |= U_MODEL3;
	if ( to->modelindex4 != from->modelindex4 )
		bits |= U_MODEL4;

	if ( to->sound != from->sound )
		bits |= U_SOUND;

	if (to->old_origin[0] != from->old_origin[0]
		|| to->old_origin[1] != from->old_origin[1]
		|| to->old_origin[2] != from->old_origin[2] )
		bits |= U_OLDORIGIN;		// FIXME: only when newly seen!!!

	//
	// write the message
	//
	if (!bits && !force)
		return;		// nothing to send!

	//----------

	if (bits & 0xff000000)
		bits |= U_MOREBITS3 | U_MOREBITS2 | U_MOREBITS1;
	else if (bits & 0x00ff0000)
		bits |= U_MOREBITS2 | U_MOREBITS1;
	else if (bits & 0x0000ff00)
		bits |= U_MOREBITS1;

	MSG_WriteByte (msg,	bits&255 );

	if (bits & 0xff000000)
	{
		MSG_WriteByte (msg,	(bits>>8)&255 );
		MSG_WriteByte (msg,	(bits>>16)&255 );
		MSG_WriteByte (msg,	(bits>>24)&255 );
	}
	else if (bits & 0x00ff0000)
	{
		MSG_WriteByte (msg,	(bits>>8)&255 );
		MSG_WriteByte (msg,	(bits>>16)&255 );
	}
	else if (bits & 0x0000ff00)
	{
		MSG_WriteByte (msg,	(bits>>8)&255 );
	}

	//----------

	if (bits & U_NUMBER16)
		MSG_WriteShort (msg, to->number);
	else
		MSG_WriteByte (msg,	to->number);

	if (bits & U_MODEL)
		MSG_WriteByte (msg,	to->modelindex);
	if (bits & U_MODEL2)
		MSG_WriteByte (msg,	to->modelindex2);
	if (bits & U_MODEL3)
		MSG_WriteByte (msg,	to->modelindex3);
	if (bits & U_MODEL4)
		MSG_WriteByte (msg,	to->modelindex4);
	if (bits & U_FRAME8)
		MSG_WriteByte (msg, to->frame);
	if (bits & U_FRAME16)
		MSG_WriteShort (msg, to->frame);

	if ((bits & U_SKIN8) && (bits & U_SKIN16))		//used for laser colors
		MSG_WriteLong (msg, to->skinnum);
	else if (bits & U_SKIN8)
		MSG_WriteByte (msg, to->skinnum);
	else if (bits & U_SKIN16)
		MSG_WriteShort (msg, to->skinnum);

	if ( (bits & (U_EFFECTS8|U_EFFECTS16)) == (U_EFFECTS8|U_EFFECTS16) )
		MSG_WriteLong (msg, to->effects);
	else if (bits & U_EFFECTS8)
		MSG_WriteByte (msg, to->effects);
	else if (bits & U_EFFECTS16)
		MSG_WriteShort (msg, to->effects);
	if ( (bits & (U_RENDERFX8|U_RENDERFX16)) == (U_RENDERFX8|U_RENDERFX16) )
		MSG_WriteLong (msg, to->renderfx);
	else if (bits & U_RENDERFX8)
		MSG_WriteByte (msg, to->renderfx);
	else if (bits & U_RENDERFX16)
		MSG_WriteShort (msg, to->renderfx);

	if (bits & U_ORIGIN1)
		MSG_WriteCoord (msg, to->origin[0]);
	if (bits & U_ORIGIN2)
		MSG_WriteCoord (msg, to->origin[1]);
	if (bits & U_ORIGIN3)
		MSG_WriteCoord (msg, to->origin[2]);

	if (bits & U_ANGLE1)
		MSG_WriteAngle(msg, to->angles[0]);
	if (bits & U_ANGLE2)
		MSG_WriteAngle(msg, to->angles[1]);
	if (bits & U_ANGLE3)
		MSG_WriteAngle(msg, to->angles[2]);

	if (bits & U_OLDORIGIN)
	{
		MSG_WriteCoord (msg, to->old_origin[0]);
		MSG_WriteCoord (msg, to->old_origin[1]);
		MSG_WriteCoord (msg, to->old_origin[2]);
	}

	if (bits & U_SOUND)
		MSG_WriteByte (msg, to->sound);
	if (bits & U_EVENT)
		MSG_WriteByte (msg, to->event);
	if (bits & U_SOLID)
		MSG_WriteShort (msg, to->solid);
}
#endif


#ifdef SIN

/*
============================
MSG_WriteDeltaConsoleForDemo
============================
*/
void MSG_WriteDeltaConsoleForDemo (console_state_t *from, console_state_t *to, sizebuf_t *msg, qboolean force, char *conbuffer)
   {
	int		bits;
   byte     num_statusbars_to_init = 0;
   byte     num_statusbars_to_change = 0;
   int      i;

	//
   // Build up the bits
   //
	bits = 0;

   bits |= C_CONPRINT;

   if (to->create_time != from->create_time)
      bits |= C_CREATECONSOLE;

   if (to->layout_update_time != from->layout_update_time)
      bits |= C_LAYOUT;

   if (to->virtual_width != from->virtual_width)
      bits |= C_VIRTUAL_WIDTH;

   if (to->virtual_height != from->virtual_height)
      bits |= C_VIRTUAL_HEIGHT;

   if (to->fraction != from->fraction)
      bits |= C_FRACTION;

   if (to->menufile_update_time != from->menufile_update_time)
      bits |= C_MENUFILE;

   if (to->focus != from->focus)
      bits |= C_FOCUS;

   if (to->rows != from->rows)
      bits |= C_ROWS;

   if (to->cols != from->cols)
      bits |= C_COLUMNS;

   if (to->consoleactive != from->consoleactive)
      bits |= C_ACTIVE;

   if (to->menuactive != from->menuactive)
      bits |= C_MENUACTIVE;

   if ( (COLOR_TO_PKTCOLOR(to->red)   != COLOR_TO_PKTCOLOR(from->red))   ||
        (COLOR_TO_PKTCOLOR(to->green) != COLOR_TO_PKTCOLOR(from->green)) ||
        (COLOR_TO_PKTCOLOR(to->blue)  != COLOR_TO_PKTCOLOR(from->blue))  ||
        (COLOR_TO_PKTCOLOR(to->alpha) != COLOR_TO_PKTCOLOR(from->alpha))
       )
      {
      bits |= C_COLOR;
      }

   if (to->spawnflags != from->spawnflags)
      bits |= C_TYPE;

   if (to->name_update_time != from->name_update_time)
      bits |= C_NAME;

   // Count up all the status bars to send for this console
   for (i=0;i<MAX_STATUSBARS;i++)
      {
      if (to->sbar[i].update_time != from->sbar[i].update_time)
         {
         num_statusbars_to_init++;
         bits |= C_STATUSBAR_INIT;
         }
      if (to->sbar[i].value != from->sbar[i].value)
         {
         num_statusbars_to_change++;
         bits |= C_STATUSBAR_VALUE;
         }
      }

   if ((from->menu_level != to->menu_level) ||
       (from->sel_menu_item != to->sel_menu_item))
      bits |= C_MENUINFO;

   //
	// write the message
	//
	if (!bits && !force)
		return;		// nothing to send!

   //---------- CONSOLE NUMBER
   MSG_WriteShort (msg, to->number);

	//---------- BITS
	MSG_WriteLong (msg, bits);

   //---------- LAYOUT
   if (bits & C_LAYOUT)
      {
      MSG_WriteString (msg, to->layout);
      if (console_debug->value)
         Com_DPrintf("Setting layout from \"%s\" to \"%s\"\n", from->layout, to->layout);
      }

   //---------- VIRTUAL WIDTH
   if (bits & C_VIRTUAL_WIDTH)
      {
      MSG_WriteFloat (msg, to->virtual_width);
      if (console_debug->value)
         Com_DPrintf("Setting virtwidth from \"%f\" to \"%f\"\n", from->virtual_width, to->virtual_width);
      }

   //---------- VIRTUAL HEIGHT
   if (bits & C_VIRTUAL_HEIGHT)
      {
      MSG_WriteFloat (msg, to->virtual_height);
      if (console_debug->value)
         Com_DPrintf("Setting virtheight from \"%f\" to \"%f\"\n", from->virtual_height, to->virtual_height);
      }

   //---------- FRACTION
   if (bits & C_FRACTION)
      {
      MSG_WriteFloat (msg, to->fraction);
      if (console_debug->value)
         Com_DPrintf("Setting fraction from \"%f\" to \"%f\"\n", from->fraction, to->fraction);
      }

   //---------- MENU FILE
   if (bits & C_MENUFILE)
      {
      MSG_WriteString (msg, to->menu_file);
      if (console_debug->value)
         Com_DPrintf("Setting menufile from \"%s\" to \"%s\"\n", from->menu_file, to->menu_file);
      }

   //---------- FOCUS
   if (bits & C_FOCUS)
      {
      MSG_WriteByte (msg, to->focus);
      if (console_debug->value)
         Com_DPrintf("Setting focus from \"%d\" to \"%d\"\n", from->focus, to->focus);
      }

   //---------- ROWS
   if (bits & C_ROWS)
      {
      MSG_WriteShort (msg, to->rows);
      if (console_debug->value)
         Com_DPrintf("Setting rows from \"%d\" to \"%d\"\n", from->rows, to->rows);
      }

   //---------- COLS
   if (bits & C_COLUMNS)
      {
      MSG_WriteShort (msg, to->cols);
      if (console_debug->value)
         Com_DPrintf("Setting columns from \"%d\" to \"%d\"\n", from->cols, to->cols);
      }
   //---------- APPEND CONSOLE BUFFER
   if (bits & C_CONPRINT)
      {
      // Send over the buffer
      MSG_WriteString (msg, conbuffer);
      }

   //---------- ACTIVE
   if (bits & C_ACTIVE)
      {
      MSG_WriteByte (msg, to->consoleactive);
      if (console_debug->value)
         Com_DPrintf("Setting active from %d to %d\n", from->consoleactive, to->consoleactive);
      }

   //---------- MENU ACTIVE
   if (bits & C_MENUACTIVE)
      {
      MSG_WriteByte (msg, to->menuactive);
      if (console_debug->value)
         Com_DPrintf("Setting menuactive from %d to %d\n", from->menuactive, to->menuactive);
      }

   //--------- FOREGROUND COLOR
   if (bits & C_COLOR)
      {
      MSG_WriteByte (msg, COLOR_TO_PKTCOLOR(to->red));
      MSG_WriteByte (msg, COLOR_TO_PKTCOLOR(to->green));
      MSG_WriteByte (msg, COLOR_TO_PKTCOLOR(to->blue));
      MSG_WriteByte (msg, COLOR_TO_PKTCOLOR(to->alpha));
      if (console_debug->value)
         Com_DPrintf("Setting color from r:%0.2f g:%0.2f b:%0.2f a:%0.2f to r:%0.2f g:%0.2f b:%0.2f a:%0.2f\n",
            from->red,from->green,from->blue,from->alpha,
            to->red,to->green,to->blue,to->alpha);
      }

   //--------- CONSOLE TYPE
   if (bits & C_TYPE)
      {
      MSG_WriteByte(msg, to->spawnflags);
      if (console_debug->value)
         Com_DPrintf("Setting type from %d to %d\n",from->spawnflags, to->spawnflags);
      }

   //--------- CONSOLE NAME
   if (bits & C_NAME)
      {
      MSG_WriteString(msg, to->console_name);
      if (console_debug->value)
         Com_DPrintf("Setting name from %s to %s\n",from->console_name, to->console_name);
      }

   // -------- STATUS BAR INIT
   if (bits & C_STATUSBAR_INIT)
      {
      MSG_WriteByte(msg, num_statusbars_to_init);
      for (i=0;i<MAX_STATUSBARS;i++)
         {
         if (to->sbar[i].update_time != from->sbar[i].update_time)
            {
            MSG_WriteByte(msg, i);
            MSG_WriteFloat(msg, to->sbar[i].width);
            MSG_WriteFloat(msg, to->sbar[i].height);
            MSG_WriteFloat(msg, to->sbar[i].min);
            MSG_WriteFloat(msg, to->sbar[i].max);
            MSG_WriteFloat(msg, to->sbar[i].value);
            MSG_WriteByte(msg, COLOR_TO_PKTCOLOR(to->sbar[i].red));
            MSG_WriteByte(msg, COLOR_TO_PKTCOLOR(to->sbar[i].green));
            MSG_WriteByte(msg, COLOR_TO_PKTCOLOR(to->sbar[i].blue));
            MSG_WriteByte(msg, COLOR_TO_PKTCOLOR(to->sbar[i].alpha));
            if (console_debug->value)
               Com_DPrintf("Creating sbar %d\n",i);
            }
         }
      }

   //--------- STATUS BAR VALUE
   if (bits & C_STATUSBAR_VALUE)
      {
      MSG_WriteByte(msg, num_statusbars_to_change);
      for (i=0;i<MAX_STATUSBARS;i++)
         {
         if (to->sbar[i].value != from->sbar[i].value)
            {
            MSG_WriteByte(msg, i);
            MSG_WriteFloat(msg, to->sbar[i].value);
            }
         if (console_debug->value)
               Com_DPrintf("Setting sbar %d from %f to %f\n",i,from->sbar[i].value,to->sbar[i].value);
         }
      }

   //--------- MENU INFORMATION
   if (bits & C_MENUINFO)
      {
      MSG_WriteShort(msg, to->menu_level);
      MSG_WriteShort(msg, to->sel_menu_item);
      if (console_debug->value)
         Com_DPrintf("Setting level to %d and item to %d\n",to->menu_level, to->sel_menu_item);
      }
   }


/*
=====================
MSG_WriteDeltaConsole
=====================
*/
void MSG_WriteDeltaConsole (console_state_t *from, console_state_t *to,
                            console_buffer_state_t *bs, sizebuf_t *msg,
                            qboolean force, int *update_index, qboolean update_return)
{
	int		bits;
   byte     num_statusbars_to_init = 0;
   byte     num_statusbars_to_change = 0;
   int      i;

	//
   // Build up the bits
   //
	bits = 0;

   if (to->create_time != from->create_time)
      bits |= C_CREATECONSOLE;

   if (to->layout_update_time != from->layout_update_time)
      bits |= C_LAYOUT;

   if (to->virtual_width != from->virtual_width)
      bits |= C_VIRTUAL_WIDTH;

   if (to->virtual_height != from->virtual_height)
      bits |= C_VIRTUAL_HEIGHT;

   if (to->fraction != from->fraction)
      bits |= C_FRACTION;

   if (to->menufile_update_time != from->menufile_update_time)
      bits |= C_MENUFILE;

   if (to->layoutfile_update_time != from->layoutfile_update_time)
      bits |= C_LAYOUTFILE;

   if (to->focus != from->focus)
      bits |= C_FOCUS;

   if (to->rows != from->rows)
      bits |= C_ROWS;

   if (to->cols != from->cols)
      bits |= C_COLUMNS;

   if (bs && update_index)
      if (bs->end_index != *update_index)
         bits |= C_CONPRINT;

   if (to->consoleactive != from->consoleactive)
      bits |= C_ACTIVE;

   if (to->menuactive != from->menuactive)
      bits |= C_MENUACTIVE;

   if ( (COLOR_TO_PKTCOLOR(to->red)   != COLOR_TO_PKTCOLOR(from->red))   ||
        (COLOR_TO_PKTCOLOR(to->green) != COLOR_TO_PKTCOLOR(from->green)) ||
        (COLOR_TO_PKTCOLOR(to->blue)  != COLOR_TO_PKTCOLOR(from->blue))  ||
        (COLOR_TO_PKTCOLOR(to->alpha) != COLOR_TO_PKTCOLOR(from->alpha))
       )
      {
      bits |= C_COLOR;
      }

   if (to->cleared_console_time != from->cleared_console_time)
      {
      *update_index = 0;
      bits |= C_CLEARCONSOLE;
      }

   if (to->spawnflags != from->spawnflags)
      bits |= C_TYPE;

   if (to->name_update_time != from->name_update_time)
      bits |= C_NAME;

   // Count up all the status bars to send for this console
   for (i=0;i<MAX_STATUSBARS;i++)
      {
      if (to->sbar[i].update_time != from->sbar[i].update_time)
         {
         num_statusbars_to_init++;
         bits |= C_STATUSBAR_INIT;
         }
      if (to->sbar[i].value != from->sbar[i].value)
         {
         num_statusbars_to_change++;
         bits |= C_STATUSBAR_VALUE;
         }
      }

   if (update_return)
      {
      if (to->console_return_time != from->console_return_time)
         bits |= C_LINERET;
      }

   if ((to->linepos - from->linepos) > 0)
      bits |= C_LINEPOS;
   else if ((to->linepos - from->linepos) < 0)
      bits |= C_LINENEG;

   if ((from->menu_level != to->menu_level) ||
       (from->sel_menu_item != to->sel_menu_item))
      bits |= C_MENUINFO;

   //
	// write the message
	//
	if (!bits && !force)
		return;		// nothing to send!

   //---------- CONSOLE NUMBER
   MSG_WriteShort (msg, to->number);

	//---------- BITS
	MSG_WriteLong (msg, bits);

   //---------- LAYOUT
   if (bits & C_LAYOUT)
      {
      MSG_WriteString (msg, to->layout);
      if (console_debug->value)
         Com_DPrintf("Setting layout from \"%s\" to \"%s\"\n", from->layout, to->layout);
      }

   //---------- VIRTUAL WIDTH
   if (bits & C_VIRTUAL_WIDTH)
      {
      MSG_WriteFloat (msg, to->virtual_width);
      if (console_debug->value)
         Com_DPrintf("Setting virtwidth from \"%f\" to \"%f\"\n", from->virtual_width, to->virtual_width);
      }

   //---------- VIRTUAL HEIGHT
   if (bits & C_VIRTUAL_HEIGHT)
      {
      MSG_WriteFloat (msg, to->virtual_height);
      if (console_debug->value)
         Com_DPrintf("Setting virtheight from \"%f\" to \"%f\"\n", from->virtual_height, to->virtual_height);
      }

   //---------- FRACTION
   if (bits & C_FRACTION)
      {
      MSG_WriteFloat (msg, to->fraction);
      if (console_debug->value)
         Com_DPrintf("Setting fraction from \"%f\" to \"%f\"\n", from->fraction, to->fraction);
      }

   //---------- MENU FILE
   if (bits & C_MENUFILE)
      {
      MSG_WriteString (msg, to->menu_file);
      if (console_debug->value)
         Com_DPrintf("Setting menufile from \"%s\" to \"%s\"\n", from->menu_file, to->menu_file);
      }

   //---------- LAYOUT FILE
   if (bits & C_LAYOUTFILE)
      {
      MSG_WriteString (msg, to->layout_file);
      if (console_debug->value)
         Com_DPrintf("Setting layoutfile from \"%s\" to \"%s\"\n", from->layout_file, to->layout_file);
      }

   //---------- ROWS
   if (bits & C_ROWS)
      {
      MSG_WriteShort (msg, to->rows);
      if (console_debug->value)
         Com_DPrintf("Setting rows from \"%d\" to \"%d\"\n", from->rows, to->rows);
      }

   //---------- COLS
   if (bits & C_COLUMNS)
      {
      MSG_WriteShort (msg, to->cols);
      if (console_debug->value)
         Com_DPrintf("Setting columns from \"%d\" to \"%d\"\n", from->cols, to->cols);
      }
   //---------- APPEND CONSOLE BUFFER
   if (bits & C_CONPRINT)
      {
      int  i,j;
      char *buffptr;
      char temp[1024];
      buffptr = &bs->buffer[0];

      if (*update_index < bs->start_index)
         {
         *update_index = bs->start_index;
         }


      for (i=*update_index,j=0;((i<bs->end_index) && (j<100));i++,j++)
         {
         temp[j] = *(buffptr+(i%MAX_BUFFER_LENGTH));
         }

      temp[j] = 0;
      MSG_WriteString (msg, temp);
      *update_index += j;//bs->end_index;
      }

   //---------- ACTIVE
   if (bits & C_ACTIVE)
      {
      MSG_WriteByte (msg, to->consoleactive);
      if (console_debug->value)
         Com_DPrintf("Setting active from %d to %d\n", from->consoleactive, to->consoleactive);
      }

   //---------- MENU ACTIVE
   if (bits & C_MENUACTIVE)
      {
      MSG_WriteByte (msg, to->menuactive);
      if (console_debug->value)
         Com_DPrintf("Setting menuactive from %d to %d\n", from->menuactive, to->menuactive);
      }

   //--------- FOREGROUND COLOR
   if (bits & C_COLOR)
      {
      MSG_WriteByte (msg, COLOR_TO_PKTCOLOR(to->red));
      MSG_WriteByte (msg, COLOR_TO_PKTCOLOR(to->green));
      MSG_WriteByte (msg, COLOR_TO_PKTCOLOR(to->blue));
      MSG_WriteByte (msg, COLOR_TO_PKTCOLOR(to->alpha));
      if (console_debug->value)
         Com_DPrintf("Setting color from r:%0.2f g:%0.2f b:%0.2f a:%0.2f to r:%0.2f g:%0.2f b:%0.2f a:%0.2f\n",
            from->red,from->green,from->blue,from->alpha,
            to->red,to->green,to->blue,to->alpha);
      }

   //--------- CONSOLE TYPE
   if (bits & C_TYPE)
      {
      MSG_WriteByte(msg, to->spawnflags);
      if (console_debug->value)
         Com_DPrintf("Setting type from %d to %d\n",from->spawnflags, to->spawnflags);
      }

   //---------- FOCUS
   if (bits & C_FOCUS)
      {
      MSG_WriteByte (msg, to->focus);
      if (console_debug->value)
         Com_DPrintf("Setting focus from \"%d\" to \"%d\"\n", from->focus, to->focus);
      }

   //--------- CONSOLE NAME
   if (bits & C_NAME)
      {
      MSG_WriteString(msg, to->console_name);
      if (console_debug->value)
         Com_DPrintf("Setting name from %s to %s\n",from->console_name, to->console_name);
      }

   // -------- STATUS BAR INIT
   if (bits & C_STATUSBAR_INIT)
      {
      MSG_WriteByte(msg, num_statusbars_to_init);
      for (i=0;i<MAX_STATUSBARS;i++)
         {
         if (to->sbar[i].update_time != from->sbar[i].update_time)
            {
            MSG_WriteByte(msg, i);
            MSG_WriteFloat(msg, to->sbar[i].width);
            MSG_WriteFloat(msg, to->sbar[i].height);
            MSG_WriteFloat(msg, to->sbar[i].min);
            MSG_WriteFloat(msg, to->sbar[i].max);
            MSG_WriteFloat(msg, to->sbar[i].value);
            MSG_WriteByte(msg, COLOR_TO_PKTCOLOR(to->sbar[i].red));
            MSG_WriteByte(msg, COLOR_TO_PKTCOLOR(to->sbar[i].green));
            MSG_WriteByte(msg, COLOR_TO_PKTCOLOR(to->sbar[i].blue));
            MSG_WriteByte(msg, COLOR_TO_PKTCOLOR(to->sbar[i].alpha));
            if (console_debug->value)
               Com_DPrintf("Creating sbar %d\n",i);
            }
         }
      }

   //--------- STATUS BAR VALUE
   if (bits & C_STATUSBAR_VALUE)
      {
      MSG_WriteByte(msg, num_statusbars_to_change);
      for (i=0;i<MAX_STATUSBARS;i++)
         {
         if (to->sbar[i].value != from->sbar[i].value)
            {
            MSG_WriteByte(msg, i);
            MSG_WriteFloat(msg, to->sbar[i].value);
            }
         if (console_debug->value)
               Com_DPrintf("Setting sbar %d from %f to %f\n",i,from->sbar[i].value,to->sbar[i].value);
         }
      }

   //--------- LINE POSITION IN CONSOLE
   if (bits & C_LINERET)
      {
      if (console_debug->value)
         Com_DPrintf("Setting carriage return\n");
      }


   //--------- LINE POSITION IN CONSOLE
   if (bits & C_LINEPOS)
      {
      MSG_WriteString(msg, &to->cmdline[from->linepos]);
      if (console_debug->value)
         Com_DPrintf("Setting linepos to %d and adding \"%s\" to buffer\n",to->linepos,&to->cmdline[from->linepos]);
      }

   //--------- LINE POSITION IN CONSOLE

   if (bits & C_LINENEG)
      {
      MSG_WriteShort(msg, to->linepos);
      if (console_debug->value)
         Com_DPrintf("Setting linepos to %d\n",to->linepos);
      }

   //--------- MENU INFORMATION
   if (bits & C_MENUINFO)
      {
      MSG_WriteShort(msg, to->menu_level);
      MSG_WriteShort(msg, to->sel_menu_item);
      if (console_debug->value)
         Com_DPrintf("Setting level to %d and item to %d\n",to->menu_level, to->sel_menu_item);
      }
   if (console_debug->value && bits)
      Com_DPrintf("\n\n");
   }



/*
=====================
MSG_WriteDeltaSurface
=====================
*/
void MSG_WriteDeltaSurface (surface_state_t *from, surface_state_t *to,
                            sizebuf_t *msg, qboolean force)
{
	int		bits = 0;

	//
   // Build up the bits
   //

   if (to->trans_state != from->trans_state)
      bits |= S_TRANS_STATE;

   if (to->trans_mag != from->trans_mag)
      bits |= S_TRANS_MAG;

   if (to->trans_angle != from->trans_angle)
      bits |= S_TRANS_ANGLE;

   if (to->translucence != from->translucence)
      bits |= S_TRANSLUCENCE;

   if (to->magnitude != from->magnitude)
      bits |= S_MAGNITUDE;

   if (to->frequency != from->frequency)
      bits |= S_FREQUENCY;

   if (to->damage_frame != from->damage_frame)
      bits |= S_DAMAGE_FRAME;

   //
	// write the message
	//
	if (!bits && !force)
		return;		// nothing to send!

   //---------- SURFACE GROUP NUMBER
   MSG_WriteShort (msg, to->groupnumber);

	//---------- BITS
	MSG_WriteByte (msg, bits);

   //---------- TRANSLATION MAGNITUDE
   if (bits & S_TRANS_MAG)
      {
      MSG_WriteFloat (msg, to->trans_mag);
      if (surface_debug->value)
         Com_DPrintf("Setting trans_mag from \"%f\" to \"%f\"\n", from->trans_mag, to->trans_mag);
      }

   //---------- TRANSLATION ANGLE
   if (bits & S_TRANS_ANGLE)
      {
      MSG_WriteFloat (msg, to->trans_angle);
      if (surface_debug->value)
         Com_DPrintf("Setting trans_angle from \"%f\" to \"%f\"\n", from->trans_angle, to->trans_angle);
      }

   //---------- TRANSLATION STATE
   if (bits & S_TRANS_STATE)
      {
      MSG_WriteByte (msg, to->trans_state);
      if (surface_debug->value)
         Com_DPrintf("Setting trans_state from \"%d\" to \"%d\"\n", from->trans_state, to->trans_state);
      }

   //---------- TRANSLUCENCE
   if (bits & S_TRANSLUCENCE)
      {
      MSG_WriteFloat (msg, to->translucence);
      if (surface_debug->value)
         Com_DPrintf("Setting translucence from \"%f\" to \"%f\"\n", from->translucence, to->translucence);
      }

   //---------- MAGNITUDE
   if (bits & S_MAGNITUDE)
      {
      MSG_WriteFloat (msg, to->magnitude);
      if (surface_debug->value)
         Com_DPrintf("Setting magnitude from \"%f\" to \"%f\"\n", from->magnitude, to->magnitude);
      }

   //---------- FREQUENCY
   if (bits & S_FREQUENCY)
      {
      MSG_WriteFloat (msg, to->frequency);
      if (surface_debug->value)
         Com_DPrintf("Setting frequency from \"%f\" to \"%f\"\n", from->frequency, to->frequency);
      }

   //---------- DAMAGE
   if (bits & S_DAMAGE_FRAME)
      {
      MSG_WriteShort (msg, to->damage_frame);
      if (surface_debug->value)
         Com_DPrintf("Setting damage frame from \"%d\" to \"%d\"\n", from->damage_frame, to->damage_frame);
      }
   }
#endif

//============================================================

//
// reading functions
//

void MSG_BeginReading (sizebuf_t *msg)
{
	msg->readcount = 0;
}

// returns -1 if no more characters are available
int MSG_ReadChar (sizebuf_t *msg_read)
{
	int	c;

	if (msg_read->readcount+1 > msg_read->cursize)
		c = -1;
	else
		c = (signed char)msg_read->data[msg_read->readcount];
	msg_read->readcount++;

	return c;
}

int MSG_ReadByte (sizebuf_t *msg_read)
{
	int	c;

	if (msg_read->readcount+1 > msg_read->cursize)
		c = -1;
	else
		c = (unsigned char)msg_read->data[msg_read->readcount];
	msg_read->readcount++;

	return c;
}

int MSG_ReadShort (sizebuf_t *msg_read)
{
	int	c;

	if (msg_read->readcount+2 > msg_read->cursize)
		c = -1;
	else
		c = (short)(msg_read->data[msg_read->readcount]
		+ (msg_read->data[msg_read->readcount+1]<<8));

	msg_read->readcount += 2;

	return c;
}

int MSG_ReadLong (sizebuf_t *msg_read)
{
	int	c;

	if (msg_read->readcount+4 > msg_read->cursize)
		c = -1;
	else
		c = msg_read->data[msg_read->readcount]
		+ (msg_read->data[msg_read->readcount+1]<<8)
		+ (msg_read->data[msg_read->readcount+2]<<16)
		+ (msg_read->data[msg_read->readcount+3]<<24);

	msg_read->readcount += 4;

	return c;
}

float MSG_ReadFloat (sizebuf_t *msg_read)
{
	union
	{
		byte	b[4];
		float	f;
		int	l;
	} dat;

	if (msg_read->readcount+4 > msg_read->cursize)
		dat.f = -1;
	else
	{
		dat.b[0] =	msg_read->data[msg_read->readcount];
		dat.b[1] =	msg_read->data[msg_read->readcount+1];
		dat.b[2] =	msg_read->data[msg_read->readcount+2];
		dat.b[3] =	msg_read->data[msg_read->readcount+3];
	}
	msg_read->readcount += 4;

	dat.l = LittleLong (dat.l);

	return dat.f;
}

const char *MSG_ReadString (sizebuf_t *msg_read)
{
	static char	string[2048];
	int		l,c;

	l = 0;
	do
	{
		c = MSG_ReadChar (msg_read);
		if (c == -1 || c == 0)
			break;
		string[l] = c;
		l++;
	} while (l < sizeof(string)-1);

	string[l] = 0;

	return string;
}

const char *MSG_ReadStringLine (sizebuf_t *msg_read)
{
	static char	string[2048];
	int		l,c;

	l = 0;
	do {
		c = MSG_ReadChar (msg_read);
		if (c == -1 || c == 0 || c == '\n') {
			break;
		}
		string[l] = c;
		l++;
	} while (l < sizeof(string)-1);

	string[l] = 0;

	return string;
}

float MSG_ReadCoord (sizebuf_t *msg_read)
{
	return MSG_ReadShort(msg_read) * (1.0/8);
}

void MSG_ReadPos (sizebuf_t *msg_read, vec3_t pos)
{
	pos[0] = MSG_ReadShort(msg_read) * (1.0/8);
	pos[1] = MSG_ReadShort(msg_read) * (1.0/8);
	pos[2] = MSG_ReadShort(msg_read) * (1.0/8);
}

float MSG_ReadAngle (sizebuf_t *msg_read)
{
	return MSG_ReadChar(msg_read) * (360.0/256);
}

float MSG_ReadAngle16 (sizebuf_t *msg_read)
{
	return SHORT2ANGLE(MSG_ReadShort(msg_read));
}

void MSG_ReadDeltaUsercmd (sizebuf_t *msg_read, usercmd_t *from, usercmd_t *move)
{
	int bits;

	FastMemCpy(move, from, sizeof(*move));

	bits = MSG_ReadByte (msg_read);

// read current angles
	if (bits & CM_ANGLE1)
		move->angles[0] = MSG_ReadShort (msg_read);
   if (bits & CM_ANGLE2)
		move->angles[1] = MSG_ReadShort (msg_read);
	if (bits & CM_ANGLE3)
		move->angles[2] = MSG_ReadShort (msg_read);

// read movement
	if (bits & CM_FORWARD)
		move->forwardmove = MSG_ReadShort (msg_read);
	if (bits & CM_SIDE)
		move->sidemove = MSG_ReadShort (msg_read);
	if (bits & CM_UP)
		move->upmove = MSG_ReadShort (msg_read);

// read buttons
	if (bits & CM_BUTTONS)
		move->buttons = MSG_ReadByte (msg_read);

	if (bits & CM_IMPULSE)
		move->impulse = MSG_ReadByte (msg_read);

// read time to run command
   move->msec = MSG_ReadByte (msg_read);

// read the light level
#ifdef SIN
   {
   int templight;

	templight = MSG_ReadByte (msg_read);
	move->lightlevel = (templight & 0xf)<<3;
   move->debris = templight >> 4;
   }
#else
	move->lightlevel = MSG_ReadByte (msg_read);
#endif
}


void MSG_ReadData (sizebuf_t *msg_read, void *data, int len)
{
	int		i;

	for (i=0 ; i<len ; i++)
		((byte *)data)[i] = MSG_ReadByte (msg_read);
}


//===========================================================================

void SZ_Init (sizebuf_t *buf, byte *data, int length)
{
	FastMemSet (buf, 0, sizeof(*buf));
	buf->data = data;
	buf->maxsize = length;
}

void SZ_Clear (sizebuf_t *buf)
{
	buf->cursize = 0;
	buf->overflowed = FALSE;
}

void *SZ_GetSpace (sizebuf_t *buf, int length)
{
	void	*data;

	if (buf->cursize + length > buf->maxsize)
	{
		if (!buf->allowoverflow)
			Com_Error (ERR_FATAL, "SZ_GetSpace: overflow without allowoverflow set");

		if (length > buf->maxsize)
			Com_Error (ERR_FATAL, "SZ_GetSpace: %i is > full buffer size", length);

		Com_Printf ("SZ_GetSpace: overflow\n");
		SZ_Clear (buf);
		buf->overflowed = TRUE;
	}

	data = buf->data + buf->cursize;
	buf->cursize += length;

	return data;
}

void SZ_Write (sizebuf_t *buf, const void *data, int length)
{
	FastMemCpy (SZ_GetSpace(buf,length),data,length);
}

void SZ_Print (sizebuf_t *buf, const char *data)
{
	int		len;

	len = Burger::StringLength(data)+1;

	if (buf->cursize)
	{
		if (buf->data[buf->cursize-1])
			FastMemCpy((byte *)SZ_GetSpace(buf, len),data,len); // no trailing 0
		else
			FastMemCpy((byte *)SZ_GetSpace(buf, len-1)-1,data,len); // write over trailing 0
	}
	else
		FastMemCpy((byte *)SZ_GetSpace(buf, len),data,len);
}


//============================================================================



//============================================================================


/*
================
COM_CheckParm

Returns the position (1 to argc-1) in the program's argument list
where the given parameter apears, or 0 if not present
================
*/
int COM_CheckParm (char *parm)
{
	int		i;

	for (i=1 ; i<com_argc ; i++)
	{
		if (!strcmp (parm,com_argv[i]))
			return i;
	}

	return 0;
}

int COM_Argc (void)
{
	return com_argc;
}

char *COM_Argv (int arg)
{
	if (arg < 0 || arg >= com_argc || !com_argv[arg])
		return "";
	return com_argv[arg];
}

void COM_ClearArgv (int arg)
{
	if (arg < 0 || arg >= com_argc || !com_argv[arg])
		return;
	com_argv[arg] = "";
}


/*
================
COM_InitArgv
================
*/
void COM_InitArgv (int argc, char **argv)
{
	int		i;

	if (argc > MAX_NUM_ARGVS)
		Com_Error (ERR_FATAL, "argc > MAX_NUM_ARGVS");
	com_argc = argc;
	for (i=0 ; i<argc ; i++)
	{
		if (!argv[i] || Burger::StringLength(argv[i]) >= MAX_TOKEN_CHARS )
			com_argv[i] = "";
		else
			com_argv[i] = argv[i];
	}
}

/*
================
COM_AddParm

Adds the given string at the end of the current argument list
================
*/
void COM_AddParm (char *parm)
{
	if (com_argc == MAX_NUM_ARGVS)
		Com_Error (ERR_FATAL, "COM_AddParm: MAX_NUM)ARGS");
	com_argv[com_argc++] = parm;
}


void Info_Print (char *s)
{
	char	key[512];
	char	value[512];
	char	*o;
	int		l;

	if (*s == '\\')
		s++;
	while (*s)
	{
		o = key;
		while (*s && *s != '\\')
			*o++ = *s++;

		l = o - key;
		if (l < 20)
		{
			FastMemSet (o, ' ', 20-l);
			key[20] = 0;
		}
		else
			*o = 0;
		Com_Printf ("%s", key);

		if (!*s)
		{
			Com_Printf ("MISSING VALUE\n");
			return;
		}

		o = value;
		s++;
		while (*s && *s != '\\')
			*o++ = *s++;
		*o = 0;

		if (*s)
			s++;
		Com_Printf ("%s\n", value);
	}
}


/*
==============================================================================

						ZONE MEMORY ALLOCATION

just cleared malloc with counters now...

==============================================================================
*/

#define	Z_MAGIC		0x1d1d

typedef struct zhead_s {
	struct zhead_s	*prev, *next;
	short	magic;
	short	tag;			// for group free
	int		size;
} zhead_t;

zhead_t z_chain = {&z_chain,&z_chain,Z_MAGIC,0,0};
int z_count, z_bytes;

/*
========================
Z_Free
========================
*/
void Z_Free (void *ptr)
{
	zhead_t	*z;

	z = ((zhead_t *)ptr) - 1;

#ifdef SIN
   assert(z->magic == Z_MAGIC);
#endif
	if (z->magic != Z_MAGIC) {
		Com_Error (ERR_FATAL, "Z_Free: bad magic");
	}
	z->prev->next = z->next;
	z->next->prev = z->prev;
	z_count--;
	z_bytes -= z->size;
	Free(z);
}


/*
========================
Z_Stats_f
========================
*/
static void Z_Stats_f (void)
{
#ifdef SIN
	if ( developer->value ) {
		int limit;
		zhead_t	*z, *next;
	
		if ( Cmd_Argc() > 1 ) {
			limit = atoi( Cmd_Argv( 1 ) );
		} else {
			limit = 0;
		}
		for (z=z_chain.next ; z != &z_chain ; z=next) {
			next = z->next;
			if ( z->size > limit ) {
				Com_Printf("Tag:%8d Size:%8d\n",z->tag,z->size);
			}
		}
	}
#endif
	Com_Printf ("%i bytes in %i blocks\n", z_bytes, z_count);
}

#if defined(_DEBUG)
extern void Z_Stats_f2(void);
void Z_Stats_f2(void)
{
	zhead_t	*z, *next;
	for (z=z_chain.next ; z != &z_chain ; z=next) {
		next = z->next;
		DebugMessage("Tag:%8d Size:%8d\n",z->tag,z->size);
	}
	DebugMessage("%i bytes in %i blocks\n", z_bytes, z_count);
}
#endif

/*
========================
Z_FreeTags
========================
*/
void Z_FreeTags (int tag)
{
	zhead_t	*z, *next;
	for (z=z_chain.next ; z != &z_chain ; z=next) {
		next = z->next;
		if (z->tag == tag) {
			Z_Free ((void *)(z+1));
		}
	}
}

/*
========================
Z_TagMalloc
========================
*/
void *Z_TagMalloc(int size, int tag)
{
	zhead_t	*z;

	size = size + sizeof(zhead_t) + 8;
	z = (zhead_t *)AllocClear(size);
	if (!z) {
		Com_Error (ERR_FATAL, "Z_Malloc: failed on allocation of %i bytes",size);
	}
	z_count++;
	z_bytes += size;
	z->magic = Z_MAGIC;
	z->tag = tag;
	z->size = size;

	z->next = z_chain.next;
	z->prev = &z_chain;
	z_chain.next->prev = z;
	z_chain.next = z;

	return (void *)(z+1);
}

/*
========================
Z_Malloc
========================
*/
void *Z_Malloc(int size)
{
	return Z_TagMalloc(size, 0);
}


static byte chktbl[1024] = {
0x84, 0x47, 0x51, 0xc1, 0x93, 0x22, 0x21, 0x24, 0x2f, 0x66, 0x60, 0x4d, 0xb0, 0x7c, 0xda,
0x88, 0x54, 0x15, 0x2b, 0xc6, 0x6c, 0x89, 0xc5, 0x9d, 0x48, 0xee, 0xe6, 0x8a, 0xb5, 0xf4,
0xcb, 0xfb, 0xf1, 0x0c, 0x2e, 0xa0, 0xd7, 0xc9, 0x1f, 0xd6, 0x06, 0x9a, 0x09, 0x41, 0x54,
0x67, 0x46, 0xc7, 0x74, 0xe3, 0xc8, 0xb6, 0x5d, 0xa6, 0x36, 0xc4, 0xab, 0x2c, 0x7e, 0x85,
0xa8, 0xa4, 0xa6, 0x4d, 0x96, 0x19, 0x19, 0x9a, 0xcc, 0xd8, 0xac, 0x39, 0x5e, 0x3c, 0xf2,
0xf5, 0x5a, 0x72, 0xe5, 0xa9, 0xd1, 0xb3, 0x23, 0x82, 0x6f, 0x29, 0xcb, 0xd1, 0xcc, 0x71,
0xfb, 0xea, 0x92, 0xeb, 0x1c, 0xca, 0x4c, 0x70, 0xfe, 0x4d, 0xc9, 0x67, 0x43, 0x47, 0x94,
0xb9, 0x47, 0xbc, 0x3f, 0x01, 0xab, 0x7b, 0xa6, 0xe2, 0x76, 0xef, 0x5a, 0x7a, 0x29, 0x0b,
0x51, 0x54, 0x67, 0xd8, 0x1c, 0x14, 0x3e, 0x29, 0xec, 0xe9, 0x2d, 0x48, 0x67, 0xff, 0xed,
0x54, 0x4f, 0x48, 0xc0, 0xaa, 0x61, 0xf7, 0x78, 0x12, 0x03, 0x7a, 0x9e, 0x8b, 0xcf, 0x83,
0x7b, 0xae, 0xca, 0x7b, 0xd9, 0xe9, 0x53, 0x2a, 0xeb, 0xd2, 0xd8, 0xcd, 0xa3, 0x10, 0x25,
0x78, 0x5a, 0xb5, 0x23, 0x06, 0x93, 0xb7, 0x84, 0xd2, 0xbd, 0x96, 0x75, 0xa5, 0x5e, 0xcf,
0x4e, 0xe9, 0x50, 0xa1, 0xe6, 0x9d, 0xb1, 0xe3, 0x85, 0x66, 0x28, 0x4e, 0x43, 0xdc, 0x6e,
0xbb, 0x33, 0x9e, 0xf3, 0x0d, 0x00, 0xc1, 0xcf, 0x67, 0x34, 0x06, 0x7c, 0x71, 0xe3, 0x63,
0xb7, 0xb7, 0xdf, 0x92, 0xc4, 0xc2, 0x25, 0x5c, 0xff, 0xc3, 0x6e, 0xfc, 0xaa, 0x1e, 0x2a,
0x48, 0x11, 0x1c, 0x36, 0x68, 0x78, 0x86, 0x79, 0x30, 0xc3, 0xd6, 0xde, 0xbc, 0x3a, 0x2a,
0x6d, 0x1e, 0x46, 0xdd, 0xe0, 0x80, 0x1e, 0x44, 0x3b, 0x6f, 0xaf, 0x31, 0xda, 0xa2, 0xbd,
0x77, 0x06, 0x56, 0xc0, 0xb7, 0x92, 0x4b, 0x37, 0xc0, 0xfc, 0xc2, 0xd5, 0xfb, 0xa8, 0xda,
0xf5, 0x57, 0xa8, 0x18, 0xc0, 0xdf, 0xe7, 0xaa, 0x2a, 0xe0, 0x7c, 0x6f, 0x77, 0xb1, 0x26,
0xba, 0xf9, 0x2e, 0x1d, 0x16, 0xcb, 0xb8, 0xa2, 0x44, 0xd5, 0x2f, 0x1a, 0x79, 0x74, 0x87,
0x4b, 0x00, 0xc9, 0x4a, 0x3a, 0x65, 0x8f, 0xe6, 0x5d, 0xe5, 0x0a, 0x77, 0xd8, 0x1a, 0x14,
0x41, 0x75, 0xb1, 0xe2, 0x50, 0x2c, 0x93, 0x38, 0x2b, 0x6d, 0xf3, 0xf6, 0xdb, 0x1f, 0xcd,
0xff, 0x14, 0x70, 0xe7, 0x16, 0xe8, 0x3d, 0xf0, 0xe3, 0xbc, 0x5e, 0xb6, 0x3f, 0xcc, 0x81,
0x24, 0x67, 0xf3, 0x97, 0x3b, 0xfe, 0x3a, 0x96, 0x85, 0xdf, 0xe4, 0x6e, 0x3c, 0x85, 0x05,
0x0e, 0xa3, 0x2b, 0x07, 0xc8, 0xbf, 0xe5, 0x13, 0x82, 0x62, 0x08, 0x61, 0x69, 0x4b, 0x47,
0x62, 0x73, 0x44, 0x64, 0x8e, 0xe2, 0x91, 0xa6, 0x9a, 0xb7, 0xe9, 0x04, 0xb6, 0x54, 0x0c,
0xc5, 0xa9, 0x47, 0xa6, 0xc9, 0x08, 0xfe, 0x4e, 0xa6, 0xcc, 0x8a, 0x5b, 0x90, 0x6f, 0x2b,
0x3f, 0xb6, 0x0a, 0x96, 0xc0, 0x78, 0x58, 0x3c, 0x76, 0x6d, 0x94, 0x1a, 0xe4, 0x4e, 0xb8,
0x38, 0xbb, 0xf5, 0xeb, 0x29, 0xd8, 0xb0, 0xf3, 0x15, 0x1e, 0x99, 0x96, 0x3c, 0x5d, 0x63,
0xd5, 0xb1, 0xad, 0x52, 0xb8, 0x55, 0x70, 0x75, 0x3e, 0x1a, 0xd5, 0xda, 0xf6, 0x7a, 0x48,
0x7d, 0x44, 0x41, 0xf9, 0x11, 0xce, 0xd7, 0xca, 0xa5, 0x3d, 0x7a, 0x79, 0x7e, 0x7d, 0x25,
0x1b, 0x77, 0xbc, 0xf7, 0xc7, 0x0f, 0x84, 0x95, 0x10, 0x92, 0x67, 0x15, 0x11, 0x5a, 0x5e,
0x41, 0x66, 0x0f, 0x38, 0x03, 0xb2, 0xf1, 0x5d, 0xf8, 0xab, 0xc0, 0x02, 0x76, 0x84, 0x28,
0xf4, 0x9d, 0x56, 0x46, 0x60, 0x20, 0xdb, 0x68, 0xa7, 0xbb, 0xee, 0xac, 0x15, 0x01, 0x2f,
0x20, 0x09, 0xdb, 0xc0, 0x16, 0xa1, 0x89, 0xf9, 0x94, 0x59, 0x00, 0xc1, 0x76, 0xbf, 0xc1,
0x4d, 0x5d, 0x2d, 0xa9, 0x85, 0x2c, 0xd6, 0xd3, 0x14, 0xcc, 0x02, 0xc3, 0xc2, 0xfa, 0x6b,
0xb7, 0xa6, 0xef, 0xdd, 0x12, 0x26, 0xa4, 0x63, 0xe3, 0x62, 0xbd, 0x56, 0x8a, 0x52, 0x2b,
0xb9, 0xdf, 0x09, 0xbc, 0x0e, 0x97, 0xa9, 0xb0, 0x82, 0x46, 0x08, 0xd5, 0x1a, 0x8e, 0x1b,
0xa7, 0x90, 0x98, 0xb9, 0xbb, 0x3c, 0x17, 0x9a, 0xf2, 0x82, 0xba, 0x64, 0x0a, 0x7f, 0xca,
0x5a, 0x8c, 0x7c, 0xd3, 0x79, 0x09, 0x5b, 0x26, 0xbb, 0xbd, 0x25, 0xdf, 0x3d, 0x6f, 0x9a,
0x8f, 0xee, 0x21, 0x66, 0xb0, 0x8d, 0x84, 0x4c, 0x91, 0x45, 0xd4, 0x77, 0x4f, 0xb3, 0x8c,
0xbc, 0xa8, 0x99, 0xaa, 0x19, 0x53, 0x7c, 0x02, 0x87, 0xbb, 0x0b, 0x7c, 0x1a, 0x2d, 0xdf,
0x48, 0x44, 0x06, 0xd6, 0x7d, 0x0c, 0x2d, 0x35, 0x76, 0xae, 0xc4, 0x5f, 0x71, 0x85, 0x97,
0xc4, 0x3d, 0xef, 0x52, 0xbe, 0x00, 0xe4, 0xcd, 0x49, 0xd1, 0xd1, 0x1c, 0x3c, 0xd0, 0x1c,
0x42, 0xaf, 0xd4, 0xbd, 0x58, 0x34, 0x07, 0x32, 0xee, 0xb9, 0xb5, 0xea, 0xff, 0xd7, 0x8c,
0x0d, 0x2e, 0x2f, 0xaf, 0x87, 0xbb, 0xe6, 0x52, 0x71, 0x22, 0xf5, 0x25, 0x17, 0xa1, 0x82,
0x04, 0xc2, 0x4a, 0xbd, 0x57, 0xc6, 0xab, 0xc8, 0x35, 0x0c, 0x3c, 0xd9, 0xc2, 0x43, 0xdb,
0x27, 0x92, 0xcf, 0xb8, 0x25, 0x60, 0xfa, 0x21, 0x3b, 0x04, 0x52, 0xc8, 0x96, 0xba, 0x74,
0xe3, 0x67, 0x3e, 0x8e, 0x8d, 0x61, 0x90, 0x92, 0x59, 0xb6, 0x1a, 0x1c, 0x5e, 0x21, 0xc1,
0x65, 0xe5, 0xa6, 0x34, 0x05, 0x6f, 0xc5, 0x60, 0xb1, 0x83, 0xc1, 0xd5, 0xd5, 0xed, 0xd9,
0xc7, 0x11, 0x7b, 0x49, 0x7a, 0xf9, 0xf9, 0x84, 0x47, 0x9b, 0xe2, 0xa5, 0x82, 0xe0, 0xc2,
0x88, 0xd0, 0xb2, 0x58, 0x88, 0x7f, 0x45, 0x09, 0x67, 0x74, 0x61, 0xbf, 0xe6, 0x40, 0xe2,
0x9d, 0xc2, 0x47, 0x05, 0x89, 0xed, 0xcb, 0xbb, 0xb7, 0x27, 0xe7, 0xdc, 0x7a, 0xfd, 0xbf,
0xa8, 0xd0, 0xaa, 0x10, 0x39, 0x3c, 0x20, 0xf0, 0xd3, 0x6e, 0xb1, 0x72, 0xf8, 0xe6, 0x0f,
0xef, 0x37, 0xe5, 0x09, 0x33, 0x5a, 0x83, 0x43, 0x80, 0x4f, 0x65, 0x2f, 0x7c, 0x8c, 0x6a,
0xa0, 0x82, 0x0c, 0xd4, 0xd4, 0xfa, 0x81, 0x60, 0x3d, 0xdf, 0x06, 0xf1, 0x5f, 0x08, 0x0d,
0x6d, 0x43, 0xf2, 0xe3, 0x11, 0x7d, 0x80, 0x32, 0xc5, 0xfb, 0xc5, 0xd9, 0x27, 0xec, 0xc6,
0x4e, 0x65, 0x27, 0x76, 0x87, 0xa6, 0xee, 0xee, 0xd7, 0x8b, 0xd1, 0xa0, 0x5c, 0xb0, 0x42,
0x13, 0x0e, 0x95, 0x4a, 0xf2, 0x06, 0xc6, 0x43, 0x33, 0xf4, 0xc7, 0xf8, 0xe7, 0x1f, 0xdd,
0xe4, 0x46, 0x4a, 0x70, 0x39, 0x6c, 0xd0, 0xed, 0xca, 0xbe, 0x60, 0x3b, 0xd1, 0x7b, 0x57,
0x48, 0xe5, 0x3a, 0x79, 0xc1, 0x69, 0x33, 0x53, 0x1b, 0x80, 0xb8, 0x91, 0x7d, 0xb4, 0xf6,
0x17, 0x1a, 0x1d, 0x5a, 0x32, 0xd6, 0xcc, 0x71, 0x29, 0x3f, 0x28, 0xbb, 0xf3, 0x5e, 0x71,
0xb8, 0x43, 0xaf, 0xf8, 0xb9, 0x64, 0xef, 0xc4, 0xa5, 0x6c, 0x08, 0x53, 0xc7, 0x00, 0x10,
0x39, 0x4f, 0xdd, 0xe4, 0xb6, 0x19, 0x27, 0xfb, 0xb8, 0xf5, 0x32, 0x73, 0xe5, 0xcb, 0x32
};

/*
====================
COM_BlockSequenceCRCByte

For proxy protecting
====================
*/
byte  COM_BlockSequenceCRCByte (byte *base, int length, int sequence)
{
   unsigned short crc;
   byte  *p;
   byte chkb[60 + 4];

   if (sequence < 0)
      Sys_Error("sequence < 0, this shouldn't happen\n");

   p = chktbl + (sequence % (sizeof(chktbl) - 4));

   if (length > 60)
      length = 60;
   FastMemCpy (chkb, base, length);

   chkb[length] = p[0];
   chkb[length+1] = p[1];
   chkb[length+2] = p[2];
   chkb[length+3] = p[3];

   length += 4;

   crc = CRC_Block(chkb, length);

   crc &= 0xff;

   return crc;
}

//========================================================

float	frand(void)
{
	return (rand()&32767)* (1.0/32767);
}

float	crand(void)
{
	return (rand()&32767)* (2.0/32767) - 1;
}

/*
=============
Com_Error_f

Just throw a fatal error to
test error shutdown procedures
=============
*/
static void Com_Error_f (void)
{
	Com_Error (ERR_FATAL, "%s", Cmd_Argv(1));
}

static void E3_CopyProt(void)
   {
#ifdef SIN_E3DEMO

   time_t ctime;
   time_t stime;
   time_t etime;
   struct tm start = { 0, 0, 12, 20, 4, 98 }; // 12 noon may 20, 1998
   struct tm end   = { 0, 0, 12, 1, 5, 98 };  // 12 noon june 1, 1998

   stime = mktime(&start);
   etime = mktime(&end);
   time(&ctime);

   if ((ctime < stime) || (ctime > etime))
      {
      goto error;
      }

   if (copy_prot = Cvar_Get("sv_gamemode", 0,0))
      {
      if (Burger::StringLength(copy_prot->string) == 5)
         {
         if (copy_prot->string[0] != 'd')
            goto error;
         if (copy_prot->string[1] != 'o')
            goto error;
         if (copy_prot->string[2] != 'm')
            goto error;
         if (copy_prot->string[3] != '6')
            goto error;
         if (copy_prot->string[4] != '9')
            goto error;

         Cvar_SetValue("sv_gamemode", 0);
         return;
         }
      else
         goto error;
      }
error:
      Sys_Error ("UPD System failed to initialize\nCheck driver disk\n");

#endif
   }

/*
=================
Qcommon_Init
=================
*/
void Qcommon_Init (int argc, char **argv)
{
	const char	*s;
	char	path[MAX_QPATH];

	if (setjmp (abortframe) )
		Sys_Error ("Error during initialization");

	z_chain.next = z_chain.prev = &z_chain;

	// prepare enough of the subsystems to handle
	// cvar and command buffer management
	COM_InitArgv (argc, argv);

	Swap_Init ();
	Cbuf_Init ();

	Cmd_Init ();
	Cvar_Init ();
#ifdef SIN
	developer = Cvar_Get ("developer", "0", 0);
#endif

	Key_Init ();

	// we need to add the early commands twice, because
	// a basedir or cddir needs to be set before execing
	// config files, but we want other parms to override
	// the settings of the config files
	Cbuf_AddEarlyCommands (FALSE);
	Cbuf_Execute ();

	FS_InitFilesystem ();
	
	Cbuf_AddText ("exec default.cfg\n");
#ifdef SIN
	Com_sprintf (path, sizeof(path),"%s/config.cfg",FS_RelativePlayerdir());
	Cbuf_AddText (va("exec %s\n", path) );
#else
	Cbuf_AddText ("exec config.cfg\n");
#endif

	Cbuf_AddEarlyCommands (TRUE);
	Cbuf_Execute ();

	E3_CopyProt();

	//
	// init commands and vars
	//
	Cmd_AddCommand ("z_stats", Z_Stats_f);
	Cmd_AddCommand ("error", Com_Error_f);

	host_speeds = Cvar_Get ("host_speeds", "0", 0);
	log_stats = Cvar_Get ("log_stats", "0", 0);
	developer = Cvar_Get ("developer", "0", 0);
	timescale = Cvar_Get ("timescale", "1", 0);
	fixedtime = Cvar_Get ("fixedtime", "0", 0);
	logfile_active = Cvar_Get ("logfile", "0", 0);
	showtrace = Cvar_Get ("showtrace", "0", 0);
#ifdef DEDICATED_ONLY
	dedicated = Cvar_Get ("dedicated", "1", CVAR_NOSET);
#else
	dedicated = Cvar_Get ("dedicated", "0", CVAR_NOSET);
#endif
#ifdef SIN
	// 2015 code 
	timescalenopitch = Cvar_Get ("timescalenopitch", "0", 0);
	fps = Cvar_Get ("fps", "0", 0);
	console_debug = Cvar_Get("console_debug", "0", 0);
	surface_debug = Cvar_Get("surface_debug", "0", 0);
#ifdef SIN_DEMO
	sindemo = Cvar_Get("sindemo", "1", CVAR_NOSET);
#else
	sindemo = Cvar_Get("sindemo", "0", CVAR_NOSET);
#endif
#endif
	s = va("%4.2f %s %s %s", VERSION, CPUSTRING, __DATE__, BUILDSTRING);
	Cvar_Get ("version", s, CVAR_SERVERINFO|CVAR_NOSET);
#ifdef SIN
	s = va("%4.2f", VERSION);
	Cvar_Get ("version_only", s, CVAR_NOSET);

	// 2015 code 
	game2015 = Cvar_Get ("game2015", "0", CVAR_NOSET);
#endif

	if (dedicated->value) {
		Cmd_AddCommand ("quit", Com_Quit);
	}
	Sys_Init ();

	NET_Init ();
	Netchan_Init ();

	SV_Init ();
	CL_Init ();

	// add + commands from command line
	if (!Cbuf_AddLateCommands ()) {		// if the user didn't give any commands, run default action
		if (!dedicated->value) {
			Cbuf_AddText ("d1");
		} else {
			Cbuf_AddText ("dedicated_start");
		}
		Cbuf_Execute ();
	} else {	// the user asked for something explicit
		// so drop the loading plaque
		SCR_EndLoadingPlaque ();
	}
	Com_Printf ("====== " GAME_NAME " Initialized ======\n\n");
}

/*
=================
Qcommon_Frame
=================
*/
void Qcommon_Frame (int msec)
{
#ifdef SIN
	static int fpsindex = 0;
	static double fpstimes[16] = {0};
	static double fpstotal = 0;
	static double fpslasttime = 0;
	static int ref_fpsindex = 0;
	static double ref_fpstimes[16] = {0};
	static double ref_fpstotal = 0;
	static double ref_fpslasttime = 0;
	
	const char	*s;
#else
	char *s;
#endif
	int time_before, time_between, time_after;

	if (setjmp(abortframe)) {
		return;			// an ERR_DROP was thrown
	}
	if (log_stats->modified) {
		log_stats->modified = FALSE;
		if ( log_stats->value ) {
			if ( log_stats_file ) {
				fclose( log_stats_file );
				log_stats_file = 0;
			}
			log_stats_file = QuakeOpenAFile( "stats.log", "w" );
			if ( log_stats_file ) {
				fprintf( log_stats_file, "entities,dlights,parts,frame time\n" );
			}
		} else {
			if ( log_stats_file ) {
				fclose( log_stats_file );
				log_stats_file = 0;
			}
		}
	}

	if (fixedtime->value) {
		msec = fixedtime->value;
	} else if (timescale->value) {
		msec *= timescale->value;
		if (msec < 1) {
			msec = 1;
		}
	}
	
	if (showtrace->value) {
		extern	int c_traces, c_brush_traces;
		extern	int	c_pointcontents;

#ifdef SIN
		Com_Printf ("%4i traces  %4i brush traces %4i points\n", c_traces, c_brush_traces, c_pointcontents);
#else
		Com_Printf ("%4i traces  %4i points\n", c_traces, c_pointcontents);
#endif
		c_traces = 0;
		c_brush_traces = 0;
		c_pointcontents = 0;
	}

	do {
		s = Sys_ConsoleInput ();
		if (s) {
			Cbuf_AddText (va("%s\n",s));
		}
	} while (s);
	Cbuf_Execute();

	if (host_speeds->value) {
		time_before = Sys_Milliseconds ();
	}
	SV_Frame (msec);

	if (host_speeds->value) {
		time_between = Sys_Milliseconds ();
	}
	CL_Frame (msec);

	if (host_speeds->value) {
		time_after = Sys_Milliseconds ();
	}

	if (host_speeds->value) {
		int			all, sv, gm, cl, rf;

		all = time_after - time_before;
		sv = time_between - time_before;
		cl = time_after - time_between;
		gm = time_after_game - time_before_game;
		rf = time_after_ref - time_before_ref;
		sv -= gm;
		cl -= rf;
		Com_Printf ("all:%3i sv:%3i gm:%3i cl:%3i rf:%3i\n",all, sv, gm, cl, rf);
	}
#ifdef SIN
    if (fps->value) {
		char text[ 255 ];
		double fpstime;
		fpstime   = Sys_Milliseconds() / 1000.0f;
		fpstotal -= fpstimes[ fpsindex ];
		fpstimes[ fpsindex ] = fpstime - fpslasttime;
		fpstotal += fpstimes[ fpsindex ];
		fpsindex  = ( fpsindex + 1 ) & 15;
		fpslasttime = fpstime;
        ref_fpstotal -= ref_fpstimes[ ref_fpsindex ];
        ref_fpstimes[ ref_fpsindex ] = (time_after_ref - time_before_ref) / 1000.0f;
		ref_fpstotal += ref_fpstimes[ ref_fpsindex ];
		ref_fpsindex  = ( ref_fpsindex + 1 ) & 15;
        sprintf( text, "TOT FPS: %4.1f\nREF FPS: %4.1f", 16.0f / fpstotal, 16.0f / ref_fpstotal );
        SCR_CenterPrint(text);
	}
#endif
}

#ifdef SIN

void (*CmdUpdateStatus) (int count, int total, const char *str)=NULL;
void (*CmdUpdateStatusString) (const char *str)=NULL;
void (*CmdIncrementStatusCount) (int i)=NULL;

void Com_RegisterUpdateFunctions( void (*fcn1) (int, int, const char *), 
                                  void (*fcn2) (const char *), 
                                  void (*fcn3) (int i) )
   {
   CmdUpdateStatus         = fcn1;
   CmdUpdateStatusString   = fcn2;
   CmdIncrementStatusCount = fcn3;
   }

void Com_UpdateStatusString( const char *str )
   {
   if ( developer->value )
      {
//      OutputDebugString( va("%s\n",str) );
      return;
      }

   if ( CmdUpdateStatusString )
      {
      CmdUpdateStatusString( str );
      }
   }

void Com_UpdateStatus( int count, int total, const char *str )
   {
   if ( developer->value )
      {
//      OutputDebugString( va("%s\n",str) );
      return;
      }

   if ( CmdUpdateStatus )
      {
      CmdUpdateStatus( count, total, str );
      }
   }

void Com_IncrementStatusCount( int i )
   {
   if ( developer->value )
      return;

   if ( CmdIncrementStatusCount )
      {
      CmdIncrementStatusCount( i );
      }
   }

#endif
/*
=================
Qcommon_Shutdown
=================
*/
void Qcommon_Shutdown (void)
{
}