//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/cvar.c                                                 $
// $Revision:: 3                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 2/5/02 7:39 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// dynamic variable tracking
//

#include "qcommon.h"
#include "console3d.h"

extern cvar_t *Cvar_FindVar (const char *var_name);

cvar_t	*cvar_vars;		/* Root pointer to the linked list */

/*
============
Cvar_InfoValidate
============
*/
static qboolean Cvar_InfoValidate (const char *s)
{
   if (strstr (s, "\\"))
      return FALSE;
   if (strstr (s, "\""))
      return FALSE;
   if (strstr (s, ";"))
      return FALSE;
   return TRUE;
}

/*
============
Cvar_FindVar
============
*/
cvar_t *Cvar_FindVar (const char *var_name)
{
	cvar_t	*var;

	for (var=cvar_vars ; var ; var=var->next)
		if (!strcmp (var_name, var->CVarName))
			return var;

	return NULL;
}

/*
============
Cvar_VariableValue
============
*/
float Cvar_VariableValue (const char *var_name)
{
	cvar_t	*var;

	var = Cvar_FindVar (var_name);
	if (!var)
		return 0;
	return Sinatof (var->CVarString);
}


/*
============
Cvar_VariableString
============
*/
const char *Cvar_VariableString (const char *var_name)
{
	cvar_t *var;

	var = Cvar_FindVar (var_name);
	if (!var)
		return "";
	return var->CVarString;
}


/*
============
Cvar_CompleteVariable
============
*/
const char *Cvar_CompleteVariable (const char *partial)
{
	cvar_t		*cvar;
	int			len;

	len = Burger::StringLength(partial);

	if (!len)
		return NULL;

	// check exact match
	for (cvar=cvar_vars ; cvar ; cvar=cvar->next)
		if (!strcmp (partial,cvar->CVarName))
			return cvar->CVarName;

	// check partial match
	for (cvar=cvar_vars ; cvar ; cvar=cvar->next)
		if (!strncmp (partial,cvar->CVarName, len))
			return cvar->CVarName;

	return NULL;
}


/*
============
Cvar_Get

If the variable allready exists, the value will not be set
The flags will be or'ed in if the variable exists.
============
*/
cvar_t *Cvar_Get (const char *var_name, const char *var_value, int flags)
{
	cvar_t	*var;

   if (flags & (CVAR_USERINFO | CVAR_SERVERINFO))
   {
      if (!Cvar_InfoValidate (var_name))
      {
         Com_Printf("invalid info cvar name\n");
         return NULL;
      }
   }

   var = Cvar_FindVar (var_name);
	if (var)
	{
		var->flags |= flags;
		return var;
	}

	if (!var_value)
		return NULL;

   if (flags & (CVAR_USERINFO | CVAR_SERVERINFO))
   {
      if (!Cvar_InfoValidate (var_value))
      {
         Com_Printf("invalid info cvar value\n");
         return NULL;
      }
   }

   var = (cvar_t *)AllocClear(sizeof(*var));
	var->CVarName = Burger::StringDuplicate(var_name);
	var->CVarString = Burger::StringDuplicate(var_value);
	var->modified = TRUE;
	var->value = Sinatof (var->CVarString);

	// link the variable in
	var->next = cvar_vars;
	cvar_vars = var;

	var->flags = flags;

	return var;
}

/*
============
Cvar_Set2
============
*/
static cvar_t *Cvar_Set2 (const char *var_name, const char *value, qboolean force)
{
	cvar_t	*var;
	
	var = Cvar_FindVar (var_name);
	if (!var) {	// create it
		return Cvar_Get (var_name, value, 0);
	}
	
	if (var->flags & (CVAR_USERINFO | CVAR_SERVERINFO)) {
		if (!Cvar_InfoValidate (value)) {
			Com_Printf("invalid info cvar value\n");
			return var;
		}
	}

	if (!force) {
		if (var->flags & CVAR_NOSET) {
			Com_Printf ("%s is write protected.\n", var_name);
			return var;
		}
	
		if (var->flags & CVAR_LATCH) {
			if (var->CVarLatchedString) {
				if (strcmp(value, var->CVarLatchedString) == 0)
					return var;
				Free(var->CVarLatchedString);
				var->CVarLatchedString = 0;
			} else {
				if (strcmp(value, var->CVarString) == 0)
					return var;
			}
	
			if (Com_ServerState()) {
				Com_Printf ("%s will be changed for next game.\n", var_name);
				var->CVarLatchedString = Burger::StringDuplicate(value);
			} else {
				Free(var->CVarString);
				var->CVarString = Burger::StringDuplicate(value);
				var->value = Sinatof (var->CVarString);
				if (!strcmp(var->CVarName, "game")) {
					FS_SetGamedir (var->CVarString);
					// 2015 code 
					if (!strcmp(var->CVarString, "2015"))
						Cvar_FullSet ("game2015", "1", CVAR_SERVERINFO|CVAR_NOSET);
					else
						Cvar_FullSet ("game2015", "0", CVAR_SERVERINFO|CVAR_NOSET);
					FS_ExecAutoexec ();
				}
			}
			return var;
		}
	} else {
		Free(var->CVarLatchedString);
		var->CVarLatchedString = NULL;
	}
	
	if (!strcmp(value, var->CVarString))
		return var;		// not changed
	
	var->modified = TRUE;
	
	if (var->flags & CVAR_USERINFO)
		userinfo_modified = TRUE;	// transmit at next oportunity
	
	Free(var->CVarString);	// free the old value string
	
	var->CVarString = Burger::StringDuplicate(value);
	var->value = Sinatof (var->CVarString);
	
	return var;
}

/*
============
Cvar_ForceSet
============
*/
cvar_t *Cvar_ForceSet (const char *var_name, const char *value)
{
	return Cvar_Set2 (var_name, value, TRUE);
}

/*
============
Cvar_Set
============
*/
cvar_t *Cvar_Set (const char *var_name, const char *value)
{
	return Cvar_Set2 (var_name, value, FALSE);
}

/*
============
Cvar_FullSet
============
*/
cvar_t *Cvar_FullSet (const char *var_name, const char *value, int flags)
{
	cvar_t	*var;

	var = Cvar_FindVar (var_name);
	if (!var)
	{	// create it
      return Cvar_Get (var_name, value, flags);
	}

	var->modified = TRUE;

	if (var->flags & CVAR_USERINFO)
		userinfo_modified = TRUE;	// transmit at next oportunity

	Free (var->CVarString);	// free the old value string

	var->CVarString = Burger::StringDuplicate(value);
	var->value = Sinatof (var->CVarString);
	var->flags = flags;

	return var;
}

/*
============
Cvar_SetValue
============
*/
void Cvar_SetValue (const char *var_name, float value)
{
	char	val[32];

	if (value == (int)value)
		Com_sprintf (val, sizeof(val), "%i",(int)value);
	else
		Com_sprintf (val, sizeof(val), "%f",value);
	Cvar_Set (var_name, val);
}


/*
============
Cvar_GetLatchedVars

Any variables with latched values will now be updated
============
*/
void Cvar_GetLatchedVars (void)
{
	cvar_t	*var;

	for (var = cvar_vars ; var ; var = var->next)
	{
		if (!var->CVarLatchedString)
			continue;
		Free (var->CVarString);
		var->CVarString = var->CVarLatchedString;
		var->CVarLatchedString = 0;
      var->value = Sinatof(var->CVarString);
      if (!strcmp(var->CVarName, "game"))
      {
         FS_SetGamedir (var->CVarString);
         FS_ExecAutoexec ();
      }
	}
}

/*
============
Cvar_Command

Handles variable inspection and changing from the console
============
*/
qboolean Cvar_Command (void)
{
	cvar_t			*v;

// check variables
	v = Cvar_FindVar (Cmd_Argv(0));
	if (!v)
		return FALSE;

// perform a variable print or set
	if (Cmd_Argc() == 1)
	{
		Com_Printf ("\"%s\" is \"%s\"\n", v->CVarName, v->CVarString);
		return TRUE;
	}

	Cvar_Set (v->CVarName, Cmd_Argv(1));
	return TRUE;
}


#ifdef SIN
/*
============
Cvar_Negate_f
============
*/
static void Cvar_Negate_f (void)
   {
   int c;

	c = Cmd_Argc();
	if (c != 2)
	   {
		Com_Printf ("usage: negate <variable>\n");
		return;
	   }

      Cvar_SetValue(Cmd_Argv(1),-Cvar_VariableValue(Cmd_Argv(1)));
   }


static void Cvar_BitSet_f (void)
   {
   int c;
   int bits = 0;
   int bitset = 0;

	c = Cmd_Argc();
	if (c != 3)
	   {
		Com_Printf ("usage: bitset <bitnum> <variable>\n");
		return;
	   }

   bits = Cvar_VariableValue( Cmd_Argv(2) );
   bitset = (1<<(atoi(Cmd_Argv(1))));
   bits |= bitset;
   Cvar_SetValue(Cmd_Argv(2),bits);
   }

static void Cvar_BitClear_f (void)
   {
   int c;
   int bits = 0;
   int bitset = 0;

	c = Cmd_Argc();
	if (c != 3)
	   {
		Com_Printf ("usage: bitclear <bitnum> <variable>\n");
		return;
	   }

   bits = Cvar_VariableValue( Cmd_Argv(2) );
   bitset = (1<<(atoi(Cmd_Argv(1))));
   bits &= ~bitset;
   Cvar_SetValue(Cmd_Argv(2),bits);
   }

// 2015 code 
// allows displaying a help text box when an item is changed
static void Cvar_DisplayText_f(void)
{
   int c, i, displaytime;
   char layout[2048];


   c = Cmd_Argc();
   if (c < 4)
   {
      Com_Printf ("usage: displaytext <time> <layout string> <text...>\n");

      for(i = 0; i < c; i++)
         Com_Printf ("       arg %i: %s\n", i, Cmd_Argv(i));

      return;
   }

   // get the text display time
   displaytime = atoi(Cmd_Argv(1));

   // add the format string to the layout string
   sprintf(layout, "%s \"", Cmd_Argv(2));

   // add all the lines of text
   i = 3;
   while(i < c)
   {
      // we're doing multiple lines of text here
      if(i > 3)
         strcat(layout, "\n");

      // tack this line of text only the layout
      strcat(layout, Cmd_Argv(i));

      i++;
   }

   // close off the end of the layout string
   strcat(layout, "\"");

   CONSOLE_ScreenPrint(layout, displaytime);
}

#endif

/*
============
Cvar_Set_f

Allows setting and defining of arbitrary cvars from console
============
*/
static void Cvar_Set_f (void)
{
	int		c;
	int		flags;

	c = Cmd_Argc();
	if (c != 3 && c != 4)
	{
		Com_Printf ("usage: set <variable> <value> [u / s]\n");
		return;
	}

	if (c == 4)
	{
		if (!strcmp(Cmd_Argv(3), "u"))
			flags = CVAR_USERINFO;
		else if (!strcmp(Cmd_Argv(3), "s"))
			flags = CVAR_SERVERINFO;
		else
		{
			Com_Printf ("flags can only be 'u' or 's'\n");
			return;
		}
		Cvar_FullSet (Cmd_Argv(1), Cmd_Argv(2), flags);
	}
	else
		Cvar_Set (Cmd_Argv(1), Cmd_Argv(2));
}


/*
============
Cvar_WriteVariables

Appends lines containing "set variable value" for all variables
with the archive flag set to TRUE.
============
*/
void Cvar_WriteVariables (char *path)
{
	cvar_t	*var;
	char	buffer[1024];
	FILE	*f;

	f = QuakeOpenAFile (path, "a");
	for (var = cvar_vars ; var ; var = var->next)
	{
		if (var->flags & CVAR_ARCHIVE)
		{
			Com_sprintf (buffer, sizeof(buffer), "set %s \"%s\"\n", var->CVarName, var->CVarString);
			fprintf (f, "%s", buffer);
		}
	}
	fclose (f);
}

/*
============
Cvar_List_f

============
*/
static void Cvar_List_f (void)
{
	cvar_t	*var;
	int		i;

#ifdef SIN
   int c,result,num=0;
   const char *mask;
	const char *pdest;

  	c = Cmd_Argc();
	if (c == 2)
	{
      mask = Cmd_Argv(1);
	}
#endif

	i = 0;
	for (var = cvar_vars ; var ; var = var->next, i++)
	{

#ifdef SIN
      if (c != 2)
         result = 1;
      else
         {
         result = 0;
         pdest = strstr( var->CVarName, mask );
         if (pdest)
            {
            result = pdest - var->CVarName + 1;
            }
         }


      if (result == 1)
         {
         num++;
#endif
         if (var->flags & CVAR_ARCHIVE)
            Com_Printf ("*");
         else
            Com_Printf (" ");
         if (var->flags & CVAR_USERINFO)
            Com_Printf ("U");
         else
            Com_Printf (" ");
         if (var->flags & CVAR_SERVERINFO)
            Com_Printf ("S");
         else
            Com_Printf (" ");
         if (var->flags & CVAR_NOSET)
            Com_Printf ("-");
         else if (var->flags & CVAR_LATCH)
            Com_Printf ("L");
         else
            Com_Printf (" ");
         Com_Printf (" %s \"%s\"\n", var->CVarName, var->CVarString);
#ifdef SIN
         }
#endif
	}
#ifdef SIN
   Com_Printf ("Printed %i of %i total cvars\n", num, i);
#else
	Com_Printf ("%i cvars\n", i);
#endif
}


qboolean userinfo_modified;


static char	*Cvar_BitInfo (int bit)
{
	static char	info[MAX_INFO_STRING];
	cvar_t	*var;

	info[0] = 0;

	for (var = cvar_vars ; var ; var = var->next)
	{
		if (var->flags & bit)
			Info_SetValueForKey (info, var->CVarName, var->CVarString);
	}
	return info;
}

// returns an info string containing all the CVAR_USERINFO cvars
char	*Cvar_Userinfo (void)
{
	return Cvar_BitInfo (CVAR_USERINFO);
}

// returns an info string containing all the CVAR_SERVERINFO cvars
char	*Cvar_Serverinfo (void)
{
	return Cvar_BitInfo (CVAR_SERVERINFO);
}

/*
============
Cvar_Init

Reads in all archived cvars
============
*/
void Cvar_Init (void)
{
	Cmd_AddCommand ("set", Cvar_Set_f);
	Cmd_AddCommand ("cvarlist", Cvar_List_f);
#ifdef SIN
	Cmd_AddCommand ("negate", Cvar_Negate_f);
	Cmd_AddCommand ("bitset", Cvar_BitSet_f);
	Cmd_AddCommand ("bitclear", Cvar_BitClear_f);
	// 2015 code 
	Cmd_AddCommand ("displaytext", Cvar_DisplayText_f);
#endif
}


/*
============
Cvar_Destroy

Dispose of all the cvars
============
*/
void Cvar_Destroy(void)
{
	cvar_t *var;

	var=cvar_vars;
	if (var) {
		do {
			cvar_t *next;
			next = var->next;
			Free(var->CVarName);
			Free(var->CVarString);
			Free(var->CVarLatchedString);
			Free(var);
			var = next;
		} while (var);
	}
	cvar_vars = 0;
}
