//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/files.c                                                $
// $Revision:: 4                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 2/5/02 7:39 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
//

#include "qcommon.h"
#include "cdaudio.h"

#if ( defined( SIN ) && defined( BURGER_WIN32 ) )
#include <windows.h>
#endif

// define this to dissalow any data but the demo pak file
//define	NO_ADDONS

extern int chdir(const char *path);
extern char *getcwd(char *buf, int size);

// if a packfile directory differs from this, it is assumed to be hacked
#define	PAK0_CHECKSUM	0x40e614e0

/*
=============================================================================

QUAKE FILESYSTEM

=============================================================================
*/

#ifdef SIN
#define	BASEDIRNAME	"base"
#else
#define	BASEDIRNAME	"baseq2"
#endif

//
// in memory
//

#define MAX_READ 0x10000     // read in blocks of 64k

typedef struct packfile_t {
	char *filename;					/* Pointer to the filename */
	Word32 filepos;				/* Offset into the pack file */
	Word32 filelen;				/* Length of the file chunk */
} packfile_t;

typedef struct pack_t {
	char filename[MAX_OSPATH];		/* Name of the pak file */
	FILE *handle;					/* Open file reference */
	Word numfiles;					/* Number of entries in the list */
	packfile_t *files;				/* Sorted list (For binary search) */
} pack_t;

struct FileRef_t {
	FILE *fp;
};

static char fs_gamedir[MAX_OSPATH];
static cvar_t *fs_basedir;
static cvar_t *fs_cddir;
cvar_t *fs_gamedirvar;
int file_from_pak;			/* True if the last opened file was in a pak file */

#ifdef SIN
static cvar_t *fs_filemask;
static cvar_t *fs_dircmd;
static cvar_t *cl_player;
static cvar_t *fs_archivedir;
static qboolean archive;
static char fs_relativeplayerdir[MAX_OSPATH];
static char fs_playerdir[MAX_OSPATH];
#endif

typedef struct filelink_t {
	struct filelink_t *next;		/* Next entry in the chain */
	char *from;						/* Pointer to the before string */
	int fromlength;					/* Length of the before string */
	char *to;						/* Pointer to the after string */
} filelink_t;

static filelink_t *fs_links;		/* Linked list of filename overrides */

typedef struct files_t {
	char **filenames;				/* All the filenames in this directory */
	Word numfiles;					/* Number of files found */
} files_t;

typedef struct searchpath_t {
	char filename[MAX_OSPATH];		/* Name of the path or the pack file */
	RezFile *pack;					/* Zero if directory, else pack file record */
	files_t *files;					/* Directory array */
	struct searchpath_t *next;		/* Linked list entry */
} searchpath_t;

static searchpath_t *fs_searchpaths;		/* Main directory search tree */
static searchpath_t *fs_base_searchpaths;	/* Internal game data search tree (No game paths) */
static searchpath_t *fs_temp_path;			/* Working directory path */
static int glob_match_after_star(const char *pattern, const char *text);

/**********************************

All of Quake's data access is through a hierchal file system, but the contents of the file system 
can be transparently merged from several sources.

The "base directory" is the path to the directory holding the quake.exe and all game directories.
The sys_* files pass this to host_init in quakeparms_t->basedir.  This can be overridden with the
"-basedir" command line parm to allow code debugging in a different directory.  The base directory is
only used during filesystem initialization.

The "game directory" is the first tree on the search path and directory that all generated files
(savegames, screenshots, demos, config files) will be saved to.  This can be overridden with the
"-game" command line parameter.  The game directory can never be changed while quake is executing.
This is a precacution against having a malicious server instruct clients to write files over areas they shouldn't.

**********************************/

/**********************************

	Takes an explicit (not game tree related) path to a pak file.

	Loads the header and directory, adding the files at the beginning
	of the list so they override previous pack files.
	
	To optimize searching and memory usage, I will concatinate all
	of the filenames into a singlular chunk, sort them and then unify
	this whole mess to reduce memory size.
	
**********************************/

static int SortPackfile(const void *One,const void *Two)
{
	return Burger::StringCaseCompare(((packfile_t *)One)->filename,((packfile_t *)Two)->filename);		/* For sorting */
}

static int SortCharArray(const void *One,const void *Two)
{
	return Burger::StringCaseCompare(((char **)One)[0],((char **)Two)[0]);		/* For sorting */
}

static RezFile *FS_LoadPackFile (char *packfile)
{
#if 1
	char PathName[2048];
	RezFile *Output;
	QuakePathToBurgerPath(PathName,sizeof(PathName),packfile);
	Output = RezFile::New(PathName,0);
	if (Output) {
		Output->LogDecompressor(1,DInflate);
	}
	return Output;
#else
	dpackheader_t header;
	Word i;
	packfile_t *newfiles;		/* Pointer to new records */
	packfile_t *DestPtr;		/* Work pointer to new records */
	Word numpackfiles;			/* Number of files found */
	pack_t *pack;				/* Pointer to return */
	FILE *packhandle;			/* Open file handle */
	dpackfile_t *info;			/* Disk file record */
	void **infohand;			/* Handle to temp disk file record */
	Word32 MemSize;			/* Amount of memory to use */
	dpackfile_t *WorkPtr;		/* Temp disk file record pointer */
	char *TextPtr;				/* Work pointer to string buffers */
	
#ifdef NO_ADDONS
	unsigned checksum;			/* Anti-hack checksum */
#endif
	
	packhandle = QuakeOpenAFile(packfile, "rb");
	if (!packhandle) {
		return 0;			/* Surrender now! */
	}

	fread (&header, 1, sizeof(header), packhandle);		/* Read in the 12 byte header */
	if (LoadIntelLongPtr(&header.ident) != SINPAKHEADER) {
		Com_Error (ERR_FATAL, "%s is not a Sin packfile", packfile);
	}
#if defined(__BIGENDIAN__)
	header.dirofs = LoadIntelLongPtr(&header.dirofs);
	header.dirlen = LoadIntelLongPtr(&header.dirlen);
#endif

	numpackfiles = header.dirlen / sizeof(dpackfile_t);		/* Each entry is 128 bytes in size */
	if (!numpackfiles) {		/* Oh please!!! */
		fclose(packhandle);
		return 0;
	}
	infohand = AllocAHandle( numpackfiles * sizeof(dpackfile_t));		/* Get a temp buffer */
	if (!infohand) {
		fclose(packhandle);
		return 0;
	}
	info = ( dpackfile_t * )LockAHandle(infohand);			/* Prevent fragmentation */
	fseek (packhandle, header.dirofs, SEEK_SET);			/* Go to the dictionary */
	fread (info, 1, header.dirlen, packhandle);				/* Read in the data */

#ifdef NO_ADDONS
// crc the directory to check for modifications
	checksum = Com_BlockChecksum ((void *)info, header.dirlen);
	
	if (checksum != PAK0_CHECKSUM) {
		Free(info);
		fclose(packhandle);
		return NULL;
	}
#endif

/* At this point, I will get the amount of memory needed to save the filenames */

	MemSize = 0;		/* Init the memory size */
	i = numpackfiles;
	WorkPtr = info;
	do {
		MemSize += FastStrLen(WorkPtr->name)+1;		/* Get the memory size for each string */
		++WorkPtr;
	} while (--i);									/* Count them! */
	
	/* Now get the memory */
	
	newfiles = (packfile_t *)Alloc(MemSize + (numpackfiles * sizeof(packfile_t)));
	TextPtr = (char *)newfiles + numpackfiles*sizeof(packfile_t);
	// parse the directory
	WorkPtr = info;
	DestPtr = newfiles;
	i = numpackfiles;
	do {
		Burger::StringCopy(TextPtr,WorkPtr->name);
		DestPtr->filename = TextPtr;
		DestPtr->filepos = LoadIntelLongPtr(&WorkPtr->filepos);
		DestPtr->filelen = LoadIntelLongPtr(&WorkPtr->filelen);
		TextPtr += FastStrLen(TextPtr)+1;
		++WorkPtr;
		++DestPtr;
	} while (--i);
	pack = (pack_t *)Alloc(sizeof (pack_t));
	Burger::StringCopy (pack->filename, packfile);
	pack->handle = packhandle;
	pack->numfiles = numpackfiles;
	pack->files = newfiles;
	
	DeallocAHandle(infohand);
	Com_Printf("Added packfile %s (%i files)\n", packfile, numpackfiles);

/* Final issue, now I need to sort all of the pack file entries */
/* so that a binary search can be performed, hence speeding up the search dramatically */

	qsort(newfiles,numpackfiles,sizeof(packfile_t),SortPackfile);
	return pack;
#endif
}

/**********************************

	Delete a pack_t structure
		
**********************************/

static void FS_DeletePack(RezFile *Input)
{
	Delete(Input);
}

/**********************************

	Given a directory name, grab all the filenames that are present
	in it...
	Return the number of names found
	
**********************************/

static Word GrabFilenames(StreamHandle_t *Output,const char *DirName)
{
	Burger::DirectorySearch MyDir;
	char temp[256];
	Word Count;
	Word i;
		
	QuakePathToBurgerPath(temp,sizeof(temp),DirName);
	Count = 0;
	if (!MyDir.Open(temp)) {
		i =  Burger::StringLength(DirName);
		while (!MyDir.GetNextEntry()) {
			FastMemCpy(temp,DirName,i+1);
			if (i && temp[i-1]!='/') {
				temp[i] = '/';
				temp[i+1] = 0;
			}
			strcat(temp,MyDir.m_Name);
			if (MyDir.m_bDir) {
				Count += GrabFilenames(Output,temp);
			} else {
				char *Foo;
				Foo = strchr(temp+2,'/');
				if (!Foo) {
					Foo = temp;
				} else {
					++Foo;
				}
				StreamHandlePutMem(Output,Foo,Burger::StringLength(Foo)+1);
				++Count;
			}
		}
		MyDir.Close();
	}
	return Count;	
}

/**********************************

	Sets fs_gamedir, adds the directory to the head of the path,
	then loads and adds pak1.pak pak2.pak ...
		
**********************************/

static void FS_AddGameDirectory (const char *dir)
{
	Word i;
	searchpath_t *search;
	RezFile *pak;
	files_t *FilesPtr;
	Word Count;
	char pakfile[MAX_OSPATH];
	StreamHandle_t MyStream;

	Burger::StringCopy(fs_gamedir, dir);

	//
	// add any pak files in the format pak0.pak pak1.pak, ...
	//
	i = 0;
	do {
		Com_sprintf (pakfile, sizeof(pakfile), "%s/pak%i.rez", dir, i);
		pak = FS_LoadPackFile (pakfile);
		if (pak) {
			search = (searchpath_t *)Alloc(sizeof(searchpath_t));
			Burger::StringCopy(search->filename,pakfile);		/* Copy the filename */
			search->pack = pak;							/* It's a pak file */
			search->files = 0;
			search->next = fs_searchpaths;				/* Link it in */
			fs_searchpaths = search;
		}
	} while (++i<32);
	
	//
	// add the directory to the search path
	//
	
	StreamHandleInitPut(&MyStream);
	Count = GrabFilenames(&MyStream,dir);
	FilesPtr = 0;
	if (Count) {
		StreamHandleEndSave(&MyStream);			/* End the saving */
		FilesPtr = (files_t *)Alloc(sizeof(files_t)+(sizeof(char *)*Count)+StreamHandleGetSize(&MyStream));
		if (FilesPtr) {
			char *SrcPtr;
			char **WorkPtr;
			WorkPtr = (char **)((char *)FilesPtr+sizeof(files_t));
			SrcPtr = (char *)FilesPtr+sizeof(files_t)+(sizeof(char *)*Count);
			FastMemCpy(SrcPtr,MyStream.DataHandle[0],StreamHandleGetSize(&MyStream));		/* Copy the text */
			FilesPtr->numfiles = Count;
			FilesPtr->filenames = WorkPtr;
			i = Count;
			do {
				WorkPtr[0] = SrcPtr;				/* Save the string pointer */
				SrcPtr += Burger::StringLength(SrcPtr)+1;		/* Next string in the list */
				++WorkPtr;							/* Next char pointer */
			} while (--i);
			qsort((char *)FilesPtr+sizeof(files_t),Count,sizeof(char *),SortCharArray);		/* Sort the directory names */
		}
	}
	StreamHandleDestroy(&MyStream);
	search = (searchpath_t *)Alloc (sizeof(searchpath_t));
	Burger::StringCopy (search->filename, dir);
	search->pack = 0;
	search->files = FilesPtr;
	search->next = fs_searchpaths;
	fs_searchpaths = search;
}

/**********************************

	Match the pattern PATTERN against the string TEXT;
	return 1 if it matches, 0 otherwise.
	
	A match means the entire string TEXT is used up in matching.
	
	In the pattern string, `*' matches any sequence of characters,
	`?' matches any character, [SET] matches any character in the specified set,
	[!SET] matches any character not in the specified set.
	
	A set is composed of characters or ranges; a range looks like
	character hyphen character (as in 0-9 or A-Z).
	[0-9a-zA-Z_] is the set of characters allowed in C identifiers.
	Any other character in the pattern must be matched exactly.
	
	To suppress the special syntactic significance of any of `[]*?!-\',
	and match the character exactly, precede it with a `\'.
	
**********************************/

static int glob_match(const char *pattern, const char *text)
{
	register const char *p;
	register const char *t;
	register char c;
	
	p = pattern;
	t = text;
	while ((c = *p++) != '\0') {
		switch (c) {
		case '?':
			if (*t == '\0') {
				return 0;
			}
			++t;
			break;
	
		case '\\':
			if (*p++ != *t++) {
				return 0;
			}
			break;
	
		case '*':
			return glob_match_after_star(p, t);
	
		case '[':
			{
				register char c1 = *t++;
				int invert;
		
				if (!c1) {
					return (0);
				}
				invert = ((*p == '!') || (*p == '^'));
				if (invert)
					p++;
		
				c = *p++;
				while (1) {
					register char cstart = c, cend = c;
		
					if (c == '\\') {
						cstart = *p++;
						cend = cstart;
					}
					if (c == '\0') {
						return 0;
					}
					c = *p++;
					if (c == '-' && *p != ']') {
						cend = *p++;
						if (cend == '\\') {
							cend = *p++;
						}
						if (cend == '\0') {
							return 0;
						}
						c = *p++;
					}
					if (c1 >= cstart && c1 <= cend) {
						goto match;
					}
					if (c == ']') {
						break;
					}
				}
				if (!invert) {
					return 0;
				}
				break;
	
match:	
				/* Skip the rest of the [...] construct that already matched.  */
				while (c != ']') {
					if (c == '\0') {
						return 0;
					}
					c = *p++;
					if (c == '\0') {
						return 0;
					} else if (c == '\\') {
						++p;
					}
				}
				if (invert) {
					return 0;
				}
				break;
			}
	
		default:
			if (c != *t++) {
				return 0;
			}
		}
	}
	return *t == '\0';
}

/**********************************

	Like glob_match, but match PATTERN against any final segment of TEXT.
	
**********************************/

static int glob_match_after_star(const char *pattern, const char *text)
{
	register const char *p;
	register const char *t;
	register char c, c1;
	
	p = pattern;
	t = text;
	while ((c = *p++) == '?' || c == '*') {
		if (c == '?' && *t++ == '\0') {
			return 0;
		}
	}
	if (c == '\0') {
		return 1;
	}
	
	if (c == '\\') {
		c1 = *p;
	} else {
		c1 = c;
	}
	while (1) {
		if (*t == '/') {
			return 0;
		}
		if ((c == '[' || *t == c1) && glob_match(p - 1, t)) {
			return 1;
		}
		if (*t++ == '\0') {
			return 0;
		}
	}
}

/**********************************

	Command "path"
	Show all the links and the search directories
	
**********************************/

static void FS_Path_f (void)
{

	Com_Printf ("Current search path:\n");
	{
		searchpath_t *s;
		s = fs_searchpaths;
		if (s) {
			do {
				if (s == fs_base_searchpaths) {		/* Divider between game paths and system paths? */
					Com_Printf ("----------\n");	/* Show the divider */
				}
				if (s->pack) {						/* Pack file? */
					RezFile *PackPtr;
					PackPtr = s->pack;
					Com_Printf ("(%i files)\n",PackPtr->GetNameArraySize());
				} else {
					Com_Printf ("%s\n", s->filename);		/* Directory only */
				}
				s = s->next;					/* Follow the list */
			} while (s);
		}
	}
	Com_Printf ("\nLinks:\n");
	{
		filelink_t *l;
	
		l=fs_links;
		if (l) {
			do {
				Com_Printf ("%s : %s\n", l->from, l->to);
				l = l->next;
			} while (l);
		}
	}
}

/**********************************

	Command "link"
	Creates a file override record
	
**********************************/

static void FS_Link_f (void)
{
	filelink_t *l;
	Word Argc;
	const char *Argv1;
	
	Argc = Cmd_Argc();
	if (Argc != 3 && Argc!=2) {		/* Must have 3 arguments */
		Com_Printf ("USAGE: link <from> <to>\n");
		return;
	}

	// see if the link allready exists

	Argv1 = Cmd_Argv(1);		/* Get the before string */

	l = fs_links;				/* Start the chain */
	if (l) {
		filelink_t *prev;
		prev = 0;					/* Init the previous record */
		do {
			if (!strcmp(l->from, Argv1)) {		/* Is this a match? */
				Free(l->to);			/* Dispose of the after string */
				if (Argc==2) {					/* delete it? */
					if (prev) {					/* Remove the link */
						prev->next = l->next;
					} else {
						fs_links = l->next;		/* New root link */
					}
					Free(l->from);	/* Dispose of the record */
					Free(l);
					return;						/* Get out now */
				}
				l->to = Burger::StringDuplicate(Cmd_Argv(2));	/* Overwrite the dest */
				return;							/* Exit now */
			}
			prev = l;							/* Save the link */
			l = l->next;
		} while (l);							/* More to check? */
	}

	/* create a new link */

	l = (filelink_t *)Alloc(sizeof(*l));
	l->next = fs_links;				/* Init the forward link */
	fs_links = l;
	l->from = Burger::StringDuplicate(Argv1);		/* Init the record */
	l->fromlength = Burger::StringLength(l->from);
	l->to = Burger::StringDuplicate(Cmd_Argv(2));
}

/**********************************

	Convert a Quake 2 style pathname into a format
	that Burgerlib understands
	
**********************************/

void QuakePathToBurgerPath(char *Output,Word OutputSize,const char *Input)
{
	Word Temp;
	
	if (OutputSize) {
		if (OutputSize>=2) {			/* Must have some room! */
			Output[0] = '9';
			Output[1] = ':';
			Output+=2;
			OutputSize-=2;
			if (OutputSize) {
				if (((Word8 *)Input)[0] == '.') {			/* Starts with a period? */
					Temp = ((Word8 *)Input)[1];
					if (Temp=='/' || Temp=='\\') {		/* End slash */
						Input+=2;				/* Discard the starting ./ */
					} else if (!Temp) {
						++Input;				/* Discard just the period */				
					}
				}
			
				do {
					Temp = ((Word8 *)Input)[0];
					++Input;
					if (Temp == '/' || Temp=='\\') {
						Temp = ':';
					}
					Output[0] = Temp;
					++Output;
					if (!--OutputSize) {
						break;
					}
				} while (Temp);
			}
			if (!OutputSize) {
				--Output;
			}
		}
		Output[0] = 0;
	}
}

Word QuakeResourceLoad(QuakeRezData_t *Output,const char *filename)
{
	searchpath_t *search;
	char netpath[MAX_OSPATH];
	char fixed_filename[MAX_OSPATH];
	char BurgerPath[MAX_OSPATH];
	Word Temp;
	RezFile *pak;
	int i;
	filelink_t *link;
	FILE *fp;
	
	/* Sometimes someone would use PC style pathnames */
	/* Convert the slashes to Quake style */
	
	Output->RezNum = 0;
	Output->Data = 0;
	Output->RezRef = 0;
	Burger::StringCopy(fixed_filename,filename);
	i = 0;
	do {
		Temp = fixed_filename[i];
		if (Temp=='\\') {
			Temp = '/';
			fixed_filename[i] = Temp;
		}
		++i;
	} while (Temp);

	file_from_pak = FALSE;			/* Assume I am loading a file */
	
	// check for links first
	link = fs_links;
	if (link) {
		do {
			if (!strncmp (fixed_filename, link->from, link->fromlength)) {
				Com_sprintf (netpath, sizeof(netpath), "%s%s",link->to, fixed_filename+link->fromlength);
				fp= QuakeOpenAFile(netpath, "rb");
				if (fp) {
					if ( developer->value > 1 ) {
						Com_DPrintf ("link file: %s\n",netpath);
					}
					Temp = Burger::GetSize(fp);
					if (Temp) {
						Output->Data = AllocAHandle(Temp);
						if (Output->Data) {
							fread(LockAHandle(Output->Data),1,Temp,fp);
							fclose(fp);
							UnlockAHandle(Output->Data);
							return FALSE;
						}
					}
					fclose(fp);
				}
				return TRUE;
			}
			link = link->next;
		} while (link);
	}

/* Scan the paths or the pak files for the file */

	search = fs_searchpaths;
	if (search) {
		do {
			if (search->pack) {			/* Pack file? */
			
				/* Note : Packfiles are sorted, therefore, you can perform a binary search */
				/* on it to reduce the number of string compares */
				
				pak = search->pack;			/* Main record */
				QuakePathToBurgerPath(BurgerPath,sizeof(BurgerPath),fixed_filename);
				
				Temp = pak->GetRezNum(BurgerPath);
				if (Temp!=RezFile::INVALIDREZNUM) {
					Output->Data = pak->LoadHandle(Temp);
					if (Output->Data) {
						if ( developer->value > 1 ) {
							Com_sprintf (netpath, sizeof(netpath), "PackFile: %s\n",fixed_filename);
							Com_DPrintf ( netpath );
						}
						Output->RezRef = pak;
						Output->RezNum = Temp;
						return FALSE;
					}
				}
				#if 0
				
				MinRec = 0;					/* Make the bounds */
				MaxRec = pak->numfiles;
				do {
					int Test;								/* For sttring compare test */
					i = ((MaxRec-MinRec)>>1)+MinRec;		/* This is record I will search */
					if (i>=MaxRec) {						/* I don't trust my math */
						i = MinRec;							/* Failsafe */
					}
					PackFilePtr = &pak->files[i];
					Test = Burger::StringCaseCompare(PackFilePtr->filename, fixed_filename);
					if (!Test) {	// found it!
						file_from_pak = TRUE;
						if ( developer->value > 1 ) {
							Com_sprintf (netpath, sizeof(netpath), "PackFile: %s : %s\n",pak->filename, fixed_filename);
							Com_DPrintf ( netpath );
						}
						// open a new file on the pakfile
						file[0] = (FileRef_t *)pak->handle;
						if (!file[0]) {
							Com_Error (ERR_FATAL, "Couldn't reopen %s", pak->filename);
						}
						fseek((FILE *)file[0], PackFilePtr->filepos, SEEK_SET);
						return PackFilePtr->filelen;
					}
					if (Test>0) {		/* Which way to go? */
						MaxRec = i;		/* Must look lower */
					} else {
						MinRec = i+1;	/* Must look higher */
					}
				} while (MinRec<MaxRec);	/* Oh oh... */
				#endif
			} else {
#ifndef NO_ADDONS
				if (search->files) {
					// check a file in the directory tree
					Word MinRec,MaxRec;			/* Bounds for binary search */
					files_t *FilePtr;	/* Work pointer */
					char *FNamePtr;
		
					// developer only stuff
					if (filename[1]!=':' && filename[1]!='/') {
						Com_sprintf (netpath, sizeof(netpath), "%s/%s",search->filename, fixed_filename);
					} else {
						Burger::StringCopy( netpath, fixed_filename );
					}
					if ( archive ) {
						char dstpath[MAX_OSPATH];
						Com_sprintf (dstpath, sizeof(dstpath), "%s/%s",fs_archivedir->CVarString, fixed_filename);
						FS_CopyFile( netpath, dstpath );
					}

					FilePtr = search->files;			/* Main record */
					MinRec = 0;					/* Make the bounds */
					MaxRec = FilePtr->numfiles;
					do {
						int Test;								/* For sttring compare test */
						i = ((MaxRec-MinRec)>>1)+MinRec;		/* This is record I will search */
						if (i>=MaxRec) {						/* I don't trust my math */
							i = MinRec;							/* Failsafe */
						}
						FNamePtr = FilePtr->filenames[i];
						Test = Burger::StringCaseCompare(FNamePtr, fixed_filename);
						if (!Test) {	// found it!
							fp = QuakeOpenAFile (netpath, "rb");
							if (fp) {
								if ( developer->value > 1 ) {
									Com_DPrintf ("FindFile: %s\n",netpath);
								}
								Temp = Burger::GetSize(fp);
								if (Temp) {
									Output->Data = AllocAHandle(Temp);
									if (Output->Data) {
										fread(LockAHandle(Output->Data),1,Temp,fp);
										fclose(fp);
										UnlockAHandle(Output->Data);
										return FALSE;
									}
								}
								fclose(fp);
								return TRUE;
							}
						}
						if (Test>0) {		/* Which way to go? */
							MaxRec = i;		/* Must look lower */
						} else {
							MinRec = i+1;	/* Must look higher */
						}
					} while (MinRec<MaxRec);	/* Oh oh... */

				}
#endif
			}
			search = search->next;		/* Continue looking */
		} while (search);
	}
	
	/* At this point, I surrender, the file is not in the cache */
	/* So I'll make one last effort and just try to manually open the file */

	search = fs_searchpaths;
	if (search) {
		do {
			if (!search->pack) {			/* Pack file? */
				if (filename[1]!=':' && filename[1]!='/') {
					Com_sprintf (netpath, sizeof(netpath), "%s/%s",search->filename, fixed_filename);
				} else {
					Burger::StringCopy( netpath, fixed_filename );
				}
				fp = QuakeOpenAFile (netpath, "rb");
				if (fp) {
					if ( developer->value > 1 ) {
						Com_DPrintf ("FindFile: %s\n",netpath);
					}
					Temp = Burger::GetSize(fp);
					if (Temp) {
						Output->Data = AllocAHandle(Temp);
						if (Output->Data) {
							fread(LockAHandle(Output->Data),1,Temp,fp);
							fclose(fp);
							UnlockAHandle(Output->Data);
							return FALSE;
						}
					}
					fclose(fp);
					return TRUE;
				}
			}
			search = search->next;		/* Continue looking */
		} while (search);
	}
	
	/* That's it! I quit! */
	
	Com_DPrintf ("FindFile: can't find %s\n", fixed_filename);		/* Oh crud, I am a loser */
	return TRUE;
}

/**********************************

	Release a loaded resource file
	
**********************************/

void QuakeResourceRelease(QuakeRezData_t *Input)
{
	if (Input) {				/* Valid pointer? */
		if (Input->RezNum) {	/* Resource? */
			Input->RezRef->Release(Input->RezNum);	/* Release it */
		} else {
			DeallocAHandle(Input->Data);	/* Dispose of the loaded file */
		}
		Input->RezRef = 0;		/* Release the resource number */
		Input->RezNum = 0;
		Input->Data = 0;
	}
}

/**********************************

	Open a standard file using Quake pathnames
	
**********************************/

FILE *QuakeOpenAFile(const char *Filename,const char *Flags)
{
	char newpath[256];
	QuakePathToBurgerPath(newpath,sizeof(newpath),Filename);		/* Convert to burger path */
	return Burger::FileManager::OpenFile(newpath,Flags);
}

/**********************************

	Delete a file using Quake pathnames
	
**********************************/

void QuakeDeleteFile(const char *Filename)
{
	char newpath[256];
	QuakePathToBurgerPath(newpath,sizeof(newpath),Filename);		/* Convert to burger path */
	Burger::FileManager::DeleteFile(newpath);
}

/**********************************

	Given a quake style pathname, create the directory
	tree for the entire path. If the path ends with a slash
	then assume that all names are directory names, otherwise
	assume the final name is an actual filename and not
	a directory
	
**********************************/

void FS_CreatePath (const char *path)
{
	char newpath[ MAX_OSPATH ];
	char *Mark;
	
	QuakePathToBurgerPath(newpath,sizeof(newpath),path);		/* Convert to burger path */

	/* Now, if the LAST character is NOT a colon */
	/* Then I assume the last entry is a filename and therefore not needed */
	/* in the pathname */
	
	Mark = strrchr(newpath,':');
	if (Mark) {
		Mark[1] = 0;		/* Truncate the string after the last colon */
	}
	Burger::FileManager::CreateDirectoryPath(newpath);								/* Create the directory */
}

/**********************************

	Copy a file.
	
**********************************/

qboolean FS_CopyFile(const char *src, const char * dst)
{
	Word Result;
	char srcpath[ MAX_OSPATH ];
	char dstpath[ MAX_OSPATH ];
	
	FS_CreatePath(dst);
	QuakePathToBurgerPath(srcpath,sizeof(srcpath),src);		/* Convert to burger path */
	QuakePathToBurgerPath(dstpath,sizeof(dstpath),dst);		/* Convert to burger path */

	Result = Burger::FileManager::CopyFile(dstpath,srcpath);
	if (!Result) {
		Com_DPrintf ("FS_CopyFile (%s, %s)\n", src, dst);
		return TRUE;
	}	
	return FALSE;
}

/**********************************

	Rename a file.
	
**********************************/

qboolean FS_RenameFile(const char *src, const char * dst)
{
	Word Result;
	char srcpath[ MAX_OSPATH ];
	char dstpath[ MAX_OSPATH ];
	
	FS_CreatePath(dst);
	QuakePathToBurgerPath(srcpath,sizeof(srcpath),src);		/* Convert to burger path */
	QuakePathToBurgerPath(dstpath,sizeof(dstpath),dst);		/* Convert to burger path */

	Result = Burger::FileManager::RenameFile(dstpath,srcpath);
	if (!Result) {
		return TRUE;
	}	
	return FALSE;
}

/**********************************

	Load in a file and return the pointer to it.
	
**********************************/

int FS_LoadFile(const char *path, void **buffer)
{
	QuakeRezData_t MyRez;
	Word8 *buf;
	int len;
	
	buf = NULL;	// quiet compiler warning
	
	// look for it in the filesystem or pack files
	if (QuakeResourceLoad(&MyRez,path)) {
		if (buffer) {
			*buffer = NULL;
		}
		return -1;
	}
	
	len = GetAHandleSize(MyRez.Data);
	if (!buffer) {
		QuakeResourceRelease(&MyRez);
		return len;
	}
	buf = (Word8 *)Alloc(len+1);
	*buffer = buf;
	if (!buf) {
		QuakeResourceRelease(&MyRez);
		return -1;
	}
	FastMemCpy(buf,MyRez.Data[0],len);
	QuakeResourceRelease(&MyRez);
	// Null terminate this thing, cause we need it for scripting.
	buf[len] = 0;
	return len;
}

/**********************************

	Release a file loaded by FS_LoadFile
	
**********************************/

void FS_FreeFile(void *buffer)
{
	Free(buffer);
}

/**********************************

	Load in a file into a stream buffer
	
**********************************/

int QuakeStreamFile(const char *path,StreamHandle_t *Dest)
{
	int len;
	QuakeRezData_t MyRez;
	
	// look for it in the filesystem or pack files
	if (QuakeResourceLoad(&MyRez,path)) {
		return 0;
	}
	len = GetAHandleSize(MyRez.Data);
	StreamHandleInitGetPtr(Dest,LockAHandle(MyRez.Data),len);
	UnlockAHandle(MyRez.Data);
	QuakeResourceRelease(&MyRez);		/* Close the file */
	return len;
}


/**********************************

	Properly handles partial reads
	
**********************************/

#if 0
void FS_Read(void *buffer, int len,FileRef_t *f)
{
	int block, remaining;
	int read;
	Word8 *buf;
	int tries;
	
	buf = (Word8 *)buffer;
	
	// read in chunks for progress bar
	remaining = len;
	tries = 0;
	while (remaining) {
		block = remaining;
		if (block > MAX_READ) {
			block = MAX_READ;
		}
		read = fread(buf,1,block,(FILE *)f);
		if (!read) {
			// we might have been trying to read from a CD
			if (!tries) {
				tries = 1;
				CDAudio_Stop();
			} else {
				Com_Error (ERR_FATAL, "FS_Read: 0 bytes read");
			}
		}
		if (read == -1) {
			Com_Error (ERR_FATAL, "FS_Read: -1 bytes read");
		}

		// do some progress bar thing here...
	
		remaining -= read;
		buf += read;
	}
}
#endif

/**********************************

	Close a file opened by FS_FOpenFile
	
**********************************/

#if 0
void FS_FCloseFile (FileRef_t *f)
{
	searchpath_t *search;
	search = fs_searchpaths;
	if (search) {
		do {
			if (search->pack) {			/* Pack file? */
				if (search->pack->handle==(FILE *)f) {
					return;
				}
			}
			search = search->next;
		} while (search);
	}			
	fclose((FILE *)f);
}
#endif

/**********************************

	Release all of the pack files
	
**********************************/

void FS_Release(void)
{
	searchpath_t *FooPtr;
	FooPtr = fs_searchpaths;
	if (FooPtr) {
		searchpath_t *NextPtr;
		do {
			FS_DeletePack(FooPtr->pack);
			Free(FooPtr->files);
			NextPtr = FooPtr->next;
			Free (FooPtr);
			FooPtr = NextPtr;
		} while (FooPtr);
	}
	fs_temp_path = 0;
	fs_base_searchpaths = 0;
	fs_searchpaths = 0;
}

/**********************************

	Open a file in the search path. The file may
	be a TRUE file or it can be a file in the PAK 
	file. A FileRef_t * is returned and must be 
	released with a call to FS_FCloseFile()
	
	The global file_from_pak is set to TRUE if
	the file came from the PAK file
	
**********************************/

#if 0
int FS_FOpenFile (const char *filename, FileRef_t **file)
{
	searchpath_t *search;
	char netpath[MAX_OSPATH];
	char fixed_filename[MAX_OSPATH];
	Word Temp;
	pack_t *pak;
	int i;
	filelink_t *link;
	
	/* Sometimes someone would use PC style pathnames */
	/* Convert the slashes to Quake style */
	
	Burger::StringCopy(fixed_filename,filename);
	i = 0;
	do {
		Temp = fixed_filename[i];
		if (Temp=='\\') {
			Temp = '/';
			fixed_filename[i] = Temp;
		}
		++i;
	} while (Temp);

	file_from_pak = FALSE;			/* Assume I am loading a file */
	
	// check for links first
	link = fs_links;
	if (link) {
		do {
			if (!strncmp (fixed_filename, link->from, link->fromlength)) {
				Com_sprintf (netpath, sizeof(netpath), "%s%s",link->to, fixed_filename+link->fromlength);
				file[0] = (FileRef_t *)QuakeOpenAFile(netpath, "rb");
				if (file[0]) {
					if ( developer->value > 1 ) {
						Com_DPrintf ("link file: %s\n",netpath);
					}
					return fgetfilesize((FILE *)file[0]);
				}
				return -1;
			}
			link = link->next;
		} while (link);
	}

/* Scan the paths or the pak files for the file */

	search = fs_searchpaths;
	if (search) {
		do {
			if (search->pack) {			/* Pack file? */
			
				/* Note : Packfiles are sorted, therefore, you can perform a binary search */
				/* on it to reduce the number of string compares */
				
				Word MinRec,MaxRec;			/* Bounds for binary search */
				packfile_t *PackFilePtr;	/* Work pointer */

				pak = search->pack;			/* Main record */
				MinRec = 0;					/* Make the bounds */
				MaxRec = pak->numfiles;
				do {
					int Test;								/* For sttring compare test */
					i = ((MaxRec-MinRec)>>1)+MinRec;		/* This is record I will search */
					if (i>=MaxRec) {						/* I don't trust my math */
						i = MinRec;							/* Failsafe */
					}
					PackFilePtr = &pak->files[i];
					Test = Burger::StringCaseCompare(PackFilePtr->filename, fixed_filename);
					if (!Test) {	// found it!
						file_from_pak = TRUE;
						if ( developer->value > 1 ) {
							Com_sprintf (netpath, sizeof(netpath), "PackFile: %s : %s\n",pak->filename, fixed_filename);
							Com_DPrintf ( netpath );
						}
						// open a new file on the pakfile
						file[0] = (FileRef_t *)pak->handle;
						if (!file[0]) {
							Com_Error (ERR_FATAL, "Couldn't reopen %s", pak->filename);
						}
						fseek((FILE *)file[0], PackFilePtr->filepos, SEEK_SET);
						return PackFilePtr->filelen;
					}
					if (Test>0) {		/* Which way to go? */
						MaxRec = i;		/* Must look lower */
					} else {
						MinRec = i+1;	/* Must look higher */
					}
				} while (MinRec<MaxRec);	/* Oh oh... */
			} else {
#ifndef NO_ADDONS
				if (search->files) {
					// check a file in the directory tree
					Word MinRec,MaxRec;			/* Bounds for binary search */
					files_t *FilePtr;	/* Work pointer */
					char *FNamePtr;
		
					// developer only stuff
					if (filename[1]!=':' && filename[1]!='/') {
						Com_sprintf (netpath, sizeof(netpath), "%s/%s",search->filename, fixed_filename);
					} else {
						Burger::StringCopy( netpath, fixed_filename );
					}
					if ( archive ) {
						char dstpath[MAX_OSPATH];
						Com_sprintf (dstpath, sizeof(dstpath), "%s/%s",fs_archivedir->CVarString, fixed_filename);
						FS_CopyFile( netpath, dstpath );
					}

					FilePtr = search->files;			/* Main record */
					MinRec = 0;					/* Make the bounds */
					MaxRec = FilePtr->numfiles;
					do {
						int Test;								/* For sttring compare test */
						i = ((MaxRec-MinRec)>>1)+MinRec;		/* This is record I will search */
						if (i>=MaxRec) {						/* I don't trust my math */
							i = MinRec;							/* Failsafe */
						}
						FNamePtr = FilePtr->filenames[i];
						Test = Burger::StringCaseCompare(FNamePtr, fixed_filename);
						if (!Test) {	// found it!
							file[0] = (FileRef_t *)QuakeOpenAFile (netpath, "rb");
							if (file[0]) {
								if ( developer->value > 1 ) {
									Com_DPrintf ("FindFile: %s\n",netpath);
								}
								return fgetfilesize ((FILE *)file[0]);
							}
						}
						if (Test>0) {		/* Which way to go? */
							MaxRec = i;		/* Must look lower */
						} else {
							MinRec = i+1;	/* Must look higher */
						}
					} while (MinRec<MaxRec);	/* Oh oh... */

				}
#endif
			}
			search = search->next;		/* Continue looking */
		} while (search);
	}
	
	/* At this point, I surrender, the file is not in the cache */
	/* So I'll make one last effort and just try to manually open the file */

	search = fs_searchpaths;
	if (search) {
		do {
			if (!search->pack) {			/* Pack file? */
				if (filename[1]!=':' && filename[1]!='/') {
					Com_sprintf (netpath, sizeof(netpath), "%s/%s",search->filename, fixed_filename);
				} else {
					Burger::StringCopy( netpath, fixed_filename );
				}
				file[0] = (FileRef_t *)QuakeOpenAFile (netpath, "rb");
				if (file[0]) {
					if ( developer->value > 1 ) {
						Com_DPrintf ("FindFile: %s\n",netpath);
					}
					return fgetfilesize ((FILE *)file[0]);
				}
			}
			search = search->next;		/* Continue looking */
		} while (search);
	}
	
	/* That's it! I quit! */
	
	Com_DPrintf ("FindFile: can't find %s\n", fixed_filename);		/* Oh crud, I am a loser */
	*file = NULL;
	return -1;
}
#endif

/**********************************

	Called to find where to write a file (demos, savegames, etc)
	
**********************************/

const char *FS_Gamedir (void)
{
	return fs_gamedir;
}

#ifdef SIN

/**********************************

	Sets the playerdir and path to a different directory.
	
**********************************/

void FS_InitPlayerdir (void)
{
	char path[MAX_OSPATH];
	char playerpath[ MAX_OSPATH ];
	char *WorkPtr;
	Word i;
	
	cl_player = Cvar_Get ("player", "blade", 0);		/* Assume the directory is blade */
	Burger::StringCopy( playerpath, cl_player->CVarString );
	
	// Convert spaces to underscores
	i = Burger::StringLength(playerpath);
	if (i) {
		WorkPtr = playerpath;
		do {
			if (WorkPtr[0] == ' ') {
				WorkPtr[0] = '_';
			}
			++WorkPtr;
		} while (--i);
	}
	
	Com_sprintf( fs_relativeplayerdir, sizeof( fs_relativeplayerdir ), "players/%s", playerpath );
	Com_sprintf( fs_playerdir, sizeof( fs_playerdir ),"%s/%s",FS_Gamedir(), fs_relativeplayerdir );
	Com_sprintf( path, sizeof( path ), "%s/", fs_playerdir );
	FS_CreatePath( path );
}

/**********************************

	Called to find where to write a file (configs, savegames, etc) relative to the basedir
	
**********************************/

const char *FS_RelativePlayerdir (void)
{
	return fs_relativeplayerdir;
}

/**********************************

	Called to find where to write a file (configs, savegames, etc)
	
**********************************/

const char *FS_Playerdir (void)
{
	return fs_playerdir;
}

/**********************************

	Called to get path to the cd drive
	
**********************************/

const char *FS_CDdir (void)
{
	static char name[MAX_QPATH];
	
	name[ 0 ] = 0;
	if ( Burger::StringLength( fs_cddir->CVarString ) > 1 ) {
		if ( fs_cddir->CVarString[ Burger::StringLength( fs_cddir->CVarString ) - 1 ] ==  '/' ) {
			sprintf( name, "%s%s", fs_cddir->CVarString, BASEDIRNAME);
		} else {
			sprintf( name, "%s/%s", fs_cddir->CVarString, BASEDIRNAME);
		}
	}
	return name;
}

#endif

/*
=============
FS_ExecAutoexec
=============
*/
void FS_ExecAutoexec (void)
{
	const char *dir;
	char name [MAX_QPATH];
	
	dir = Cvar_VariableString("gamedir");
	if (*dir)
		Com_sprintf(name, sizeof(name), "%s/%s/autoexec.cfg", fs_basedir->CVarString, dir);
	else
		Com_sprintf(name, sizeof(name), "%s/%s/autoexec.cfg", fs_basedir->CVarString, BASEDIRNAME);
	dir = Sys_FindFirst(name, 0, SFF_SUBDIR | SFF_HIDDEN | SFF_SYSTEM);
	if (dir) {
		Free((void *)dir);
		Cbuf_AddText ("exec autoexec.cfg\n");
	}
	Sys_FindClose();
}

/*
================
FS_SetGamedir

Sets the gamedir and path to a different directory.
================
*/
void FS_SetGamedir (const char *dir)
{
	searchpath_t	*next;

	if (strstr(dir, "..") || strstr(dir, "/")
		|| strstr(dir, "\\") || strstr(dir, ":") ) {
		Com_Printf ("Gamedir should be a single filename, not a path\n");
		return;
	}

	//
	// free up any current game dir info
	//
	while (fs_searchpaths != fs_base_searchpaths) {
		FS_DeletePack(fs_searchpaths->pack);
		Free(fs_searchpaths->files);
		next = fs_searchpaths->next;
		Free (fs_searchpaths);
		fs_searchpaths = next;
	}

	//
	// flush all data, so it will be forced to reload
	//
   //
	if (dedicated && !dedicated->value)
		Cbuf_AddText ("vid_restart\nsnd_restart\n");
	
	Com_sprintf (fs_gamedir, sizeof(fs_gamedir), "%s/%s", fs_basedir->CVarString, dir);
	
	if (!strcmp(dir,BASEDIRNAME) || (*dir == 0)) {
		Cvar_FullSet ("gamedir", "", CVAR_SERVERINFO|CVAR_NOSET);
		Cvar_FullSet ("game", "", CVAR_LATCH|CVAR_SERVERINFO);
	} else {
		Cvar_FullSet ("gamedir", dir, CVAR_SERVERINFO|CVAR_NOSET);
#ifdef SIN
		// 2015 code, add the 2015 game directory if game2015 is set but not a 2015 game
		if ( strcmp( dir, "2015" ) && strstr( dir, "2015" ) ) {
			if ( fs_basedir->CVarString[ Burger::StringLength( fs_basedir->CVarString) - 1 ] == '/' ) {
				FS_AddGameDirectory (va("%s2015", fs_basedir->CVarString) );
			} else {
				FS_AddGameDirectory (va("%s/2015", fs_basedir->CVarString) );
			}
			Cvar_FullSet ("game2015", "1", CVAR_SERVERINFO|CVAR_NOSET);
		}
		if (fs_cddir->CVarString[0]) {
			if ( fs_cddir->CVarString[ Burger::StringLength( fs_cddir->CVarString ) - 1 ] ==  '/' ) {
				FS_AddGameDirectory (va("%s%s", fs_cddir->CVarString, dir) );
			} else {
				FS_AddGameDirectory (va("%s/%s", fs_cddir->CVarString, dir) );
			}
		}
		if ( fs_basedir->CVarString[ Burger::StringLength( fs_basedir->CVarString ) - 1 ] ==  '/' ) {
			FS_AddGameDirectory (va("%s%s", fs_basedir->CVarString, dir) );
		} else {
			FS_AddGameDirectory (va("%s/%s", fs_basedir->CVarString, dir) );
		}
#else
		if (fs_cddir->CVarString[0])
			FS_AddGameDirectory (va("%s/%s", fs_cddir->CVarString, dir) );
		FS_AddGameDirectory (va("%s/%s", fs_basedir->CVarString, dir) );
#endif
	}
}



/*
** FS_ListFiles
*/
static char **FS_ListFiles( char *findname, int *numfiles, unsigned musthave, unsigned canthave )
{
	const char *s;
	int nfiles = 0;
	char **list = 0;
	
	s = Sys_FindFirst( findname, musthave, canthave );
	while ( s ) {
		if ( s[Burger::StringLength(s)-1] != '.' )
			nfiles++;
		Free((void *)s);
		s = Sys_FindNext( musthave, canthave );
	}
	Sys_FindClose ();

	if ( !nfiles )
		return NULL;
	
	nfiles++; // add space for a guard
	*numfiles = nfiles;
	
	list = (char **)Alloc( sizeof( char * ) * nfiles );
	FastMemSet( list, 0, sizeof( char * ) * nfiles );
	
	s = Sys_FindFirst( findname, musthave, canthave );
	nfiles = 0;
	while ( s ) {
      if ( s[Burger::StringLength(s)-1] != '.' )
      {
         list[nfiles] = Burger::StringDuplicate( s );
#ifdef BURGER_WIN32
         Burger::StringLowercase( list[nfiles] );
#endif
         nfiles++;
      }
      Free((void *)s);
      s = Sys_FindNext( musthave, canthave );
   }
   Sys_FindClose ();

   return list;
}

#ifdef SIN
static void FS_AddTempPath (const char *dir)
{
	char tempgamedir[MAX_OSPATH];
	
	if ( fs_temp_path ) {
		Burger::StringCopy (fs_temp_path->filename, dir);
	} else {
		Burger::StringCopy ( tempgamedir, FS_Gamedir() );
		FS_AddGameDirectory( dir );
		fs_temp_path = fs_searchpaths;
		Burger::StringCopy ( fs_gamedir, tempgamedir );
	}
}

static const char *FS_GetTempPath ( void )
{
	if ( fs_temp_path ) {
		return fs_temp_path->filename;
	}
	return "";
}

#endif


//===========================================================


//**********************************
// DIRECTORY FUNCTIONS
//**********************************
#if ( defined(SIN) && defined(BURGER_WIN32) )

#include <io.h>
#include <direct.h>
#include <windows.h>
#include <sys/types.h>
#include <sys/stat.h>
#define NUM_COMPARE_FUNCS 5
static int			dirinclude;
static int			direxclude;
static int			numsorts;
static int			( *compare[ NUM_COMPARE_FUNCS ] )( struct _finddata_t *file1, struct _finddata_t *file2 );
static qboolean	reverse[ NUM_COMPARE_FUNCS ];

static int Dir_SortByName(struct _finddata_t *file1,struct _finddata_t *file2)
{
	return _stricmp( file1->name, file2->name );
}

static int Dir_SortBySize(struct _finddata_t *file1,struct _finddata_t *file2)
{
	return file1->size - file2->size;
}

static int Dir_SortByExtension(struct _finddata_t *file1,struct _finddata_t *file2)
{
	char ext1[ 10 ];
	char ext2[ 10 ];

	Burger::StringCopy( ext1, COM_FileExtension( file1->name ) );
	Burger::StringCopy( ext2, COM_FileExtension( file2->name ) );

	return _stricmp( ext1, ext2 );
}

static int Dir_SortByDate(struct _finddata_t *file1,struct _finddata_t *file2)
{
	return file1->time_write - file2->time_write;
}

static int Dir_DirectoriesFirst(struct _finddata_t *file1,struct _finddata_t *file2)
{
	if ( ( file1->attrib & _A_SUBDIR ) == ( file2->attrib & _A_SUBDIR ) ) {
		return 0;
	}

	if ( file2->attrib & _A_SUBDIR ) {
		return 1;
	}

	return -1;
}

static int Dir_CompareFunc(const void *arg1,const void *arg2 )
{
	struct _finddata_t *file1;
	struct _finddata_t *file2;
	int i;
	int ret;

	file1 = ( struct _finddata_t * )arg1;
	file2 = ( struct _finddata_t * )arg2;

	for( i = 0; i < numsorts; i++ ) {
		ret = compare[ i ]( file1, file2 );
		if ( ret ) {
			return reverse[ i ] ? -ret : ret;
		}
	}
	return 0;
}

static void Dir_Help(void)
{
	Com_Printf ("Displays a list of files and subdirectories in a directory.\n"
					"\n"
					"DIR [drive:][path][filename]\n"
					"  [/A[[:]attributes]]\n"
					"  [/O[[:]sortorder]]\n"
					"\n"
					"  [drive:][path][filename]\n"
					"             Specifies drive,\n"
					"             directory, and/or files\n"
					"             to list.\n"
					"\n"
					"  /A        Displays files with\n"
					"            specified attributes.\n"
					"  attributes D  Directories\n"
					"             R  Read-only files\n"
					"             H  Hidden files\n"
					"             A  Files ready for\n"
					"                archiving\n"
					"             S  System files\n"
					"             -  Prefix meaning not\n"
					"  /O         List by files in sorted\n"
					"             order.\n"
					"\n"
					"  sortorder  N By name\n"
					"             S By size\n"
					"             E By extension\n"
					"             D By date & time\n"
					"             G Group directories\n"
					"             - Prefix to reverse order\n"
					"\n"
					"Switches may be preset in the DIRCMD environment variable.\n"
					"\n"
					);
}

static void Dir_AddAttributeMask(const char *arg)
{
	int		j;
	int		len;
	int		attrib;
	qboolean reverse;

	reverse = FALSE;
	direxclude = 0;
	dirinclude = 0;

	len = Burger::StringLength( arg );
	for( j = 0; j < len; j++ )
		{
		attrib = 0;
		switch( arg[ j ] )
			{
			case 'D' :
			case 'd' :
				attrib = _A_SUBDIR;
				break;

			case 'R' :
			case 'r' :
				attrib = _A_RDONLY;
				break;
			
			case 'H' :
			case 'h' :
				attrib = _A_HIDDEN;
				break;
			
			case 'A' :
			case 'a' :
				attrib = _A_ARCH;
				break;

			case 'S' :
			case 's' :
				attrib = _A_SYSTEM;
				break;

			case '-' :
				reverse = TRUE;
				break;
			}

		if ( attrib )
			{
			if ( reverse )
				{
				direxclude |= attrib;
				dirinclude &= ~attrib;
				reverse = FALSE;
				}
			else
				{
				dirinclude |= attrib;
				direxclude &= ~attrib;
				}
			}
		}
	}

static void Dir_AddSortArg(const char *arg)
{
	int		( *sortfunc )( struct _finddata_t *file1, struct _finddata_t *file2 );
	int		j;
	int		i;
	int		len;
	qboolean reverseflag;

	reverseflag = FALSE;

	len = Burger::StringLength( arg );
	for( j = 0; j < len; j++ )
		{
		sortfunc = NULL;
		switch( arg[ j ] )
			{
			case 'N' :
			case 'n' :
				sortfunc = Dir_SortByName;
				break;

			case 'S' :
			case 's' :
				sortfunc = Dir_SortBySize;
				break;
			
			case 'E' :
			case 'e' :
				sortfunc = Dir_SortByExtension;
				break;
			
			case 'D' :
			case 'd' :
				sortfunc = Dir_SortByDate;
				break;

			case 'G' :
			case 'g' :
				sortfunc = Dir_DirectoriesFirst;
				break;

			case '-' :
				reverseflag = TRUE;
				break;
			}

		if ( sortfunc )
			{
			for( i = 0; i < numsorts; i++ )
				{
				if ( compare[ i ] == sortfunc )
					{
					return;
					}
				}
			reverse[ numsorts ] = reverseflag;
			reverseflag = FALSE;
			compare[ numsorts++ ] = sortfunc;
			}
		}
	}

static qboolean Dir_ParseCommand(const char *arg)
{
	switch( arg[ 1 ] ) {
	case 'O' :
	case 'o' :
		Dir_AddSortArg( &arg[ 2 ] );
		break;

	case 'A' :
	case 'a' :
		Dir_AddAttributeMask( &arg[ 2 ] );
		break;

	case '?' :
	default :
		return FALSE;
		break;
	}

	return TRUE;
}

/*
=================
FS_Dir_f 

This is a big fucking mess.
=================
*/
static void FS_Dir_f(void)
{
	long filehandle;
	struct _finddata_t fileinfo;
	int   numfiles;
	int	  bytes;
	char  currentdir[_MAX_PATH];
	char  originaldir[_MAX_PATH];
	qboolean error;
	char  filemask[_MAX_PATH];
	char  path[_MAX_PATH];
	char  text[1024];
	char  buffer[1024];
	struct tm *tms;
	struct _finddata_t *files;
	int	i;
	int	args;
	const char *arg;
	int	len;
	int	originaldrive;
	int	drive;
	const char *s;

	qboolean parms;

	error = FALSE;
	Burger::StringCopy( filemask, "" );

	numsorts = 0;
	direxclude = _A_HIDDEN | _A_SYSTEM;
	dirinclude = 0;

	args = Cmd_Argc();
	parms = FALSE;
	for( i = 1; i < args; i++ )
		{
		arg = Cmd_Argv( i );
		if ( arg[ 0 ] != '/' && arg[ 0 ] != '-' )
			{
			Burger::StringCopy( filemask, arg );
			}
		else
			{
			parms = TRUE;
			if ( !Dir_ParseCommand( arg ) )
				{
				Dir_Help();
				return;
				}
			}
		}

	if ( !parms )
		{
		Dir_ParseCommand( fs_dircmd->string );
		}


	FastMemSet( originaldir, 0, sizeof( originaldir ) );

   originaldrive = _getdrive();
	drive = -1;

	len = Burger::StringLength( filemask );

	// Get the drive letter
	if ( ( len > 1 ) && ( filemask[ 1 ] == ':' ) )
		{
		drive = toupper( filemask[ 0 ] ) - 'A' + 1;
      if( _chdrive( drive ) )
			{
			Com_Printf ("%s : %s\n\n", filemask, strerror( errno ) );
			return;
			}
		Burger::StringCopy( filemask, &filemask[ 2 ] );
		}

	getcwd( originaldir, _MAX_PATH );
	if ( drive == -1 )
		{
		chdir( FS_GetTempPath() );
		}

	// find the filename
	s = COM_SkipPath( filemask );

	// Get the path
	FastMemSet( path, 0, sizeof( path ) );
	strncpy( path, filemask, s - filemask );

	// Get the filename
	Burger::StringCopy( filemask, s );
	
	// Special case for previous directory
	if ( !strcmp( filemask, ".." ) || !strcmp( filemask, "." ) )
		{
		strcat( path, filemask );
		strcat( path, "\\" );
		Burger::StringCopy( filemask, fs_filemask->string );
		}

	// if the filemask refers to a directory, then we need to handle that
   if( ( path[ 0 ] ) && chdir( path ) )
		{
		Com_Printf ("%s\n\n", strerror( errno ) );
		return;
		}

	if ( !filemask[ 0 ] )
		{
		Burger::StringCopy( filemask, fs_filemask->string );
		}

	if	( !strchr( filemask, '*' ) && !strchr( filemask, '?' ) )
		{
		filehandle = _findfirst( filemask, &fileinfo );
		if ( filehandle != -1 )
			{
			_findclose( filehandle );
			if ( fileinfo.attrib & _A_SUBDIR )
				{
				if( chdir( filemask ) )
					{
					Com_Printf ("%s\n\n", strerror( errno ) );
					return;
					}
				strcat( path, filemask );
				Burger::StringCopy( filemask, fs_filemask->string );
				}
			}
		}

	FastMemSet( currentdir, 0, sizeof( currentdir ) );
	getcwd( currentdir, _MAX_PATH );

	bytes = 0;
	
	Com_Printf ("\n Directory of %s\n\n", currentdir );

	// Count how many files we have
	numfiles = 0;
	filehandle = _findfirst( filemask, &fileinfo );
	if ( filehandle != -1L )
		{
		do
			{
			if ( ( fileinfo.attrib & dirinclude ) == dirinclude && !( fileinfo.attrib & direxclude ) )
				{
				numfiles++;
				bytes += fileinfo.size;
				}
			}
		while( _findnext( filehandle, &fileinfo ) == 0 );
		_findclose( filehandle );
		}

	
	files = ( struct _finddata_t * )AllocClear( sizeof( struct _finddata_t ) * numfiles );

	i = 0;
	buffer[ 0 ] = 0;
	filehandle = _findfirst( filemask, &files[ i ] );
	if ( filehandle != -1L )
		{
		do
			{
			if ( ( files[ i ].attrib & dirinclude ) == dirinclude && !( files[ i ].attrib & direxclude ) )
				{
				i++;
				if ( i >= numfiles )
					{
					break;
					}
				}
			}
		while( _findnext( filehandle, &files[ i ] ) == 0 );

		if ( numsorts )
			{
			qsort( files, numfiles, sizeof( struct _finddata_t ), Dir_CompareFunc );
			}

		for( i = 0; i < numfiles; i++ )
			{
			tms = localtime( &files[ i ].time_write );
			sprintf( text, "%2d/%02d/%02d %2d:%02d%c ",
				tms->tm_mon + 1, tms->tm_mday, tms->tm_year,
				( tms->tm_hour + 12 ) % 12, tms->tm_min,
				( tms->tm_hour >= 12 ) ? 'p' : 'a' );
			strcat( buffer, text );

			if (files[ i ].attrib & _A_SUBDIR)
				{
				sprintf( text, "%6s", "<dir>" );
				}
			else
				{
				if ( files[ i ].size > 1024 * 1024 )
					{
					sprintf( text, "%5.1fm", ( float )files[ i ].size / ( 1024.0 * 1024.0 ) );
					}
				else
					{
					sprintf( text, "%5.1fk", ( float )files[ i ].size / 1024.0 );
					}
				}
			strcat( buffer, text );

			sprintf( text, " %-14s", files[ i ].name );
			strcat( buffer, text );
			
			strcat( buffer, "\n" );

			if ( Burger::StringLength( buffer ) > 512 )
				{
				Com_Printf ( "%s", buffer );
				buffer[ 0 ] = 0;
				}
			}

		sprintf( text, " %d file(s)   %d bytes\n\n", numfiles, bytes );
		strcat( buffer, text );
		Com_Printf ( "%s", buffer );
		}
	else
		{
		Com_Printf ("%s : %s\n\n", filemask, strerror( errno ) );
		}

	_findclose( filehandle );

	chdir( originaldir );
   _chdrive( originaldrive );

	Free (files);
	}

/*
=================
FS_chdir_f 
=================
*/
static void FS_chdir_f(void)
{
	char  dir[_MAX_PATH];
	char  currentdir[_MAX_PATH];
	char  originaldir[_MAX_PATH];

	if ( Cmd_Argc() == 1 ) {
		Com_Printf ("%s\n\n", FS_GetTempPath() );
		return;
	}

	if ( Cmd_Argc() > 2 ) {
		Com_Printf ("chdir [directory name] : change to specified directory\n");
		return;
	}

	FastMemSet( originaldir, 0, sizeof( originaldir ) );
	getcwd( originaldir, _MAX_PATH );
	chdir( FS_GetTempPath() );

	Burger::StringCopy( dir, Cmd_Argv( 1 ) );
	if ( chdir( dir ) != 0 ) {
		Com_Printf ("\n%s : %s\n\n", dir, strerror( errno ) );
	} else {
		getcwd( currentdir, _MAX_PATH );
		FS_AddTempPath( currentdir );
		}

	chdir( originaldir );
}

#else

/*
** FS_Dir_f
*/

static void FS_Dir_f( void )
{
   const char  *path = NULL;
   char  findname[1024];
   char  wildcard[1024] = "*.*";
   char  **dirnames;
   int      ndirs;

   if ( Cmd_Argc() != 1 )
   {
      Burger::StringCopy( wildcard, Cmd_Argv( 1 ) );
   }

   while ( ( path = FS_NextPath( path ) ) != NULL )
   {
      char *tmp = findname;

      Com_sprintf( findname, sizeof(findname), "%s/%s", path, wildcard );

      while ( *tmp != 0 )
      {
         if ( *tmp == '\\' )
            *tmp = '/';
         tmp++;
      }
      Com_Printf( "Directory of %s\n", findname );
      Com_Printf( "----\n" );

      if ( ( dirnames = FS_ListFiles( findname, &ndirs, 0, 0 ) ) != 0 )
      {
         int i;

         for ( i = 0; i < ndirs-1; i++ )
         {
            if ( strrchr( dirnames[i], '/' ) )
               Com_Printf( "%s\n", strrchr( dirnames[i], '/' ) + 1 );
            else
               Com_Printf( "%s\n", dirnames[i] );
			Free(dirnames[i]);
         }
         Free(dirnames);
      }
      Com_Printf( "\n" );
   };
}



#endif



/*
================
FS_NextPath

Allows enumerating all of the directories in the search path
================
*/
const char *FS_NextPath (const char *prevpath)
{
	searchpath_t	*s;
	const char		*prev;

	if (!prevpath)
		return fs_gamedir;

	prev = fs_gamedir;
	for (s=fs_searchpaths ; s ; s=s->next)
	{
		if (s->pack)
			continue;
		if (prevpath == prev)
			return s->filename;
		prev = s->filename;
	}

	return NULL;
}


/*
================
FS_InitFilesystem
================
*/
void FS_InitFilesystem (void)
{
	Cmd_AddCommand ("path", FS_Path_f);
	Cmd_AddCommand ("link", FS_Link_f);
	Cmd_AddCommand ("dir", FS_Dir_f );
#if ( defined(SIN) && defined(BURGER_WIN32) )
	Cmd_AddCommand ("cd", FS_chdir_f );
#endif

#ifdef SIN
	fs_filemask = Cvar_Get ("filemask", "*.*", 0);
	fs_dircmd = Cvar_Get ("dircmd", "", 0);
	//
	// see if we are archiving
	//
	fs_archivedir = Cvar_Get ("archive", "", CVAR_NOSET);
	if ( fs_archivedir->CVarString[ 0 ] )
		archive = TRUE;
	else
		archive = FALSE;
#endif
	
	//
	// basedir <path>
	// allows the game to run from outside the data tree
	//
	fs_basedir = Cvar_Get ("basedir", ".", CVAR_NOSET);
	
	//
	// cddir <path>
	// Logically concatenates the cddir after the basedir for
	// allows the game to run from outside the data tree
	//
	fs_cddir = Cvar_Get ("cddir", "", CVAR_NOSET);
	if (fs_cddir->CVarString[0])
#ifdef SIN
		{
		if ( fs_cddir->CVarString[ Burger::StringLength( fs_cddir->CVarString ) - 1 ] ==  '/' ) {
			FS_AddGameDirectory (va("%s%s", fs_cddir->CVarString, BASEDIRNAME) );
		} else {
			FS_AddGameDirectory (va("%s/%s", fs_cddir->CVarString, BASEDIRNAME) );
		}
	}
#else
		FS_AddGameDirectory (va("%s/"BASEDIRNAME, fs_cddir->CVarString) );
#endif

	//
	// start up with baseq2 by default
	//
#ifdef SIN
   if ( fs_basedir->CVarString[ Burger::StringLength( fs_basedir->CVarString ) - 1 ] ==  '/' )
	   FS_AddGameDirectory (va("%s"BASEDIRNAME, fs_basedir->CVarString) );
   else
	   FS_AddGameDirectory (va("%s/"BASEDIRNAME, fs_basedir->CVarString) );
#else
	FS_AddGameDirectory (va("%s/"BASEDIRNAME, fs_basedir->CVarString) );
#endif

	// any set gamedirs will be freed up to here
	fs_base_searchpaths = fs_searchpaths;

	// check for game override
	fs_gamedirvar = Cvar_Get ("game", "", CVAR_LATCH|CVAR_SERVERINFO);
	if (fs_gamedirvar->CVarString[0])
		FS_SetGamedir (fs_gamedirvar->CVarString);

#ifdef SIN
	FS_InitPlayerdir();
#endif
}

#ifdef SIN



/* Return nonzero if PATTERN has any special globbing chars in it.  */
static int glob_pattern_p(const char *pattern)
{
   register const char *p = pattern;
   register char c;
   int open = 0;

   while ((c = *p++) != '\0')
      switch (c) {
      case '?':
      case '*':
         return 1;

      case '[':      /* Only accept an open brace if there is a close */
         open++;     /* brace to match it.  Bracket expressions must be */
         continue;   /* complete, according to Posix.2 */
      case ']':
         if (open)
            return 1;
         continue;

      case '\\':
         if (*p++ == '\0')
            return 0;
      }

   return 0;
}


#define MAX_FIND_FILES 1024

/*
** FS_FindFiles
*/

const char **FS_FindFiles( const char *path, const char * pattern, int *numfiles )
{
	const char **ListPtr;
	searchpath_t *search;
	const char *ptr;
	RezFile	*pak;
	int	i, j, len, pathlen;
	qboolean found;
	char searchpath[ MAX_OSPATH ];
	char searchpath2[ MAX_OSPATH ];
	char searchpath3[ MAX_OSPATH ];
	char globpattern[ 32 ];
	
	//
	// bullet proof the code
	//
	if ( !path || !pattern || !numfiles )
		return NULL;
	
	// initialize
	*numfiles = 0;
	
	// copy over the path
	// fix the slashes and make sure we end with one
	len = Burger::StringLength( path );
	for ( i = 0; i < len; i++ ) {
		if ( path[ i ] == '\\' ) {
			searchpath[ i ] = '/';
		} else {
			searchpath[ i ] = path[ i ];
		}
	}
	// check for trailing slash
	if ( searchpath[ i-1 ] != '/' ) {
		searchpath[ i ] = '/';
		i++;
	}
	// terminate it
	searchpath[ i ] = '\0';
	pathlen = i;

	for ( i = 0; i < pathlen; i++ ) {
		if ( searchpath[ i ] == '/' ) {
			searchpath2[ i ] = ':';
		} else {
			searchpath2[ i ] = searchpath[ i ];
		}
	}
	searchpath2[i] = 0;
	// copy over the pattern
	strncpy( globpattern, pattern, sizeof( globpattern )-1 );
	// see if the pattern should be simplified
	if ( !strcmp( globpattern, "*.*" ) )
		Burger::StringCopy( globpattern, "*" );
	
	//
	// allocate the space necessary
	//
	ListPtr = (const char **)AllocClear( sizeof( char * ) * MAX_FIND_FILES );
	if ( !ListPtr ) {
		Com_Printf( "Could not allocate list for FS_FindFiles\n" );
		return NULL;
	}
	
	//
	// search through the path, one element at a time
	//
	// TODO BURGER Trace this function to ensure it's working as needed 
	// using the new resource manager
	//

	for (search = fs_searchpaths ; search ; search = search->next) {
		// is the element a pak file?
		if (search->pack) {
			// look through all the pak file elements
			pak = search->pack;
			Word uRezNameCount = pak->GetNameArraySize();
			const RezFile::FilenameToRezNum_t *pRezNameList = pak->GetNameArray();
			for ( i = 0; i < uRezNameCount ; i++ ) {
				if ( !StringCaseCompare( searchpath2,pRezNameList[i].m_pRezName, pathlen ) ) {
					ptr = pRezNameList[i].m_pRezName + pathlen;
					if ( glob_match( pattern, ptr ) ) {
						found = FALSE;
						Burger::StringCopy(searchpath3,pRezNameList[i].m_pRezName);
						j = 0;
						while (searchpath3[j]) {
							if (searchpath3[j]==':') {
								searchpath3[j] = '/';
							}
							++j;
						}
						for( j = 0; j < *numfiles; j++ ) {
							if ( !Burger::StringCaseCompare(searchpath3, ListPtr[ j ] ) ) {
								found = TRUE;
								break;
							}
						}
						if ( !found && *numfiles<MAX_FIND_FILES) {
							ListPtr[ *numfiles ] = Burger::StringDuplicate(searchpath3);
							*numfiles = *numfiles + 1;
						}
					}
				}
			}
		} else {
			char foundname[ MAX_OSPATH ];
			char findname[ MAX_OSPATH ];
			char ** files;
			int numf;
		
			// check a file in the directory tree
			Com_sprintf( findname, sizeof( findname ), "%s/%s%s", search->filename, searchpath, pattern );
			files = FS_ListFiles( findname, &numf, 0, 0 );
		
				// see if it exists in the list and if it doesn't, add it to the list
			if ( files ) {
				for( i = 0; i < ( numf - 1 ); i++ ) {
					ptr = strstr( files[ i ], searchpath );
					// make sure we found the path
					if ( ptr ) {
						Burger::StringCopy( foundname, ptr );
						found = FALSE;
						for( j = 0; j < *numfiles; j++ ) {
							if ( !Burger::StringCaseCompare( ListPtr[ j ], foundname ) ) {
								found = TRUE;
								break;
							}
						}
						if ( !found && *numfiles<MAX_FIND_FILES) {
							ListPtr[ *numfiles ] = Burger::StringDuplicate( foundname );
							*numfiles = *numfiles + 1;
						}
					}
					if ( files[ i ] ) {
						Free(files[i]);
					}
				}
				Free(files);
			}
		}
	}
	return ListPtr;
}

void FS_FreeList( const char **list, int n )
{
	int i;
	
	if (list ) {
		for ( i = 0; i < n; i++ ) {
			if ( list[i] ) {
				Free((void *)list[i]);
				list[i] = 0;
			}
		}
		Free((void *)list);
	}
}
	

/*
===========
FS_GameDirectoryExists

Checks to see if a particular game directory exists
===========
*/

qboolean FS_GameDirectoryExists( const char *gamedirname )
{
	Burger::DirectorySearch MyDir;
	char newpath[512];

	QuakePathToBurgerPath(newpath,sizeof(newpath),gamedirname);		/* Convert to burger path */
	if (!MyDir.Open(newpath)) {
		MyDir.Close();
		return TRUE;
	}
	return FALSE;
}
	
#endif

